/*
 * dbcopy.ec - Program to copy a database table without using too many
 * locks. 
 */
/*
 * RCS Header: 
 * ----------- 
 * $Header: /home/art/GoogleDrive/Latest/RCS/dbcopy.ec,v 1.102 2023/02/13 22:14:46 art Exp $
 * 
 * RCS Log: 
 * -------- 
 * $Log: dbcopy.ec,v $
 * Revision 1.102  2023/02/13 22:14:46  art
 * Fix the fix.
 *
 * Revision 1.101  2023/02/13 22:07:11  art
 * Changed the initialization of selstmt array to avoid possible memory
 * issue.
 *
 * Revision 1.100  2022/08/23 19:53:30  art
 * Fixed problem with -D option.
 *
 * Revision 1.99  2022/05/03 16:39:02  art
 * Changed Usage output to stdout to simplify piping to less/ more.
 * Removed old debug trace.
 * Added new -J option to alter the target table to TYPE(RAW) before
 * the copy and alter it back to TYPE(STANDARD) when the copy completes.
 * If the target is already a RAW table -J is ignored and the table's
 * type is preserved either way. Dbcopy will exit if the alter cannot
 * be processed.
 *
 * Revision 1.98  2021/12/21 17:12:40  art
 * Updated Copyright notice.
 *
 * Revision 1.97  2021/07/05 15:16:08  art
 * Fixed the copy rate calculation and display (-G).
 *
 * Revision 1.96  2021/04/19 22:09:08  art
 * Added -v - verbose option.
 * Added -c - disable source table's active triggers
 * Added -C - disable target table's active triggers
 *   Disabled triggers are re-enabled after copy completes.
 * Added -G - to display run time and data copy rate per second
 * Added -A - lock source table in exclusive mode.
 *
 * Revision 1.95  2021/01/18 17:59:28  art
 * Fixed a trace that should be a debug trace. Thanks to Doug 
 * Lawry for pointing this one out!
 *
 * Revision 1.94  2020/07/30 17:54:33  art
 * Added -G to print data copy rate and total copy time.
 * Added -a to lock the source table.
 *
 * Revision 1.93  2020/01/15 19:22:24  art
 * Replaced int8toasc() with ifx_int8toasc(). The older Informix Corp. function 
 * has finally been deprecated and is no longer supported in the include files.
 *
 * Revision 1.92  2019/03/15 15:44:38  art
 * Fixed some compiler warnings.
 *
 * Revision 1.91  2016/06/27 21:15:27  art
 * Updated the copyright notice.
 *
 * Revision 1.90  2013/12/25 14:25:51  art
 * *** empty log message ***
 *
 * Revision 1.89  2013/12/25 14:05:24  art
 * Fixed compile time warning on HPUX.
 *
 * Revision 1.88  2013/12/24 17:12:57  art
 * Fixed compile time warnings on HPUX.  Updated copyright notice.
 *
 * Revision 1.87  2013/11/20 15:37:03  art
 * Fixed HPUX detection YET AGAIN!
 *
 * Revision 1.86  2013/04/29 19:12:32  art
 * Added #define of atoll() to strtoll() for HPUX which doesn't have atoll.
 *
 * Revision 1.85  2013/02/18 21:24:27  art
 * Added code to make sure this will work as expected with compilers/engines prior to x.70.xC6!
 *
 * Revision 1.84  2013/02/18 21:12:50  art
 * Fixed unnecessary flushes and commits in block flush mode. Looks like big buffers are
 * working on the PUT buffer side!  Predicted buffer size is being used.  But fetch
 * is returning fewer rows than predicted by BigFetBufSize/pos!
 *
 * Revision 1.83  2013/02/18 20:17:28  art
 * OK, big buffers are working somewhat.
 * Can set BigFetBufSize > 32K and calculate FetArrSize up to 32K but max data returned is just over 50K.
 *
 * Revision 1.82  2013/02/12 21:28:55  art
 * Added -E maxerrs option to control the maximum number of errors permitted.
 *
 * Revision 1.81  2013/02/10 22:18:19  art
 * Updated Copyright
 *
 * Revision 1.80  2013/01/21 16:51:39  art
 * fixed WITHOUT REPLICATION code.  Added -O to override the restriction against copying
 * a table into itself (really bad idea, but caviat emptor).
 *
 * Revision 1.79  2011/12/21 20:47:07  art
 * Added new option -R which begins all transactions with the WITHOUT REPLICATION
 * clause which prevents the operation from being replicated via Enterprise
 * Replication.  Thanks to Øyvind Gjerstad for the suggestion.
 *
 * Revision 1.78  2011/06/28 00:17:54  art
 * Fixed a missing #ifdef that caused a compile time problem for esql/c versions
 * earlier than 3.50.
 *
 * Revision 1.77  2011/05/23 16:04:57  art
 * Updated the copyright statement.
 *
 * Revision 1.76  2011/05/23 15:11:32  art
 * Removed a debugging trace left in by mistake.  Removed some redundant DEBUG_DBCOPY output.
 * Added column names do more of the dumprec lines.
 *
 * Revision 1.75  2011/05/22 21:31:04  art
 * typo.
 *
 * Revision 1.74  2011/05/22 21:29:53  art
 * Corrected trace and dump output for INT8/SERIAL8 types.
 *
 * Revision 1.73  2011/05/22 20:18:45  art
 * I think I fixed the problem with DECIMAL numbers.
 *
 * Revision 1.72  2010/08/23 14:00:07  art
 * Fixed a variable declaration that causes compile time problems on some platforms.
 *
 * Revision 1.71  2010/03/24 16:27:19  art
 * Updated copyright notice.
 *
 * Revision 1.70  2010/03/21 04:41:35  art
 * FINALLY fixed DATETIME and INTERVAL handling.  Thanks to Jonathan Leffler for pitching
 * in on this old problem with my understanding and for patching the code.  Jonathan also
 * cleaned up some compile time warnings.
 *
 * Revision 1.68  2010/03/18 11:01:43  art
 * Fixed usage.
 *
 * Revision 1.67  2010/03/18 10:56:28  art
 * Fixed problem with -W & -b.  Removed debugging trace.
 *
 * Revision 1.66  2010/03/18 02:13:25  art
 * Cleanup.
 *
 * Revision 1.65  2010/03/18 01:40:22  art
 * Finally!  Fixed the DATETIME handling once and for all time.  I hope.
 *
 * Revision 1.64  2010/03/17 20:13:09  art
 * Fixed LVARCHAR again.  The ifdef condition protecting older servers/compilers was backwards.
 *
 * Revision 1.63  2009/09/09 18:05:22  art
 * Fixed handling of BOOLEAN columns.  Also improved version detection
 * for compile time handling of BIGINT, INT8, BOOLEAN, etc.
 *
 * Revision 1.62  2009/07/19 14:53:46  art
 * ifdef'd BigFetBufSize to FetBufSize for older ESQL/C Compilers that do not
 * support the new parameter.
 *
 * Revision 1.61  2009/05/20 18:06:52  art
 * typo
 *
 * Revision 1.60  2009/05/20 18:00:22  art
 * Added support for BigFetBufSize to allow fetching more than 32K of data at
 * at time.  Default set to 64K if new -W option is passed.  New option -b
 * will allow sizing the expanded fetch buffer between 32767 and 4,194,303.
 *
 * Revision 1.59  2009/03/25 12:59:12  art
 * Just pretty formatting changes.
 *
 * Revision 1.58  2008/12/10 05:48:22  art
 * Added one to length for variable LVARCHAR data (no -e) to allow space
 * for the NULL terminator if the column was full.
 *
 * Revision 1.57  2008/12/10 05:20:10  art
 * Updated copyright notice.
 *
 * Revision 1.56  2008/12/10 05:16:41  art
 * Made code 64bit compile safe.
 * Added new -e option to control binary/fixed length character copies.
 * Added new -B option to save error records in a binary format that is
 * compatible with ul.ec which can be used to reload the file.
 * Removed some redundant code.
 * Deprecated -L option in favor of "-e char".  Option accepted but ignored.
 *
 * Revision 1.55  2008/04/06 18:24:46  art
 * Fixed default for wait mode to NO WAIT and fixed entry of zero to be default.
 *
 * Revision 1.54  2008/04/06 18:19:36  art
 * Moved source to Galadriel
 *
 * Revision 1.54  2007/10/22 17:31:37  kagel
 * Fixed typo
 *
 * Revision 1.53  2007/10/22 17:30:13  kagel
 * Made string changes from 1.51.1.1 but reversed datetime changes for now.
 *
 * Revision 1.52  2007/10/22 17:18:55  kagel
 * Recheckin 1.50.
 *
 * Revision 1.50  2007/05/22 17:32:19  kagel
 * Added new -I option to support the new IDS 11.10 isolation mode,
 * COMMITTED READ LAST COMMITTED.
 *
 * Revision 1.49  2007/05/14 15:09:29  kagel
 * Made PDQPRIORITY user controlled.  Was set by 10's.
 * Also fixed -w so that -w 0 is default which is NOT WAIT.
 *
 * Revision 1.48  2007/05/14 14:31:55  kagel
 * *** empty log message ***
 *
 * Revision 1.47  2006/07/11 13:22:56  kagel
 * Fixed what output support.
 *
 * Revision 1.46  2006/07/11 13:12:35  kagel
 * Fixed several type casts in diagnostic routines that do not compile
 * correctly on HPUX (probably with an 'H' 32/64bit release of the SDK).
 *
 * Revision 1.45  2006/06/08 17:40:30  kagel
 * Fixed problem with secured connections.  Source connection was not right.
 *
 * Revision 1.44  2006/02/01 18:22:33  kagel
 * Cosmetic.
 *
 * Revision 1.43  2004/10/22 16:34:38  kagel
 * Added -P & -U to enter password and username respectively and -a to cause
 * dbcopy to acquire an exclusive table level lock on the target table.
 *
 * Revision 1.42  2004/10/22 14:50:28  kagel
 * OK, REALLY added INFORMIXPASS and INFORMIXUSER support as in dostats.
 *
 * Revision 1.40  2003/07/15 19:35:18  kagel
 * Lost Log RCS keyword.  Fixed.
 *
 *
 * Revision 1.39  2003/07/15 19:16:07  kagel
 * Fixed SQL & ISAM error printouts to use %d instead of %ld which is wrong in
 * 64bit compiles as the codes are 'int' (32bit) not 'long' (64bit).  Sure
 * wish the 'C' committee had taken my advise and made 'int' 64bit instead
 * of 'long'.
 * 
 * Revision 1.38  2003/03/25 16:02:18  kagel
 * Updated copyright
 *
 * Revision 1.37  2003/03/25 15:59:08  kagel
 * Edited the usage message for -L.
 *
 * Revision 1.36  2003/03/25 15:52:50  kagel
 * Fixed omission of CVCHARTYPE from dumprec() error logging function.  Thanks
 * to Yvind Gjerstad for picking that up.  While in there I decided CVCHARTYPE
 * and CFIXCHAR should not have trailing spaces stripped in the log, but normal
 * CCHARTYPE should.  However, if using CHAR columns to pack binary data, yeah
 * it's dumb but we do it, you do not want to strip.  So default is to continue
 * to treat CHAR columns as FIXCHAR but the new -L option will change that to
 * use CCHARTYPE instead if the user knows there's no binary in the CHAR cols
 * of the table being copied.
 *
 * Revision 1.35  2002/08/20 21:24:03  kagel
 * Fixed problem with copying INFORMIXSERVER to source host if -h/-H not
 * provided.
 *
 * Revision 1.34  2002/08/20 20:27:17  kagel
 * Updated copyright notice
 *
 * Revision 1.33  2002/08/20 19:57:38  kagel
 * Made long object name compatible for use with 9.xx servers.
 *
 * Revision 1.32  2001/10/24 18:43:46  kagel
 * Updated copyright notice.  Improved -? and no args usage message.
 * Protected _H_LOCALEDEF from multiple definition.  Fixed a compiler dependent
 * warning.
 *
 * Revision 1.31  2001/04/10 18:13:45  kagel
 * Fixed log messages messed up in some editor hiccough.
 *
 * Revision 1.30  2001/01/18 23:15:49  kagel
 * Added some new data type support and fixed VARCHARs.  Support added for
 * INTEGER8, SERIAL8, BOOLEAN, NCHAR, NVARCHAR, LVARCHAR
 * 
 */

static const char RCSHeader[] = "@(#)$Header: /home/art/GoogleDrive/Latest/RCS/dbcopy.ec,v 1.102 2023/02/13 22:14:46 art Exp $";
static const char RCSWhat[] = "<@(#) dbcopy.ec,v	$Revision: 1.102 $ > $Date: 2023/02/13 22:14:46 $";

#if defined( _AIX )		/*
				 * Needed as AIX defines a locale control
				 * structure loc_t 
				 */
#if !defined( _H_LOCALEDEF )
#define _H_LOCALEDEF
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include"getopt.h"
#include <sys/time.h>

#include <ifxtypes.h>
EXEC SQL include sqlca;
EXEC SQL include sqlda;
EXEC SQL include "sqltypes.h";
EXEC SQL include decimal;
EXEC SQL include datetime;
EXEC SQL include locator;
EXEC SQL include sqlhdr;

#define FLDSIZE  40
#define BUFFSIZE 256
#define DELIM "|"
#define STMTSIZE 2001
#define DefaultBufSize 32767
#define MaxBufSize 2147483648
#define MaxArrSize 32767

#if defined( DBLOG )
#define DEBUG_LOGGING 1==1
#else
#define DEBUG_LOGGING 1==0
#endif

#if defined( __hpux )
#define atoll( t ) strtoll( (t), (char **)NULL, 10 )
#endif

/*
 * PROTOTYPES 
 */
int             copy_table(void);
void            usage(char *progname);
void            dumprec(struct sqlda *);
int             strip(char *str, int len);
int             charcpy(unsigned char *target, unsigned char *source, int len);
void            dumpSqlvar(struct sqlda *desc, int nrows);
void            dumpSQLDA(char *msg, struct sqlda *desc, int nrows);

/*
 * GLOBALS - I hate these things. 
 */
struct		timeval hold_time, start_time, curr_time;
struct		timezone tz;
double		 dcount, dtime;
FILE           *errlog;
char            Revision[] = "$Revision: 1.102 $";
char            errfile[1025] = "dbcopy.unl";
char            delim[2] = DELIM;
int             begin_ok = 0;
int             commit_ok = 0;	/* True if DB has transactions */
long            display_count = 10000;
int             use_ins_curs = 0;
long            SyncRate = 0;
int             LoseDups = 0;
int		verbose = 0;
int             debugFlag = 0;
int             BlockFlush = 0;
int             Silent = 0;
int             HasBlob = 0;
int             conn5 = 0;
int             AcquireTargLock = 0;
int             AcquireSrcLock = 0;
int		last_committed = 0;
int		FixChar = 1;
int		FixVarChar = 0;
int		FixLVarChar = 0;
int             LogBinary = 0;
int		UseWideBuf = 1;
int		NoSourceTriggers = 0;
int		NoTargetTriggers = 0;
unsigned long 	WideBufSize = 0;
int		DebugDbcopy = 0;
int		no_replicate = 0;
int		PrintProgress = 0;
int		Override = 0;  /* Restrict from copying into the source table! */
long long	MaxErrors = 0;
long long	errcount = 0;
long long	NErr = 0;

EXEC SQL BEGIN DECLARE SECTION;
int             priority = 0;
char            dbname_from[130];
char            dbname_to[130];
char            host_from[1300];
char            dbhost_from[230];
char            host_to[1300];
char            dbhost_to[230];
char            tabname_from[1300];
char            tabname_to[130];
char		trgname[130], trgname2[140];
string		src_trig_list[2049], targ_trig_list[2049];
char            targ_lock_stmt[250];
char            src_lock_stmt[250];
char		targ_no_trig_stmt[250];
char		src_no_trig_stmt[250];
char		targ_trglst_stmt[1000];
char		src_trglst_stmt[1000];
char            qrystmt[103001];
char            selstmt[102001];
char		alter_stmt[200];
char           *insert_stmt;
char            pdqstmt[100];
char            waitstmt[100];
int             towait = 0;
char		username[100] = "", passwd[100] = "";
char		isolation[500];
char           *fet_buf_size_env = (char *)NULL;
long            fet_buf_size_set = 0;
int		do_raw = 0, is_raw = 0;

EXEC SQL END DECLARE SECTION;

int main(int argc, char **argv)
{
    int ch,
	ii,
	errval;
    char *hold;
    selstmt[0] = (char)0;

    if (argc == 1) {
	usage(argv[0]);
	exit(2);
    }
    if (getenv("DBCOPY_DEBUG"))
      DebugDbcopy = 1;
    else
      DebugDbcopy = 0;

    while ((ch = getopt( argc, argv, "5ABCFGIJLORSVWacgvi:D:E:H:P:T:U:b:d:e:f:h:l:p:r:s:t:w:") ) != EOF)
	switch (ch) {
	case 'G':
	  PrintProgress = 1;
	  break;
	case 'c':
	  NoSourceTriggers = 1;
	  break;
	case 'C':
	  NoTargetTriggers = 1;
	  break;
	case 'E':
	    MaxErrors = atoll( optarg );
	    break;
	case 'O':
	    Override = 1;
	    break;
	case 'R':
	    no_replicate = 1;
	    break;
	case 'v':
	    verbose = 1;
	    break;
	case 'V':
	    fprintf( stderr,
		     "\ndbcopy.ec: %s Copyright 1996-2022 by Art S. Kagel.\n",
		     strtok(Revision, "$"));
	    fprintf( stderr,
		     "\tLicense for private usage (including use by "
		     "commercial enterprise\n"
		     "\tfor its own needs) is granted.  Commercial and "
		     "exploitation rights\n"
		     "\tare reserved.\n\n" );
	    exit(0);
	    break;
	case 'W':
	    break;
	case 'b':
	    WideBufSize = strtoul( optarg, &hold, 10 );
	    break;
	case 'B':
	    LogBinary = 1;
	    break;
	case 'I':
	    last_committed = 1;
	    break;
	case 'a':
	    AcquireTargLock = 1;
	    break;
	case 'A':
	    AcquireSrcLock = 1;
	    break;
	case 'U':
	    snprintf( username, sizeof username, "%s", optarg );
	    break;
	case 'P':
	    snprintf( passwd, sizeof passwd, "%s", optarg );
	    break;
	case 'L':
	    break;
	case '5':
	    conn5 = 1;
	    break;
	case 'S':
	    Silent = 1;
	    break;
	case 'i':
	    LoseDups = 1;
	    break;
	case 'r':
	    delim[0] = optarg[0];
	    delim[1] = (char) 0;
	    break;
	case 'l':
	    strcpy(errfile, optarg);
	    break;
	case 'e':
	    if (strcmp( optarg, "varchar" ) == 0) {
		FixVarChar = !FixVarChar;
	    } else if (strcmp( optarg, "lvarchar" ) == 0) {
		FixLVarChar = !FixLVarChar;
	    } else if (strcmp( optarg, "char" ) == 0) {
		FixChar = !FixChar;
	    }
	    break;
	case 'F':
	    BlockFlush = 1;
	    break;
	case 'f':
	    SyncRate = (int) atoi(optarg);
	    break;
	case 'h':
	    /*
	     * Copy from host. 
	     */
	    snprintf(host_from, sizeof host_from, "@%1.99s", optarg);
	    host_from[99] = (char) 0;
	    break;
	case 'H':
	    /*
	     * Copy to host. 
	     */
	    snprintf(host_to, sizeof host_to, "@%1.99s", optarg);
	    host_to[99] = (char) 0;
	    break;
	case 'd':
	    /*
	     * Copy from database. 
	     */
	    snprintf( dbname_from, sizeof dbname_from, "%s", optarg );
	    if (strlen(dbname_to) == 0)
		strcpy(dbname_to, dbname_from);
	    break;
	case 'D':
	    /*
	     * Copy to database. 
	     */
	    snprintf( dbname_to, sizeof dbname_to, "%s", optarg );
	    if (strlen(dbname_from) == 0)
		strcpy(dbname_from, dbname_to);
	    break;
	case 's':
	    snprintf( selstmt, sizeof selstmt, "%s", optarg);
	    break;
	case 't':
	    /*
	     * Table to copy from. 
	     */
	    snprintf( tabname_from, sizeof tabname_from, "%s", optarg );
	    break;
	case 'T':
	    /*
	     * Table to copy to. 
	     */
	    snprintf( tabname_to, sizeof tabname_to, "%s", optarg );
	    break;
	case 'w':
	    towait = atoi(optarg);
	    break;
	case 'p':
	    priority = atoi(optarg);
	    break;
	case 'g':
	    debugFlag = 1;
	    break;
	case 'J':
	    do_raw = 1;
	    break;
	case '?':
	    usage( argv[0] );
	    exit(0);
	}

    fet_buf_size_env = getenv( "FET_BUF_SIZE" );
    if (fet_buf_size_env)
	fet_buf_size_set = atol( fet_buf_size_env );

    if (fet_buf_size_env && fet_buf_size_set) {
	if (WideBufSize) fprintf( stderr, "Overriding -b with FET_BUF_SIZE = %ld.\n", fet_buf_size_set );
	WideBufSize = fet_buf_size_set;
    } else {
	if (WideBufSize < 4096) {
	    if (WideBufSize) {
		fprintf( stderr, "Minimum buffer is 4096! Overridding with minimum." );
		WideBufSize = 4096;
	    } else
		WideBufSize = DefaultBufSize;
	}
    }

    if (UseWideBuf && WideBufSize > MaxBufSize) {
	fprintf( stderr, "Maximum buffer is %llu! Overridding.", (unsigned long long)MaxBufSize );
    } else if (!UseWideBuf && WideBufSize > DefaultBufSize) {
	fprintf( stderr, "Maximum buffer for this C-SDK version is %d! Overridding.\n", (int)32767 );
	WideBufSize = DefaultBufSize;
    }

    if (debugFlag) fprintf( stderr, "FixChar %s\n", FixChar ? "set" : "clear" );
    if (debugFlag) fprintf( stderr, "FixVarChar %s\n", FixVarChar ? "set" : "clear" );
    if (debugFlag) fprintf( stderr, "FixLVarChar %s\n", FixLVarChar ? "set" : "clear" );
    if (debugFlag) fprintf( stderr, "Fetch buffer set to: %lu.\n", WideBufSize );
    if (debugFlag) fprintf( stderr, "Maximum errors: %lld.\n", MaxErrors );

    if (strlen(tabname_from) == 0 && strlen(tabname_to))
	strcpy(tabname_from, tabname_to);
    else if (strlen(tabname_to) == 0 && strlen(tabname_from))
	strcpy(tabname_to, tabname_from);
    if (strlen(tabname_to) == 0) {
	fprintf(stderr, "Neither -t nor -T option specified!\n");
	usage( argv[0] );
	exit(22);
    }

    if (strlen(host_to) == 0) {
	snprintf(host_to, sizeof host_to, "@%s", getenv("INFORMIXSERVER"));
    }

    if (!Override && 
	strcmp( host_from, host_to ) == 0 &&
	strcmp( dbname_from, dbname_to ) == 0 &&
	strcmp( tabname_from, tabname_to ) == 0)
    {
	fprintf( stderr, "Cannot copy from a table to itself without the override flag!\n" );
	usage( argv[0] );
	exit(22);
    }

    if (SyncRate) {
	display_count = SyncRate;
	fprintf( stdout, "Commit every %ld rows.\n", display_count );
    }

    if (strlen(selstmt) == 0) {
	/*
	 * Put table name in default statement. 
	 */
	snprintf( qrystmt, sizeof qrystmt, "SELECT * FROM %s", tabname_from);
    } else if (strstr(selstmt, "%s") != (char *) NULL) {
	/*
	 * Put the table name into the user's statement. 
	 */
	snprintf( qrystmt, sizeof qrystmt, selstmt, tabname_from);
    } else {
	/*
	 * Assume that the user has given us a complete statement. 
	 */
	snprintf( qrystmt, sizeof qrystmt, "%s", selstmt);
    }

    if (username[0] == (char)0) {
	hold = getenv( "INFORMIXUSER" );
	if (hold)
	    snprintf( username, sizeof username, "%s", hold );
    }
    
    
    if (passwd[0] == (char)0) {
	hold = getenv( "INFORMIXPASS" );
	if (hold)
	    snprintf( passwd, sizeof passwd, "%s", hold );
    }

    if (passwd[0] == 0 && username[0] != 0) {
	fputs( "Password> ", stderr );
	ii=0;
	do {
	    passwd[ii] = fgetc( stdin );
	    if (passwd[ii] == '\n' )
		break;
	    fputc( '\b', stderr );
	    fputc( '*', stderr );
	    ii++;
	} while (ii < sizeof passwd - 1);
	passwd[ii] = (char)0;
	if (strlen(passwd) == 0) {
	    fprintf( stderr,
		     "Must supply a password if supplying username!\n" );
	    usage( argv[0] );
	    exit( 1 );
	}
    }

    /*
     * Set up connection for target database connection. 
     */
    if (conn5) {
	snprintf(dbhost_to, sizeof dbhost_to, "%s%s", dbname_to, host_to);
	EXEC SQL CONNECT TO :dbhost_to AS "target" WITH CONCURRENT TRANSACTION;
	if (sqlca.sqlcode < 0) {
	    fprintf(stderr,
		    "Error opening target host connection/database: %d, %d\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    return 1;
	}
    } else {
	if (username[0] != 0) 
	    EXEC SQL CONNECT TO :host_to
		AS "target"
		USER :username USING :passwd
		WITH CONCURRENT TRANSACTION;
	else
	    EXEC SQL CONNECT TO :host_to
		AS "target"
		WITH CONCURRENT TRANSACTION;
	if (sqlca.sqlcode < 0) {
	    fprintf(stderr,
		    "Error opening target host (%s) connection: %d, %d\n",
		    host_to,
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    return 1;
	}
	EXEC SQL DATABASE :dbname_to;
	if (sqlca.sqlcode < 0) {
	    fprintf(stderr,
		    "Error opening target database (%s): %d, %d\n",
		    dbname_to,
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    return 1;
	}
    }
    commit_ok = (sqlca.sqlwarn.sqlwarn1 == 'W');
    begin_ok = (commit_ok && sqlca.sqlwarn.sqlwarn2 != 'W');

    if ((!commit_ok || !begin_ok) && no_replicate) {
	fprintf( stderr,
		 "Transactions without replication (-R) specified but database %s %s.  Ignoring.\n",
		 dbname_to,
		 !begin_ok ? "is ANSI mode." : "does not support transactions" );
	no_replicate = 0;
    }

    if (towait) {
	if (towait > 0)
	    snprintf(waitstmt, sizeof waitstmt, "SET LOCK MODE TO WAIT %d", towait);
	else if (towait < 0)
	    strcpy(waitstmt, "SET LOCK MODE TO WAIT");
	EXEC SQL PREPARE s_lock FROM :waitstmt;
	if (sqlca.sqlcode < 0) {
	    fprintf(stderr,
		    "Error PREPARING lock mode statement for target: %d, %d\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    if (sqlca.sqlcode == -201) {
		fprintf(stderr, "%s\nLocation: %d.\n", waitstmt, sqlca.sqlerrd[4]);
	    }
	    return 1;
	}
	EXEC SQL EXECUTE s_lock;
	if (sqlca.sqlcode < 0) {
	    fprintf(stderr,
		    "Error EXECUTING lock mode statement for target: %d, %d\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    return 1;
	}
    } else
        snprintf( waitstmt, sizeof waitstmt, "SET LOCK MODE TO NOT WAIT;" );

    EXEC SQL IFNDEF ifmx_6_0;
    snprintf( pdqstmt, sizeof pdqstmt, "SET PDQPRIORITY %d", priority );
    EXEC SQL EXECUTE IMMEDIATE :pdqstmt;

    if (sqlca.sqlcode < 0) {
	fprintf(stderr,
		"Warning: Error setting PDQPRIORITY on target: %d, %d\n",
		sqlca.sqlcode,
		sqlca.sqlerrd[1]);
    }
    EXEC SQL        ENDIF;

    /*
     * Set up connection for source database connection. 
     */
    if (strlen(host_from) == 0) {
	snprintf(host_from, sizeof host_from, "@%s", getenv("INFORMIXSERVER"));
    }
    if (conn5) {
        snprintf(dbhost_from, sizeof dbhost_from, "%s%s", dbname_from, host_from);
	EXEC SQL CONNECT TO:dbhost_from
	    AS "source"
	    WITH CONCURRENT TRANSACTION;
	if (sqlca.sqlcode < 0) {
	    fprintf(stderr,
		    "Error opening source host connection/database: %d, %d\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    return 1;
	}
    } else {
	if (username[0] != 0) 
	    EXEC SQL CONNECT TO :host_from
		AS "source"
		USER :username USING :passwd
		WITH CONCURRENT TRANSACTION;
	else
	    EXEC SQL CONNECT TO :host_from
		AS "source"
		WITH CONCURRENT TRANSACTION;
	if (sqlca.sqlcode < 0) {
	    fprintf(stderr,
		    "Error opening source host connection(%s): %d, %d\n",
		    host_from,
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    return 1;
	}
	EXEC SQL DATABASE :dbname_from;
	if (sqlca.sqlcode < 0) {
	    fprintf(stderr,
		    "Error opening source database(%s): %d, %d\n",
		    dbname_from,
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    return 1;
	}
    }

    if (last_committed) {
	/* Give it a try.  If it fails, use lock mode, it fails silently. */
	snprintf( isolation,
		  sizeof isolation,
		  "SET ISOLATION COMMITTED READ LAST COMMITTED" );
	EXEC SQL EXECUTE IMMEDIATE :isolation;
	if  (sqlca.sqlcode < 0) {
	    fprintf(stderr,
		    "Warning LAST COMMITTED isolation mode not available: %d, %d.  Ignoring.\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	}
    }
    if (towait) {
	EXEC SQL PREPARE s_lock2 FROM :waitstmt;
	if (sqlca.sqlcode < 0) {
	    fprintf(stderr,
		    "Error PREPARING lock mode statement for source: %d, %d\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    return 1;
	}
	EXEC SQL EXECUTE s_lock2;
	if (sqlca.sqlcode < 0) {
	    fprintf(stderr,
		    "Error EXECUTING lock mode statement for source: %d, %d\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    return 1;
	}
    }
    EXEC SQL IFNDEF ifmx_6_0;
    EXEC SQL EXECUTE IMMEDIATE :pdqstmt;

    if (sqlca.sqlcode < 0) {
	fprintf(stderr,
		"Warning setting PDQPRIORITY on source: %d, %d\n",
		sqlca.sqlcode,
		sqlca.sqlerrd[1]);
    }
    EXEC SQL        ENDIF;

    EXEC SQL SET    CONNECTION "source";
    if (sqlca.sqlcode != 0) {
	fprintf(stderr,
		"Error: %d(%d) on SET CONNECTION TO source.\n",
		sqlca.sqlcode,
		sqlca.sqlerrd[1]);
	return 1;
    }
    errval = copy_table();
    if (errlog != (FILE *) NULL) {
	fflush(errlog);
	fclose(errlog);
    }
    return errval;
}

/*
 * ---------------------------------------------------------------------- 
 */

int 
copy_table(void)
{
    int             pos;
    char           *cp;
    int             i,
    j,
    nrows = 0,
    mrows = 0,
    stmtsize = STMTSIZE,
    siz;
    struct sqlda   *udesc,
    *cdesc;
    struct sqlvar_struct *col,
    *col2;
    char           *buffer = NULL;
    long            counter = 0,
	thiscnt = 0,
	ii = 0,
	errlvl = 0,
	reccount = 0,
	dupslost = 0;
    long            uncommitted = 0;
    loc_t          *loc;
    char          **bufary,
    *ptr;
    short         **ipary;

    if (!Silent) {
	printf("Selecting data from %s%s:%s with:\n\t%s;\n\n",
	       dbname_from,
	       host_from,
	       tabname_from,
	       qrystmt);
	fflush(stdout);
    }
    EXEC SQL PREPARE fetcher FROM :qrystmt;
    if (sqlca.sqlcode != 0) {
	fprintf(stderr,
		"Error: %d(%d) on prepare of fetch.\n",
		sqlca.sqlcode,
		sqlca.sqlerrd[1]);
	if (sqlca.sqlcode == -201)
	    fprintf(stderr,
		    "Error at char %d in:\n%s\n",
		    sqlca.sqlerrd[4],
		    qrystmt);
	return 1;
    }
    EXEC SQL DESCRIBE fetcher INTO udesc;
    if (sqlca.sqlcode != 0) {
	fprintf(stderr,
		"Error: %d(%d) on describe.\n",
		sqlca.sqlcode,
		sqlca.sqlerrd[1]);
	return 1;
    }
    /*
     * Create an sqlda struture for the insert statement that does not use 
     * the  insert array code. 
     */
    cdesc = (struct sqlda *) malloc(sizeof(struct sqlda));
    memset(cdesc, 0, sizeof(struct sqlda));
    cdesc->sqld = udesc->sqld;
    cdesc->sqlvar = (struct sqlvar_struct *)
	malloc(cdesc->sqld * sizeof(struct sqlvar_struct));
    memset(cdesc->sqlvar, 0, cdesc->sqld * sizeof(struct sqlvar_struct));
    strncpy(cdesc->desc_name, udesc->desc_name, 19);
    cdesc->desc_occ = udesc->desc_occ;
    cdesc->desc_next = (struct sqlda *) NULL;

    /*
     * Step 1: analyze udesc to determine type and memory requirements *
     *       of each item in the select list.  rtypalign() returns a *
     *     pointer to the next appropriate boundary (starting at *
     * pos) for the indicated data type. 
     */
    pos = 0;

    /*
     * Allocate pointer array to hold the fetch array pointers. 
     */
    if (DebugDbcopy)
      dumpSQLDA("udesc", udesc, 0);

    for (col = udesc->sqlvar, i = 0, col2 = cdesc->sqlvar;
	 i < udesc->sqld;
	 col++, i++, col2++) {
        int clen;

	switch (col->sqltype & SQLTYPE) {
	  case SQLSMFLOAT:
	      col->sqltype = CFLOATTYPE;
	      col2->sqltype = CFLOATTYPE;
	      col2->sqllen = col->sqllen = rtypmsize(col->sqltype, col->sqllen);
	      clen = col2->sqllen;
	      break;

	  case SQLFLOAT:
	      col->sqltype = CDOUBLETYPE;
	      col2->sqltype = CDOUBLETYPE;
	      col2->sqllen = col->sqllen = rtypmsize(col->sqltype, col->sqllen);
	      clen = col2->sqllen;
	      break;

	  case SQLMONEY:
	  case SQLDECIMAL:
	      col->sqltype = CDECIMALTYPE;
	      col2->sqltype = CDECIMALTYPE;
	      col2->sqllen = col->sqllen = rtypmsize(col->sqltype, col->sqllen);
	      clen = col2->sqllen;
	      break;

#if defined( SQLINT8 )
	  case SQLSERIAL8:
	  case SQLINT8:
	      col->sqltype = CINT8TYPE;
	      col2->sqltype = CINT8TYPE;
	      col2->sqllen = col->sqllen = rtypmsize(col->sqltype, col->sqllen);
	      clen = col2->sqllen;
	      break;
#endif

#if defined( SQLINFXBIGINT )
	  case SQLBIGSERIAL:
	  case SQLINFXBIGINT:
	      col->sqltype =CBIGINTTYPE;
	      col2->sqltype = CBIGINTTYPE;
	      col2->sqllen = col->sqllen = rtypmsize(col->sqltype, col->sqllen);
	      clen = col2->sqllen;
	      break;
#endif

	  case SQLSERIAL:
	  case SQLINT:
	      col->sqltype = CINTTYPE;
	      col2->sqltype = CINTTYPE;
	      col2->sqllen = col->sqllen = rtypmsize(col->sqltype, col->sqllen);
	      clen = col2->sqllen;
	      break;

	  case SQLSMINT:
	      col->sqltype = CSHORTTYPE;
	      col2->sqltype = CSHORTTYPE;
	      col2->sqllen = col->sqllen = rtypmsize(col->sqltype, col->sqllen);
	      clen = col2->sqllen;
	      break;

#if defined( SQLBOOL )
	  case SQLBOOL:
	      //	      col->sqltype = CINTTYPE;
	      //	      col2->sqltype = CINTTYPE;
	      //	      col->sqltype = CFIXCHARTYPE;
	      //	      col2->sqltype = CFIXCHARTYPE;
	      col->sqltype = CBOOLTYPE;
	      col2->sqltype = CBOOLTYPE;
	      col2->sqllen = col->sqllen = rtypmsize(col->sqltype, col->sqllen);
	      clen = col2->sqllen;
	      break;
#endif

	  case SQLVCHAR:
	  case SQLNVCHAR:
	      if (FixVarChar) {
		  col->sqltype = CFIXCHARTYPE;	/* get all bytes */
		  col2->sqltype = CFIXCHARTYPE;	/* get all bytes */
	      } else {
		  col->sqltype = CVCHARTYPE;	/* get all bytes */
		  col2->sqltype = CVCHARTYPE;	/* get all bytes */
	      }
	      col2->sqllen = col->sqllen = rtypmsize(col->sqltype, col->sqllen);
	      clen = col2->sqllen;
	      break;

	  case SQLCHAR:
	  case SQLNCHAR:
	      if (FixChar) {
		  col->sqltype = CFIXCHARTYPE;	/* get all bytes */
		  col2->sqltype = CFIXCHARTYPE;	/* get all bytes */
	      } else {
		  col->sqltype = CCHARTYPE;	/* get all bytes */
		  col2->sqltype = CCHARTYPE;	/* get all bytes */
	      }
	      col2->sqllen = col->sqllen = rtypmsize(col->sqltype, col->sqllen);
	      clen = col2->sqllen;
	      break;

#if defined( SQLLVARCHAR )
	  case SQLLVARCHAR:
	      if (FixLVarChar) {
		  col->sqltype = CFIXCHARTYPE;	/* get all bytes */
		  col2->sqltype = CFIXCHARTYPE;	/* get all bytes */
	      } else {
		  col->sqltype = CLVCHARTYPE;
		  col2->sqltype = CLVCHARTYPE;
		  col->sqllen += 1;  /* Allow for the NULL! */
		  col2->sqllen += 1;  /* Allow for the NULL! */
	      }
	      col2->sqllen = col->sqllen = rtypmsize(col->sqltype, col->sqllen);
	      clen = col2->sqllen;
	      break;
#endif

	  case SQLDATE:
	      col->sqltype = CDATETYPE;
	      col2->sqltype = CDATETYPE;
	      col2->sqllen = col->sqllen = rtypmsize(col->sqltype, col->sqllen);
	      clen = col2->sqllen;
	      break;

	  case SQLDTIME:
	      col->sqltype = CDTIMETYPE;
	      col2->sqltype = CDTIMETYPE;
	      col2->sqllen = col->sqllen;
	      clen = sizeof(dtime_t);
	      if (DebugDbcopy) fprintf( stderr, "New datetime code col(%d), sqllen = %d(0x%x).\n", i, col2->sqllen, col2->sqllen );
	      if (DebugDbcopy) fprintf( stderr, "\tNew datetime code, C size = %d.\n", clen );
	      break;

	  case SQLINTERVAL:
	      col->sqltype = CINVTYPE;
	      col2->sqltype = CINVTYPE;
	      col2->sqllen = col->sqllen;
	      clen = sizeof(intrvl_t);
	      if (DebugDbcopy) fprintf( stderr, "New interval code col(%d), sqllen = %d(0x%x).\n", i, col2->sqllen, col2->sqllen );
	      if (DebugDbcopy) fprintf( stderr, "\tNew interval code, C size = %d.\n", clen );
	      break;

	  case SQLBYTES:
	  case SQLTEXT:
	      col->sqltype = CLOCATORTYPE;
	      col2->sqltype = CLOCATORTYPE;
	      col2->sqllen = col->sqllen = rtypmsize(col->sqltype, col->sqllen);
	      HasBlob++;
	      clen = col2->sqllen;
	      break;

	  default:
	      /*
	       * Other data types.  Do nothing if we see an unsupported
	       * type. 
	       */
	      fprintf(stderr, "copy_table: Invalid data type: %d \n",
		      col->sqltype);
	      return 1;
	}
	if (DebugDbcopy) {
	    fprintf( stderr,
		     "Col%02d (Typ %d): Adding %d(len) + %d(align) bytes to pos (%d).\n",
		     i,
		     col2->sqltype,
		     clen,
		     (int) rtypalign(pos, col2->sqltype),
		     pos );
	    fprintf( stderr, 
		     "Line: %d rtypesize = %d (struct decimal), DECLEN = %d, sqllen = 0x%x.\n",
		     __LINE__,
		     rtypmsize(col2->sqltype, col2->sqllen ),
		     DECLENGTH( col2->sqllen ),
		     col2->sqllen );
	}
	pos = (int) rtypalign(pos, col2->sqltype) + clen;
	if (DebugDbcopy) fprintf( stderr, "   After align, pos=%d.\n", pos );
	col2->sqlname = (char *) malloc(strlen(col->sqlname) + 1);
	strcpy(col2->sqlname, col->sqlname);
    }

    /*
     * Step 2: Allocate memory to receive rows of the table returned *
     *     by the SELECT statement.  The variable pos has an integer *
     *     value equal to the number of bytes occupied by the row. 
     */
    /*
     * Allocate a buffer for the insert sqlda structure. 
     */
    buffer = malloc(pos);
    memset(buffer, 0, pos);
    /*
     * OK buffer is now the insert buffer.  Bufary is now the array of
     * pointers to the fetch buffers for each column. 
     */
    /*
     * Set FetBufSize and FetArrSize and allocate bufary. 
     */
    if (UseWideBuf && pos > WideBufSize) {
	/*
	 * For large records, can only fetch one at a time. 
	 */
	BigFetBufSize = FetBufSize = pos;
	FetArrSize = 1;
    } else if (!UseWideBuf && pos > DefaultBufSize) {
	/*
	 * For large records, can only fetch one at a time. 
	 */
	BigFetBufSize = DefaultBufSize;
	FetArrSize = 1;
    } else {
	if (UseWideBuf) {
	    FetArrSize = (WideBufSize / pos);
	    if ((WideBufSize / pos) > MaxArrSize)
		FetArrSize = MaxArrSize;
	    BigFetBufSize = (long)FetArrSize * pos;
	    if (BigFetBufSize > DefaultBufSize)
		FetBufSize = DefaultBufSize;
	    else
		FetBufSize = (short)BigFetBufSize;
	} else {
	    FetArrSize = (DefaultBufSize / pos);
	    BigFetBufSize = FetBufSize = FetArrSize * pos;
	}
    }

    if (DebugDbcopy || debugFlag) {
	fprintf( stderr,
		 "Fetch Array set to %ld rows.\nBufsize set to %ld bytes, Bigbufsize set to %ld bytes.\n",
		 (long)FetArrSize,
		 (long)FetBufSize,
		 (long)BigFetBufSize );
	fprintf( stderr, "Rowsize = %d.\n", pos );
    }
    if (BlockFlush && !Silent) {
	printf("Block flush enabled.  Blocks contain %d rows.  "
	       "Rowsize is %d bytes.\n",
	       FetArrSize,
	       pos);
	fflush(stdout);
    }
    /*
     * Bufary needs one pointer per column in the result set. 
     */
    bufary = (char **) malloc(udesc->sqld * sizeof(char *));

    if (DebugDbcopy)
	fprintf(stderr,
		"Malloc'd and cleared %d bytes from %p.\n",
		pos,
		buffer);
    /*
     * Allocate indicator list. 
     */
    /*
     * Array of indicators for fetch sqlda_t structure. 
     */
    ipary = (short **) malloc(udesc->sqld * sizeof(short *));
    if (buffer == NULL || ipary == NULL) {
	fprintf(stderr,
		"Out of memory (%p, %p)\n",
		buffer,
		ipary );
	return 1;
    }
    /*
     * Step 3: Set pointers in the allocated memory to receive each *
     *    item in the select list. 
     */
    loc = (loc_t *) malloc(sizeof(loc_t));
    /*
     * Initialize pointer to insert buffer. 
     */
    cp = buffer;
    for (col = udesc->sqlvar, col2 = cdesc->sqlvar, i = 0;
	 i < udesc->sqld;
	 col2++, col++, i++) {
        int tmp, clen;

	if (col->sqltype == CDTIMETYPE)
	    clen = sizeof(dtime_t);
	else if (col->sqltype == CINVTYPE)
	    clen = sizeof(intrvl_t);
	else if (col->sqltype == CDECIMALTYPE)
	    clen = sizeof(dec_t);
	else
	    clen = col->sqllen;

	/*
	 * Allocate Fetch array for this column. 
	 */
	if (DebugDbcopy) {
	    fprintf( stderr, "Allocating Bufary element for col[%02d] of %d.\n", i, (clen * FetArrSize) );
	    fprintf( stderr, "    FetArrSize=%d, element=%d.\n", FetArrSize, clen );
	}
	bufary[i] = malloc(clen * FetArrSize);
	if (bufary[i] == NULL) {
	    fprintf( stderr, "Out of memory allocating column fetch arrays.\n" );
	    exit(1);
	}
	if (DebugDbcopy)
	    fprintf( stderr,
		     "bufary[%d] allocated %d bytes at %p.\n", i, (clen * FetArrSize), bufary[i] );
	/*
	 * Allocate indicator array for this column. 
	 */
	ipary[i] = (short *) malloc(FetArrSize * sizeof(short));

	/*
	 * Adjust pointer for each column for alignment in insert sqlda 
	 */
	cp = buffer + (tmp = rtypalign((int) (cp - buffer), col->sqltype));
	if (DebugDbcopy) {
	    fprintf( stderr, "Output for col(%d) at %p, align was %d.\n", i, cp, tmp );
	}
	/*
	 * Assign indicator addresses. 
	 */
	col->sqlind = ipary[i];
	/*
	 * Assign the data pointers. 
	 */
	col->sqldata = bufary[i];
	col2->sqldata = cp;
	if (DebugDbcopy)
	    fprintf(stderr,
		    "Setting insert pointer(%d) to %p for %d bytes.\n",
		    i,
		    cp,
		    clen );
	/*
	 * For any BLOB columns set up to fetch blob to memory. 
	 */
	switch (col->sqltype) {
	case CLOCATORTYPE:
	    /*
	     * Set up locator for fetch sqlda, copy sqlda will be set up
	     * later from the current row. 
	     */
	    loc = (loc_t *) col->sqldata;
	    for (j = 0; j < FetArrSize; j++, loc++) {
		loc->loc_loctype = LOCMEMORY;
		loc->loc_bufsize = -1;
		loc->loc_oflags = 0;
		loc->loc_mflags = 0;
		loc->loc_buffer = (char *) NULL;
	    }
	    break;
	}
	cp += clen;
    }

    if (DebugDbcopy)
	{
	    dumpSQLDA("udesc - after setup", udesc, 0);
	    dumpSQLDA("cdesc - after setup", cdesc, 0);
	}


    if (NoSourceTriggers) {
	if (verbose)
	    fprintf( stderr, "Disabling triggers for source table.\n" );
	snprintf( src_trglst_stmt, sizeof src_trglst_stmt,
		  "SELECT trigname "
		  "FROM systables AS t "
		  "JOIN systriggers AS g "
		  "  ON t.tabid = g.tabid "
		  "  AND t.tabname = '%s' "
		  "LEFT OUTER JOIN sysobjstate AS o "
		  "  ON t.tabid = o.tabid "
		  "  AND o.name = g.trigname "
		  "  AND o.state = 'D' "
		  "WHERE o.tabid IS NULL;",
		  tabname_from );
	EXEC SQL PREPARE strg_s FROM :src_trglst_stmt;
	if (sqlca.sqlcode != 0) {
	    fprintf(stderr,
		    "Error: %d(%d) Preparing source trigger list.\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    errlvl = 1;
	    return errlvl;
	}
	EXEC SQL DECLARE strg_c CURSOR FOR strg_s;
	if (sqlca.sqlcode != 0) {
	    fprintf(stderr,
		    "Error: %d(%d) Declaring cursor for source trigger list.\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    errlvl = 1;
	    return errlvl;
	}
	EXEC SQL OPEN strg_c;
	if (sqlca.sqlcode != 0) {
	    fprintf(stderr,
		    "Error: %d(%d) Opening source trigger list.\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    errlvl = 1;
	    return errlvl;
	}
	src_trig_list[0] = (char)0;
	while (1) {
	    EXEC SQL FETCH strg_c INTO :trgname;
	    if (sqlca.sqlcode == 100) {
		break;
	    } else if (sqlca.sqlcode < 0) {
		fprintf(stderr,
			"Error: %d(%d) Fetching source trigger list.\n",
			sqlca.sqlcode,
			sqlca.sqlerrd[1]);
		fprintf( stderr,
			 "Stmt: %s\n",
			 src_trglst_stmt );
		errlvl = 1;
		return errlvl;
	    } else if ((strlen( src_trig_list ) + strlen( trgname ) + 2) > sizeof src_trig_list ) {
		fprintf( stderr, "Error: Trigger list too long!\n" );
		errlvl = 1;
		return errlvl;
	    }
	    if (strlen(src_trig_list) == 0)
		snprintf( trgname2, sizeof trgname2, "%s", trgname );
	    else
		snprintf( trgname2, sizeof trgname2, ", %s", trgname );
	    strcat( src_trig_list, trgname2 );
	}
	EXEC SQL CLOSE strg_c;
	EXEC SQL FREE strg_c;
	EXEC SQL FREE strg_s;
	
	if (strlen( src_trig_list ) > 0) {
	    snprintf( src_no_trig_stmt, sizeof src_no_trig_stmt, "SET TRIGGERS %s DISABLED;", src_trig_list );
	    EXEC SQL EXECUTE IMMEDIATE :src_no_trig_stmt;
	    if (sqlca.sqlcode != 0) {
		fprintf(stderr,
			"Error: %d(%d) disabling source table triggers.\n",
			sqlca.sqlcode,
			sqlca.sqlerrd[1]);
		fprintf( stderr, "Stmt: %s\n", src_no_trig_stmt );
		errlvl = 1;
		return errlvl;
	    }
	}
    }
    if (AcquireSrcLock) {
        if (commit_ok) {
	    EXEC SQL BEGIN WORK;
	}
	if (verbose)
	    fprintf( stderr, "Locking source table in exclusive mode.\n" );
	snprintf( src_lock_stmt,
		  sizeof src_lock_stmt,
		  "LOCK TABLE %s IN EXCLUSIVE MODE",
		  tabname_from );
	EXEC SQL EXECUTE IMMEDIATE :src_lock_stmt;
	if (sqlca.sqlcode != 0) {
	    fprintf(stderr,
		    "Error: %d(%d) on acquiring source table lock.\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    errlvl = 1;
	    return errlvl;
	}
    }

    /*
     * Step 4: Fetch each row of the query, convert to ASCII format, and * 
     * write to the output file. 
     */
    EXEC SQL DECLARE fetch_curs CURSOR WITH HOLD FOR fetcher;
    if (sqlca.sqlcode != 0) {
	fprintf(stderr,
		"Error: %d(%d) on declare of fetch_curs.\n",
		sqlca.sqlcode,
		sqlca.sqlerrd[1]);
	return 1;
    }
    gettimeofday( &hold_time, &tz );
    memcpy( &start_time, &hold_time, sizeof hold_time );
    EXEC SQL OPEN   fetch_curs;
    if (sqlca.sqlcode != 0) {
	fprintf(stderr,
		"Error: %d(%d) on OPEN fetch_curs.\n",
		sqlca.sqlcode,
		sqlca.sqlerrd[1]);
	return 1;
    }
    if (!Silent) {
	printf("\n");
	fflush(stdout);
    }
    /*
     * OK, the source connection and statement are setup, now set up the
     * target connection and statement. 
     */

    EXEC SQL SET    CONNECTION "target";
    if (sqlca.sqlcode != 0) {
	fprintf(stderr,
		"Error: %d(%d) on SET CONNECTION TO target.\n",
		sqlca.sqlcode,
		sqlca.sqlerrd[1]);
	return 1;
    }

    if (NoTargetTriggers) {
	if (verbose)
	    fprintf( stderr, "Disabling triggers for target table.\n" );
	snprintf( targ_trglst_stmt, sizeof targ_trglst_stmt,
		  "SELECT trigname "
		  "FROM systables AS t "
		  "JOIN systriggers AS g "
		  "  ON t.tabid = g.tabid "
		  "  AND t.tabname = '%s' "
		  "LEFT OUTER JOIN sysobjstate AS o "
		  "  ON t.tabid = o.tabid "
		  "  AND o.name = g.trigname "
		  "  AND o.state = 'D' "
		  "WHERE o.tabid IS NULL;",
		  tabname_to );
	EXEC SQL PREPARE ttrg_s FROM :targ_trglst_stmt;
	if (sqlca.sqlcode != 0) {
	    fprintf(stderr,
		    "Error: %d(%d) Preparing target trigger list.\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    errlvl = 1;
	    return errlvl;
	}
	EXEC SQL DECLARE ttrg_c CURSOR FOR ttrg_s;
	if (sqlca.sqlcode != 0) {
	    fprintf(stderr,
		    "Error: %d(%d) Declaring cursor for target trigger list.\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    errlvl = 1;
	    return errlvl;
	}
	EXEC SQL OPEN ttrg_c;
	if (sqlca.sqlcode != 0) {
	    fprintf(stderr,
		    "Error: %d(%d) Opening target trigger list.\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    errlvl = 1;
	    return errlvl;
	}
	targ_trig_list[0] = (char)0;
	while (1) {
	    EXEC SQL FETCH ttrg_c INTO :trgname;
	    if (sqlca.sqlcode == 100) {
		break;
	    } else if (sqlca.sqlcode < 0) {
		fprintf(stderr,
			"Error: %d(%d) Fetching target trigger list.\n",
			sqlca.sqlcode,
			sqlca.sqlerrd[1]);
		fprintf( stderr,
			 "Stmt: %s\n",
			 targ_trglst_stmt );
		errlvl = 1;
		return errlvl;
	    } else if ((strlen( targ_trig_list ) + strlen( trgname ) + 2) > sizeof targ_trig_list ) {
		fprintf( stderr, "Error: Trigger list too long!\n" );
		errlvl = 1;
		return errlvl;
	    }
	    if (strlen(targ_trig_list) == 0)
		snprintf( trgname2, sizeof trgname2, "%s", trgname );
	    else
		snprintf( trgname2, sizeof trgname2, ", %s", trgname );
	    strcat( targ_trig_list, trgname2 );
	}
	EXEC SQL CLOSE ttrg_c;
	EXEC SQL FREE ttrg_c;
	EXEC SQL FREE ttrg_s;

	if (strlen( targ_trig_list ) > 0) {
	    snprintf( targ_no_trig_stmt, sizeof targ_no_trig_stmt, "SET TRIGGERS %s DISABLED;", targ_trig_list );
	    EXEC SQL EXECUTE IMMEDIATE :targ_no_trig_stmt;
	    if (sqlca.sqlcode != 0) {
		fprintf(stderr,
			"Error: %d(%d) disabling target table triggers.\n",
			sqlca.sqlcode,
			sqlca.sqlerrd[1]);
		fprintf( stderr, "Stmt: %s\n", targ_no_trig_stmt );
		errlvl = 1;
		return errlvl;
	    }
	}
    }

    insert_stmt = (char *) malloc(stmtsize);
    siz = snprintf(insert_stmt, stmtsize, "INSERT INTO %s ( \n\t\t", tabname_to);
    for (col = udesc->sqlvar, ii = 1; ii < udesc->sqld; col++, ii++) {
	if (stmtsize < (siz + 65)) {
	    stmtsize += 1000;
	    insert_stmt = (char *) realloc(insert_stmt, stmtsize);
	}
	strcat(insert_stmt, col->sqlname);
	strcat(insert_stmt, ", \n\t\t");
	siz += (strlen(col->sqlname) + 5);
    }
    if (stmtsize < (siz + 45)) {
	stmtsize += 1000;
	insert_stmt = (char *) realloc(insert_stmt, stmtsize);
    }
    strcat(insert_stmt, col->sqlname);
    strcat(insert_stmt, "\n\t) values (\n\t\t");
    siz += (strlen(col->sqlname) + 15);
    for (ii = 1; ii < udesc->sqld; ii++) {
	if (stmtsize < (siz + 12)) {
	    stmtsize += 1000;
	    insert_stmt = (char *) realloc(insert_stmt, stmtsize);
	}
	strcat(insert_stmt, "?, ");
	siz += 3;
	if (i % 18 == 0) {
	    strcat(insert_stmt, "\n\t\t");
	    siz += 3;
	}
    }
    strcat(insert_stmt, "? )");
    siz += 3;

    if (!Silent) {
	printf("Inserting data to %s%s:%s with: \n\t%s;\n\n",
	       dbname_to,
	       host_to,
	       tabname_to,
	       insert_stmt);
	fflush(stdout);
    }

    if (do_raw == 1) {
	/* First see if the table is already RAW! */
	EXEC SQL SELECT (bitand( flags, 16) / 16)::INT
	    INTO :is_raw
	    FROM systables
	    WHERE tabname = :tabname_to;
	if (!Silent && is_raw == 1) {
	    printf( "Target table (%s) is already type RAW. Ignoring -J.\n", tabname_to );
	    do_raw = 0;
	} else {
	    snprintf( alter_stmt, sizeof alter_stmt, "ALTER TABLE %s TYPE( RAW );", tabname_to );
	    EXEC SQL EXECUTE IMMEDIATE :alter_stmt;
	    if (sqlca.sqlcode != 0) {
		fprintf(stderr,
			"Error: %d(%d) ALTERing target to RAW.\n",
			sqlca.sqlcode,
			sqlca.sqlerrd[1]);
		return 1;
	    }
	}
    }

    EXEC SQL PREPARE insertstmt FROM :insert_stmt;
    if (sqlca.sqlcode != 0) {
	fprintf(stderr,
		"Error: %d(%d) on PREPARE insertstmt TO target.\n",
		sqlca.sqlcode,
		sqlca.sqlerrd[1]);
	fprintf(stderr, "    sqlerrm: %s.\n", sqlca.sqlerrm );
	return 1;
    }
    EXEC SQL DECLARE insert_curs CURSOR WITH HOLD FOR insertstmt;
    if (sqlca.sqlcode != 0) {
	fprintf(stderr,
		"Error: %d(%d) on DECLARE insert_curs CURSOR.\n",
		sqlca.sqlcode,
		sqlca.sqlerrd[1]);
	return 1;
    }
    if (begin_ok) {
	if (no_replicate)
	    EXEC SQL EXECUTE IMMEDIATE "BEGIN WORK WITHOUT REPLICATION";
	else
	    EXEC SQL BEGIN WORK;
	if (sqlca.sqlcode != 0) {
	    fprintf(stderr,
		    "Error: %d(%d) on BEGIN WORK.\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    return 1;
	}
    }

    if (AcquireTargLock) {
	if (verbose)
	    fprintf( stderr, "Locking target table in exclusive mode.\n" );
	snprintf( targ_lock_stmt,
		  sizeof targ_lock_stmt,
		  "LOCK TABLE %s IN EXCLUSIVE MODE",
		  tabname_to );
	EXEC SQL EXECUTE IMMEDIATE :targ_lock_stmt;
	if (sqlca.sqlcode != 0) {
	    fprintf(stderr,
		    "Error: %d(%d) on acquiring target table lock.\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    errlvl = 1;
 	    return errlvl;
	}
    }

    EXEC SQL OPEN   insert_curs;
    if (sqlca.sqlcode != 0) {
	fprintf(stderr,
		"Error: %d(%d) on OPEN insert_curs CURSOR.\n",
		sqlca.sqlcode,
		sqlca.sqlerrd[1]);
	errlvl = 1;
	return errlvl;
    }
    for (;;) {
	EXEC SQL SET    CONNECTION "source";
	if (sqlca.sqlcode != 0) {
	    fprintf(stderr,
		    "Error: %d(%d) on SET CONNECTION TO source.\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    errlvl = 1;
	    break;
	}
    fetchit:
	fprintf( stderr, "Fetching.\n" );
	EXEC SQL FETCH fetch_curs USING DESCRIPTOR udesc;

	if (sqlca.sqlcode < 0) {
	    fprintf(stderr,
		    "Error fetching row!  Code=%d, ISAM=%d.\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    NErr++;
	    if (MaxErrors && NErr >= MaxErrors) {
		fprintf( stderr, "Maximum error count exceeded!\n" );
		exit(1);
	    }
	    goto fetchit; /* Skip it. */
	    /*break;*/
	} else if (sqlca.sqlcode == 100)
	    break;
	if (debugFlag) fprintf( stderr, "\nFetched %d rows. ", sqlca.sqlerrd[2] );

	reccount += (nrows = sqlca.sqlerrd[2]);

	if (getenv("DbcopyDumpSQLDA")) {
	    dumpSQLDA("After fetch, udesc", udesc, nrows);
	}
	EXEC SQL SET CONNECTION "target";
	if (sqlca.sqlcode != 0) {
	    fprintf(stderr,
		    "Error: %d(%d) on SET CONNECTION TO target.\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    errlvl = 1;
	    break;
	}
	col = udesc->sqlvar;
	col2 = cdesc->sqlvar;
	mrows = 0;
	i = 0;
    re_put:
	for (; i < nrows; i++) {
	    /*
	     * Copy data from FETCH sqlda to INSERT sqlda. 
	     */
	    if (DebugDbcopy)
		fprintf( stderr, "Row number: %d in block of %d.\n", (i + 1), nrows);
	    for (j = 0; j < udesc->sqld; j++) {
	        int adj;
		
		adj = (int)rtypalign( (int)0, col[j].sqltype );
		switch( col[j].sqltype ) {
		case CDTIMETYPE:
		    ptr = col[j].sqldata + (i * (sizeof(dtime_t)));
		    break;
		case CINVTYPE:
		    ptr = col[j].sqldata + (i * (sizeof(intrvl_t)));
		    break;
		case CDECIMALTYPE:
		    ptr = col[j].sqldata + (i * (rtypmsize(col[j].sqltype, col[j].sqllen) + adj));
		    break;
		default:
		    ptr = col[j].sqldata + (i * (col[j].sqllen + adj));
		    break;
		}
		if (DebugDbcopy) {
		    char tstring[50];

		    switch( col[j].sqltype ){
		    case CSHORTTYPE:
			fprintf( stderr, "\tValue is: <%hd>\n", *(short *)ptr );
			break;
		    case CDATETYPE:
		    case CINTTYPE:
			fprintf( stderr, "\tValue is: <%d>\n", *(int *)ptr );
			break;
		    case CFIXCHARTYPE:
			fprintf( stderr, "\tValue is: <%*.*s>\n", col[j].sqllen, col[j].sqllen, (char *)ptr );
			break;
		    case CCHARTYPE:
		    case CVCHARTYPE:
		    case CLVCHARTYPE:
			fprintf( stderr, "\tValue is: <%s>\n", (char *)ptr );
			break;
		    case CDTIMETYPE:
			dttoasc((dtime_t *)ptr, tstring);
			fprintf( stderr, "\tDT Value is(%p): <%s>\n", ptr, tstring );
			break;
		    case CDECIMALTYPE:
			dectoasc( (dec_t *)ptr, tstring, 49, -1 );
			tstring[49] = (char)0;
			fprintf( stderr, "\tDEC Value is: <%s>.\n", tstring );
			break;
		    }
		}

		if (col2[j].sqltype == CDTIMETYPE)
		    memcpy( col2[j].sqldata,
			    ptr,
			    sizeof(dtime_t) );
		else if (col2[j].sqltype == CINVTYPE)
		    memcpy( col2[j].sqldata,
			    ptr,
			    sizeof(intrvl_t) );
		else if (col2[j].sqltype == CDECIMALTYPE)
		    memcpy( col2[j].sqldata,
			    ptr,
			    sizeof(dec_t) );
		else
		    memcpy( col2[j].sqldata,
			    ptr,
			    col2[j].sqllen );

		col2[j].sqlind = &(col[j].sqlind[i]);
	    }

	    EXEC SQL PUT insert_curs USING DESCRIPTOR cdesc;
	    if (sqlca.sqlerrd[2] > 0 && debugFlag)
		fprintf( stderr,
			 "\nAuto-flushed %d rows, uncommitted = %ld.",
			 sqlca.sqlerrd[2],
			 (uncommitted + sqlca.sqlerrd[2]) );
	    if (sqlca.sqlcode || DEBUG_LOGGING) {
		if (!LoseDups || (sqlca.sqlcode != -239 &&
				  sqlca.sqlcode != -268)) {
		    fprintf( stderr,
			     "Error(%d): %d(%d) on PUT insert_curs, "
			     "rec# %d, Msg: %s.\n",
			     __LINE__,
			     sqlca.sqlcode,
			     sqlca.sqlerrd[1],
			     mrows + sqlca.sqlerrd[2] + 1,
			     sqlca.sqlerrm );
		}
		/*
		 * Recover from error in block. 
		 */
		if (BlockFlush) {
		    i = sqlca.sqlerrd[2] + mrows;
		    for (j = 0; j < udesc->sqld; j++) {
			ptr = col[j].sqldata + (i * col[j].sqllen);
			memcpy(col2[j].sqldata, ptr, col[j].sqllen);
			col2[j].sqlind = &(col[j].sqlind[i]);
		    }
		    thiscnt += sqlca.sqlerrd[2];
		    counter += (sqlca.sqlerrd[2]);
		    uncommitted += sqlca.sqlerrd[2];
		    mrows += sqlca.sqlerrd[2] + 1;
		    if (LoseDups && (sqlca.sqlcode == -239 ||
				     sqlca.sqlcode == -268)) {
			dupslost++;
		    } else {
			errcount++;
			dumprec(cdesc);
		    }
		} else {
		    errcount++;
		    if (LoseDups && (sqlca.sqlcode == -239 ||
				     sqlca.sqlcode == -268)) {
			dupslost++;
		    } else {
			dumprec(cdesc);
		    }
		}
	    } else if (!BlockFlush) {
		EXEC SQL FLUSH  insert_curs;
		if (sqlca.sqlcode == 0 || DEBUG_LOGGING) {
		    counter++;
		    thiscnt++;
		    uncommitted++;
		} else if (LoseDups && (sqlca.sqlcode == -239 ||
					sqlca.sqlcode == -268))
		    dupslost++;
		else if (sqlca.sqlcode < 0) {
		    errcount++;
		    fprintf(stderr,
			    "Error: %d(%d) on FLUSH insert_curs, rec# %lld.\n",
			    sqlca.sqlcode,
			    sqlca.sqlerrd[1],
			    (long long)counter + (long long)errcount);
		    dumprec(cdesc);
		}
	    } else if (sqlca.sqlerrd[2] > 0) {
		thiscnt = sqlca.sqlerrd[2];
		counter += (sqlca.sqlerrd[2]);
		uncommitted += sqlca.sqlerrd[2];
		/*
		 * Keep track of beginning of next recovery block within
		 * the input block. 
		 */
		mrows += sqlca.sqlerrd[2];
	    }
	}			/*
				 * End of PUT Loop. 
				 */

	if (display_count && uncommitted >= display_count) {
	    if (BlockFlush) {
		EXEC SQL FLUSH  insert_curs;
		if (debugFlag) fprintf( stderr, "\nPush-flushed %d rows.", sqlca.sqlerrd[2] );
		thiscnt += sqlca.sqlerrd[2];
		counter += (sqlca.sqlerrd[2]);
		uncommitted += sqlca.sqlerrd[2];
		if (sqlca.sqlcode != 0 || DEBUG_LOGGING) {
		    if (!LoseDups || (sqlca.sqlcode != -239 &&
				      sqlca.sqlcode != -268)) {
			fprintf ( stderr,
				  "Error(%d): %d(%d) on PUT insert_curs, "
				  "rec# %d, Msg: %s.\n",
				  __LINE__,
				  sqlca.sqlcode,
				  sqlca.sqlerrd[1],
				  mrows + sqlca.sqlerrd[2] + 1,
				  sqlca.sqlerrm );
		    }
		    /*
		     * Recover from error in block. 
		     */
		    i = sqlca.sqlerrd[2] + mrows;
		    for (j = 0; j < udesc->sqld; j++) {
			ptr = col[j].sqldata + (i * col[j].sqllen);
			memcpy(col2[j].sqldata, ptr, col[j].sqllen);
			col2[j].sqlind = &(col[j].sqlind[i]);
		    }
		    mrows += sqlca.sqlerrd[2] + 1;
		    if (LoseDups && (sqlca.sqlcode == -239 ||
				     sqlca.sqlcode == -268)) {
			dupslost++;
		    } else {
			errcount++;
			dumprec(cdesc);
		    }
		    i++;
		    goto re_put;
		}
	    }
	    gettimeofday( &hold_time, &tz );

	    if (commit_ok) {
		EXEC SQL COMMIT WORK;
		if (sqlca.sqlcode != 0) {
		    fprintf(stderr,
			    "Error: %d(%d) on COMMIT WORK in loop.\n",
			    sqlca.sqlcode,
			    sqlca.sqlerrd[1]);
		    errlvl = 1;
		    break;
		}
		if (begin_ok) {
		    if (no_replicate)
	    		EXEC SQL EXECUTE IMMEDIATE "BEGIN WORK WITHOUT REPLICATION";
		    else
			EXEC SQL BEGIN WORK;
		    if (sqlca.sqlcode != 0) {
			fprintf(stderr,
				"Error: %d(%d) on BEGIN WORK.\n",
				sqlca.sqlcode,
				sqlca.sqlerrd[1]);
			return 1;
		    }
		}

		if (AcquireTargLock) {
		    snprintf( targ_lock_stmt,
			      sizeof targ_lock_stmt,
			      "LOCK TABLE %s IN EXCLUSIVE MODE",
			      tabname_to );
		    EXEC SQL EXECUTE IMMEDIATE :targ_lock_stmt;
		    if (sqlca.sqlcode != 0) {
			fprintf(stderr,
				"Error: %d(%d) on acquiring target table lock.\n",
				sqlca.sqlcode,
				sqlca.sqlerrd[1]);
			errlvl = 1;
			return errlvl;
		    }
		}
	    }
	    if (!Silent) {
	        if (PrintProgress == 1) {
		    gettimeofday( &curr_time, &tz );
		    dtime = ((curr_time.tv_sec - start_time.tv_sec) + ((double)(curr_time.tv_usec - start_time.tv_usec)/1000000.0));
		    dcount = (counter / dtime);
		    printf("Committed: %ld, Rows/sec: %lld\r", counter, (long long)dcount);
	        } else {
		    printf("Committed: %ld\r", counter);
		}
		thiscnt = 0;
		fflush(stdout);
	    }
	    uncommitted = 0;
	}

	
#if defined(DOCLEANUP)
	/*
	 * Reinitialize any Locators. 
	 */
	if (HasBlob) {
	    int             Freed = HasBlob;	/*
						 * Keep track of multiple
						 * BLOB columns. 
						 */

	    for (j = 0; j < udesc->sqld && Freed; j++) {
		if (udesc->sqlvar[j].sqltype == CLOCATORTYPE) {
		    for (i = 0; i < FetArrSize; i++) {
			if (((loc_t *) cdesc->sqlvar[j].sqldata)[i].loc_buffer) {
			    free(((loc_t *) cdesc->sqlvar[j].sqldata)[i].loc_buffer);
			    ((loc_t *) cdesc->sqlvar[j].sqldata)[i].loc_buffer
				= (loc_t *) NULL;
			    ((loc_t *) cdesc->sqlvar[j].sqldata)[i].loc_bufsize
				= -1;
			}
		    }
		    Freed--;
		}
	    }
	}
#endif
    }				/*
				 * end fetch loop 
				 */

    EXEC SQL SET    CONNECTION "target";
    if (sqlca.sqlcode != 0) {
	fprintf(stderr,
		"Error: %d(%d) on SET CONNECTION TO target.\n",
		sqlca.sqlcode,
		sqlca.sqlerrd[1]);
	return 1;
    }
    /*
     * Flush any unflushed rows in the buffer (s/b none). 
     */
    EXEC SQL FLUSH  insert_curs;
    if (debugFlag) fprintf( stderr, "\nFinal-flushed %d rows.", sqlca.sqlerrd[2] );
    thiscnt += sqlca.sqlerrd[2];
    counter += (sqlca.sqlerrd[2]);
    uncommitted += sqlca.sqlerrd[2];
    if (sqlca.sqlcode != 0 || DEBUG_LOGGING) {
	if (!LoseDups || (sqlca.sqlcode != -239 &&
			  sqlca.sqlcode != -268)) {
	    fprintf( stderr,
		     "Error(%d): %d(%d) on PUT insert_curs, "
		     "rec# %lld, Msg: %s.\n",
		     __LINE__,
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1],
		     (long long)counter + (long long)errcount,
		     sqlca.sqlerrm );
	}
	/*
	 * Recover from error in block. 
	 */
	i = sqlca.sqlerrd[2] + mrows;
	for (j = 0; j < udesc->sqld; j++) {
	    ptr = col[j].sqldata + (i * col[j].sqllen);
	    memcpy(col2[j].sqldata, ptr, col[j].sqllen);
	    col2[j].sqlind = &(col[j].sqlind[i]);
	}
	mrows += sqlca.sqlerrd[2] + 1;
	if (LoseDups && (sqlca.sqlcode == -239 || sqlca.sqlcode == -268)) {
	    dupslost++;
	} else {
	    errcount++;
	    dumprec(cdesc);
	}
	i++;
	if (BlockFlush)
	    goto re_put;
    }
    /*
     * Commit any flushed rows not yet committed (there will normally be
     * some). 
     */
    if (commit_ok && uncommitted) {
	EXEC SQL COMMIT WORK;
	if (sqlca.sqlcode != 0) {
	    fprintf(stderr,
		    "Error: %d(%d) on COMMIT WORK after loop.\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    errlvl = 1;
	}
	if (!Silent) {
	    if (PrintProgress == 1) {
		gettimeofday( &curr_time, &tz );
		dtime = ((curr_time.tv_sec - start_time.tv_sec) + ((double)(curr_time.tv_usec - start_time.tv_usec)/1000000.0));
		dcount = (counter / dtime);
		// printf("Committed: %ld, Rows/sec: %lld\r", counter, (long long)dcount);
	    } else {
		// printf("Committed: %ld\r", counter);
	    }
	    thiscnt = 0;
	}
	fflush( stdout );
	uncommitted = 0;
    }
    EXEC SQL CLOSE  insert_curs;
    if (sqlca.sqlcode != 0) {
	fprintf(stderr,
		"Error: %d(%d) on CLOSE insert_curs.\n",
		sqlca.sqlcode,
		sqlca.sqlerrd[1]);
	return 1;
    }
    EXEC SQL FREE   insert_curs;
    if (sqlca.sqlcode != 0) {
	fprintf(stderr,
		"Error: %d(%d) on FREE insert_curs.\n",
		sqlca.sqlcode,
		sqlca.sqlerrd[1]);
	return 1;
    }

    if (!Silent) {
	printf("\nInput:  %ld records.\n", reccount);
	printf("Copied: %ld records to %s%s:%s.\n",
	       counter,
	       dbname_to,
	       host_to,
	       tabname_to);
	if (PrintProgress == 1) {
	  printf("  Runtime: %10.2f seconds.\n", dtime );
	  printf("  Copy rate: %lld rows per second.\n", (long long)(counter / dtime ) );
	}
	if (LoseDups)
	    printf("Ignored %ld duplicated rows.\n", dupslost);
	printf("Logged: %lld records to error log.\n", errcount);
	fflush(stdout);
    }
    if (NoTargetTriggers) {
	if (verbose)
	    fprintf( stderr, "Re-enabling triggers for target table.\n" );
	if (strlen( targ_trig_list ) > 0) {
	    snprintf( targ_no_trig_stmt, sizeof targ_no_trig_stmt, "SET TRIGGERS %s ENABLED;", targ_trig_list );
	    EXEC SQL EXECUTE IMMEDIATE :targ_no_trig_stmt;
	    if (sqlca.sqlcode != 0) {
		fprintf(stderr,
			"Error: %d(%d) enabling target table triggers.\n",
			sqlca.sqlcode,
			sqlca.sqlerrd[1]);
		fprintf( stderr, "Stmt: %s\n", targ_no_trig_stmt );
		errlvl = 1;
		return errlvl;
	    }
	}
    }

    EXEC SQL FREE   insertstmt;
    if (sqlca.sqlcode != 0 && sqlca.sqlcode != -404) {
	fprintf(stderr,
		"Error: %d(%d) on FREE statement insertstmt.\n",
		sqlca.sqlcode,
		sqlca.sqlerrd[1]);
	return 1;
    }

    EXEC SQL SET    CONNECTION "source";
    if (sqlca.sqlcode != 0) {
	fprintf(stderr,
		"Error: %d(%d) on SET CONNECTION TO source.\n",
		sqlca.sqlcode,
		sqlca.sqlerrd[1]);
	return 1;
    }
    EXEC SQL CLOSE  fetchcurs;
    if (sqlca.sqlcode != 0 && sqlca.sqlcode != -404) {
	fprintf(stderr,
		"Error: %d(%d) on CLOSE fetchcurs.\n",
		sqlca.sqlcode,
		sqlca.sqlerrd[1]);
	return 1;
    }
    EXEC SQL FREE   fetchcurs;
    if (sqlca.sqlcode != 0 && sqlca.sqlcode != -404) {
	fprintf(stderr,
		"Error: %d(%d) on FREE fetchcurs.\n",
		sqlca.sqlcode,
		sqlca.sqlerrd[1]);
	return 1;
    }
    EXEC SQL FREE   fetcher;
    if (sqlca.sqlcode != 0 && sqlca.sqlcode != -404) {
	fprintf(stderr,
		"Error: %d(%d) on FREE statments fetcher.\n",
		sqlca.sqlcode,
		sqlca.sqlerrd[1]);
	return 1;
    }

    if (NoSourceTriggers) {
	if (verbose)
	    fprintf( stderr, "Re-enabling triggers for source table.\n" );
	if (strlen( src_trig_list ) > 0) {
	    snprintf( src_no_trig_stmt, sizeof src_no_trig_stmt, "SET TRIGGERS %s ENABLED;", src_trig_list );
	    EXEC SQL EXECUTE IMMEDIATE :src_no_trig_stmt;
	    if (sqlca.sqlcode != 0) {
		fprintf(stderr,
			"Error: %d(%d) enabling source table triggers.\n",
			sqlca.sqlcode,
			sqlca.sqlerrd[1]);
		fprintf( stderr, "Stmt: %s\n", src_no_trig_stmt );
		errlvl = 1;
		return errlvl;
	    }
	}
    }
    if (do_raw == 1) {
	snprintf( alter_stmt, sizeof alter_stmt, "ALTER TABLE %s TYPE (STANDARD);", tabname_to );
	EXEC SQL EXECUTE IMMEDIATE :alter_stmt;
	if (sqlca.sqlcode != 0) {
	    fprintf(stderr,
		    "Error: %d(%d) ALTERing target to STANDARD.\n",
		    sqlca.sqlcode,
		    sqlca.sqlerrd[1]);
	    errlvl = 1;
	}
    }
    return errlvl;
}


/*
 * ----------------------------------------------------------------------- 
 */

int 
strip(char *str, int len)
{
    char           *p;
    int             alen = len;

    for (p = (str + len - 1); p > str; p--, alen--) {
	if (*p == ' ' || *p == (char) 0)
	    *p = (char) 0;
	else
	    break;
    }
    return alen;
}

int 
charcpy(unsigned char *target, unsigned char *source, int len)
{
    int             rlen = 0;

    while (len) {
	if (*source == *delim) {
	    *target++ = '\\';
	    *target = *source;
	    target++;
	} else if (*source < ' ' || *source > '~') {
	    /*
	     * Non-printable, convert to hex. 
	     */
	    target += sprintf((char *) target, "\\%2.2x", *source);
	} else {
	    *target = *source;
	    target++;
	}
	source++;
	len--;
	rlen++;
    }
    /*
     * *target++ = *delim; 
     */
    *target = (char) 0;
    rlen++;

    return rlen;
}

void 
dumprec(struct sqlda *ldesc)
{
    static char    *iobuf = (char *) NULL;
    static int      iobsz = 10240,
	nxt = 1,
	recurs = 0;
    static char     string[2048],
	tstring[1025];

    loc_t          *loc;
    struct sqlvar_struct *col;
    int             i, j,
	slen = 0,
	flen = 0,
	alen;

    NErr++;

    if (iobuf == (char *) NULL) {
	iobuf = (char *) malloc(iobsz);
	if (!iobuf) {
	    fprintf(stderr,
		    "Cannot allocate memory for blob.  Errno=%d\n",
		    errno);
	    exit(errno);
	}
    }
    if (errlog == (FILE *) NULL) {
	errlog = fopen(errfile, "w");
	if (errlog == (FILE *) NULL) {
	    fprintf(stderr, "Cannot open error log file!  Errno=%d", errno);
	    exit(errno);
	}
    }
    if (fseek(errlog, 0, SEEK_END) == -1) {
	fprintf(stderr, "Error seeking end of error log: %d\n", errno);
	fclose(errlog);
	snprintf(string, sizeof string, "%s.%d", errfile, nxt++);
	fprintf(stderr, "Opening next log: %s\n", errfile);
	errlog = fopen(string, "w");
	if (errlog == (FILE *) NULL) {
	    fprintf(stderr,
		    "Cannot open next log file (%d), exiting.\n",
		    errno);
	    fprintf(stderr, "Record in error was:\n%s\n", iobuf);
	    exit(99);
	}
    }
    if (LogBinary) {
        int bufsz = 0;
	char *buff, *buffer, *ptr;
	/*
	 * Determine size of dump buffer.
	 */
	for (col = ldesc->sqlvar, j = 0; j < ldesc->sqld; j++) {
	    bufsz += col[j].sqllen;
	}
	/* Create dump buffer. */
	buff = buffer = (char *)malloc( bufsz );

	/*
	 * Copy data from INSERT sqlda to dump buffer. 
	 */
	for (col = ldesc->sqlvar, j = 0; j < ldesc->sqld; j++) {
	    ptr = col[j].sqldata;
	    memcpy(buff, ptr, col[j].sqllen);
	    buff += col[j].sqllen;
	}
        do {
	    if (fwrite( buffer, bufsz, 1, errlog ) != 1) {
		/* Error writing this record  If no more space open a new file
		   else error out. */
		if (errno == ENOSPC) {
		    fclose(errlog);
		    snprintf(string, sizeof string, "%s.%d", errfile, nxt++);
		    fprintf(stderr, "Opening next log: %s\n", errfile);
		    errlog = fopen(string, "w");
		    if (errlog == (FILE *) NULL) {
			fprintf(stderr,
				"Cannot open next log file (%d), exiting.\n",
				errno);
			fprintf(stderr, "Record in error was:\n%s\n", iobuf);
			exit(99);
		    }
		} else {
		    fprintf( stderr,
			     "Unexpected error: %d encountered logging failed record.\n",
			     errno );
		    exit(99);
		}
	    }
	} while (errno != 0);
    } else {
	for (col = ldesc->sqlvar, i = 0; i < ldesc->sqld; col++, i++) {
	    errno = 0;
	    if (*col->sqlind == -1) {
		flen = 0;
		string[0] = (char) 0;
	    } else {
		switch (col->sqltype) {
		  case CFLOATTYPE:
		      flen = snprintf(string,
				      sizeof string,
				      "%f",
				      (double) *(float *) col->sqldata);
		      if (DebugDbcopy) fprintf( stderr, "ColFLT(%d:%s), len: %d. Val: %f.\n", i, col->sqlname, col->sqllen, *(float *)col->sqldata );
		      break;

		  case CDOUBLETYPE:
		      flen = snprintf(string,
				      sizeof string,
				     "%f",
				     *(double *) col->sqldata);
		      if (DebugDbcopy) fprintf( stderr, "ColDBL(%d:%s), len: %d. Val: %f.\n", i, col->sqlname, col->sqllen, *(double *)col->sqldata );
		      break;

		  case CDECIMALTYPE:
		      dectoasc((dec_t *) col->sqldata, tstring, 1025, -1);
		      strip(tstring, 1024);
		      flen = snprintf(string, sizeof string, "%s", tstring);
		      if (DebugDbcopy) fprintf( stderr, "ColDEC(%d:%s), len: %d. Val: ", i, col->sqlname, rtypmsize(col->sqltype, col->sqllen) );
		      {
			  char decstr[40];

			  if (dectoasc( (dec_t *)col->sqldata, decstr, 33, -1 ) == 0)
			      fprintf( stderr, "%s", decstr );
			  else
			      fprintf( stderr, "<Conversion Error!>" );
		      }
		      fprintf( stderr, ".\n" );
		      break;

		  case CSHORTTYPE:  /* 16 bit */
		      flen = snprintf(string, sizeof string, "%hd", *(short *) col->sqldata);
		      if (DebugDbcopy) fprintf( stderr, "ColSHRT(%d:%s), len: %d. Val: %hd.\n", i, col->sqlname, col->sqllen, *(short *)col->sqldata );
		      break;

		  case CINTTYPE:    /* 32 bit */
		      flen = snprintf(string, sizeof string, "%d", *(int *) col->sqldata);
		      if (DebugDbcopy) fprintf( stderr, "ColINT(%d:%s), len: %d. Val: %d.\n", i, col->sqlname, col->sqllen, *(int *)col->sqldata );
		      break;

#if defined( SQLINT8 )
		  case CINT8TYPE:   /* 64 bit */
		      ifx_int8toasc( (ifx_int8_t *)col->sqldata, (char *)tstring, 20 );
		      flen = snprintf(string, sizeof string, "ColINT8(%d:%s)%s", i, col->sqlname, string );
		      break;
#endif

#if defined( SQLINFXBIGINT )
		  case CBIGINTTYPE:
		      flen = snprintf(string, sizeof string, "ColBIG(%d:%s), %lld", i, col->sqlname, *(long long *) col->sqldata);
		      break;
#endif

		  case CFIXCHARTYPE:
		  case CCHARTYPE:
		  case CVCHARTYPE:
		  case CLVCHARTYPE:
		      if (!FixChar)
			  alen = strip((char *) col->sqldata, col->sqllen);
		      else
			  alen = col->sqllen;
		      flen = charcpy((unsigned char *) string,
				     (unsigned char *) col->sqldata,
				     alen);
		      if (DebugDbcopy) fprintf( stderr, "ColFC(%d:%s), len: %d. Val: %s.\n", i, col->sqlname, col->sqllen, (char *)col->sqldata );
		      break;

		  case CDATETYPE:
		      rdatestr(*(int *) col->sqldata, tstring);
		      flen = snprintf(string, sizeof string, "%s", tstring);
		      if (DebugDbcopy) fprintf( stderr, "ColDATE(%d:%s), len: %d. Val: %d.\n", i, col->sqlname, col->sqllen, *(int *)col->sqldata );
		      break;

		  case CDTIMETYPE:
		      dttoasc((dtime_t *) col->sqldata, tstring);
		      flen = snprintf(string, sizeof string, "%s", tstring);
		      if (DebugDbcopy) fprintf( stderr, "colDT(%d:%s) at: %p.  Val: %s.\n", i, col->sqlname, col->sqldata, string );
		      break;

		  case CINVTYPE:
		      intoasc((intrvl_t *) col->sqldata, tstring);
		      flen = snprintf(string, sizeof string, "%s", tstring);
		      if (DebugDbcopy) fprintf( stderr, "colINTVL(%d:%s) at: %p.  Val: %s.\n", i, col->sqlname, col->sqldata, string );
		      break;

		  case CLOCATORTYPE:
		      loc = (loc_t *) col->sqldata;
		      if (iobsz < (3 * loc->loc_size)) {
			  iobsz = 3 * loc->loc_size;
			  iobuf = realloc(iobuf, iobsz);
		      }
		      flen = charcpy((unsigned char *) iobuf,
				     (unsigned char *) loc->loc_buffer,
				     loc->loc_size);
		      if (!fwrite(iobuf, flen, 1, errlog)) {
			  /*
			   * Recursively call me, once, so whole rec is output. 
			   */
			  if (!recurs) {
			      fprintf(stderr,
				      "Error writing to error log: %d at %d\n",
				      errno,
				      __LINE__);
			      fclose(errlog);
			      snprintf(string, sizeof string, "%s.%d", errfile, nxt++);
			      fprintf(stderr, "Opening next log: %s\n", errfile);
			      errlog = fopen(string, "w");
			      if (errlog == (FILE *) NULL) {
				  fprintf(stderr,
					  "Cannot open next log file (%d), exiting.\n",
					  errno);
				  fprintf(stderr,
					  "Record in error was:\n%s\n",
					  iobuf);
				  exit(99);
			      }
			      recurs = 1;
			      dumprec(ldesc);
			      recurs = 0;
			  } else {
			      fprintf(stderr,
				      "Failed to write to new logfile also!\n");
			      exit(1);
			  }
			  return;
		      }
		      if (fputs(delim, errlog) == EOF) {
			  /*
			   * Recursively call me, once, so whole rec is output. 
			   */
			  if (!recurs) {
			      fprintf(stderr,
				      "Error writing to error log: %d at %d\n",
				      errno,
				      __LINE__);
			      fclose(errlog);
			      snprintf(string, sizeof string, "%s.%d", errfile, nxt++);
			      fprintf(stderr, "Opening next log: %s\n", errfile);
			      errlog = fopen(string, "w");
			      if (errlog == (FILE *) NULL) {
				  fprintf(stderr,
					  "Cannot open next log file (%d), exiting.\n",
					  errno);
				  fprintf(stderr,
					  "Record in error was:\n%s\n",
					  iobuf);
				  exit(99);
			      }
			      recurs = 1;
			      dumprec(ldesc);
			      recurs = 0;
			  } else {
			      fprintf(stderr,
				      "Failed to write to new logfile also!\n");
			      exit(1);
			  }
			  return;
		      }
		      slen += flen + 1;
		      continue;

		  default:
		      fprintf(stderr, "dumprec: Invalid data type: %d \n",
			      col->sqltype);
		      return;
		}
	    }
	    if (fprintf(errlog, "%s%s", string, delim) <= 0) {
		/*
		 * Recursively call me, once, so whole rec is output. 
		 */
		if (!recurs) {
		    fprintf(stderr,
			    "Error writing to error log: %d at %d\n",
			    errno,
			    __LINE__);
		    fclose(errlog);
		    snprintf(string, sizeof string, "%s.%d", errfile, nxt++);
		    fprintf(stderr, "Opening next log: %s\n", errfile);
		    errlog = fopen(string, "w");
		    if (errlog == (FILE *) NULL) {
			fprintf(stderr,
				"Cannot open next log file (%d), exiting.\n",
				errno);
			fprintf(stderr, "Record in error was:\n%s\n", iobuf);
			exit(99);
		    }
		    recurs = 1;
		    dumprec(ldesc);
		    recurs = 0;
		} else {
		    fprintf(stderr,
			    "Failed to write to new logfile also!\n");
		    exit(1);
		}
	    }
	    fflush(errlog);
	    slen += flen + 1;
	}
	slen += fprintf(errlog, "\n");  /* Write out EOL mark. */
    }
    if (MaxErrors && NErr >= MaxErrors) {
	fprintf( stderr, "Maximum error count exceeded!\n" );
	exit(1);
    }
}

/*
 * ----------------------------------------------------------------------- 
 */

void 
usage(char *progname)
{
    char           *p;

    p = strrchr(progname, '/');
    if (!p)
	p = progname;
    fprintf(stdout, "\n");
    fprintf(stdout,
	    "%s - Copy Informix tables from one server/database to another.\n\n",
	    p);

    fprintf(stdout, "Usage: \n");
    fprintf(stdout,
	    "\t%s\t<-f FlushCount>\n"
	    "\t\t<-V> <-i> <-G> <-I> <-F> <-S> <-5> <-L> <-B> <-a> <-A> <-R> <-O>\n"
	    "\t\t<-c> <-C> <-J> <-v>\n"
	    "\t\t<-U username> <-P password>\n"
	    "\t\t<-l errorlog>\n"
	    "\t\t<-E maxerrors>\n"
	    "\t\t<-r delimiter>\n"
	    "\t\t<-h fromhost>\n"
	    "\t\t<-H tohost>\n"
	    "\t\t<-d fromdb>\n"
	    "\t\t<-D todb>\n"
	    "\t\t[-t table_from &| -T table_to]\n"
	    "\t\t<-s select statement>\n"
	    "\t\t<-w lock wait time>\n"
	    "\t\t<-p PDQPRIORITY>\n"
	    "\t\t<-W> <-b widebufsize>\n"
	    "\t\t<-e typestring>\n\n",
	    p );

    fprintf(stdout,
	    "\t-a - Acquire an exclusive mode table lock on the target table.\n" );
    fprintf(stdout,
	    "\t-A - Acquire an exclusive mode table lock on the source table.\n" );
    fprintf(stdout,
	    "\t-c - Disable triggers on the source table during the copy to avoid \n"
	    "\t     SELECT trigger effects. Triggers are reenabled at the end.\n" );
    fprintf(stdout,
	    "\t-C - Disable triggers on the target table during the copy to avoid \n"
	    "\t     INSERT trigger effects Triggers are reenabled at the end..\n" );
    fprintf(stdout,
	    "\t-J - ALTER target table to TYPE(RAW) prior to copying data and revert\n"
	    "\t     it to TYPE(STANDARD) afterwards. If the target table is already \n"
	    "\t     a RAW table the -J option is ignored. (Default: no alter).\n" );
    fprintf(stdout,
	    "\t-O - Override restriction against the source and target db@server:table\n"
	    "\t     being the same table. Note: You can still copy a table into itself \n"
	    "\t     without -O if the source and target hosts are aliases or the source\n"
	    "\t     and target tables are synonyms - caviat emptor!\n" );
    fprintf(stdout,
	    "\t-R - Do not replicate insertions.  Begins transactions with the WITHOUT \n"
            "\t     REPLICATION clause included to prevent ER from replicating the data.\n" );
    fprintf(stdout,
	    "\t-U - Connect to both servers using 'username'.  If -P not passed and \n"
	    "\t     INFORMIXPASS not set, prompt for a password.\n" );
    fprintf(stdout,
	    "\t-P - If a username is provided, either with -U or through INFORMIXUSER, \n"
	    "\t     -P can be used to supply a password.  Otherwise the user is \n"
	    "\t     prompted.\n" );
    fprintf(stdout,
	    "\t-f - Number of records between database commits (Default: 10000).\n"
	    "\t     Commits actually occur on the next completed input buffer "
	    "that \n\t     exceeds FlushCount if the -F option is invoked.\n");
    fprintf(stdout, "\t-G - Print data copy rate in rows per second and total copy time.\n" );
    fprintf(stdout, "\t-v - Verbose output.\n" );
    fprintf(stdout, "\t-V - Print version information and exit.\n");
    fprintf(stdout,
	    "\t-i - Ignore duplicate record errors and do not log dups.\n");
    fprintf(stdout,
	    "\t-I - Enable COMMITTED READ LAST COMMITTED ISOLATION on the source.\n" );
    fprintf(stdout,
	    "\t-F - Flush only when an optimally sized buffer fills.\n"
	    "\t     Note that this option will improve throughput greatly,\n"
	    "\t     however, some rows may not have been flushed if %s is "
	    "killed.\n",
	    progname);
    fprintf(stdout,
	    "\t-W - Deprecated.  Ignored. \n" );
    fprintf(stdout,
	    "\t-b - Fetch Buffer size.  Sets FetBufSize.  ( Default: %d)\n"
	    "\t     Range: 4096 to %llu.  Setting this too large may cause memory \n"
	    "\t     issues and can reduce performance!  Use carefully. Note large \n"
	    "\t     fetch buffers not supported prior to CSDK version 3.70.xC6 and IDS \n"
	    "\t     version 11.70.xC6!\n", 
	    DefaultBufSize,
	    (unsigned long long)MaxBufSize );
    fprintf(stdout, "\t-S - Silent mode.  Minimize screen output.\n");
    fprintf(stdout,
	    "\t-l - Any records incurring an insert error will be written to the log\n"
	    "\t     file in ASCII delimited format suitable for load/dbload or in\n"
	    "\t     binary format suitable for ul.ec if -B is specified.  See below.\n"
	    "\t     (Default name: bcopy.unl)\n");
    fprintf(stdout,
	    "\t-E - Limit the number of errors that will be logged to maxerrors.\n" );
    fprintf(stdout,
	    "\t-B - Create error log file using binary records.  Log may be reloaded \n"
	    "\t     using the ul.ec utility. Required if character type columns contain\n"
	    "\t     binary or other non-printable data copies with a FIXCHAR format.\n"
	    "\t     See -e below. (Default: Delimited log records.)\n");
    fprintf(stdout,
	    "\t-r - Delimiter character to use for error log records.\n");
    fprintf(stdout, "\t-5 - Connect to hosts using 5.0 compatible source.\n");
    fprintf(stdout,
	    "\t-L - Was: Log CHAR columns stripping trailing spaces. Deprecated in\n"
	    "\t     favor of '-e CHAR'.  Currently ignored for backward compatibility.\n");
    fprintf(stdout,
	    "\t-e - Toggle trimming or extension of specific character types to FIXCHAR\n"
	    "\t     Set 'typestring' to:\n"
	    "\t         \"char\" - CHAR and NCHAR type columns (default: extend)\n"
	    "\t         \"varchar\" - VARCHAR and NVCHAR types (default: right trim)\n"
	    "\t         \"lvarchar\" - LVARCHAR type columns (default: right trim)\n"
	    "\t     Extending to FIXCHAR permits correctly copying binary data.\n"
	    "\t     NOTE: Binary data may NOT be recoverable from delimited error \n"
	    "\t     files.  Use the -B option if CHAR or VARCHAR columns contain any \n"
	    "\t     binary/non-printing characters.  Also note that binary data cannot\n"
	    "\t     be reliably retrieved from LVARCHAR columns by any method known\n"
	    "\t     to me.  The -e lvarchar option is ONLY used for non-printing ASCII\n"
	    "\t     characters.  It is NOT supported for use copying binary data!  Also\n"
	    "\t     note -e lvarchar will cause the target column to be space padded to\n"
	    "\t     its the full length of the source table column.\n");
    fprintf(stdout, "\t-h - Host to copy from ($INFORMIXSERVER).\n");
    fprintf(stdout, "\t-H - Host to copy to ($INFORMIXSERVER).\n");
    fprintf(stdout, "\t-d - Database from which to copy (Default: to_db).\n");
    fprintf(stdout, "\t-D - Database to which to copy (Default: from_db).\n");
    fprintf(stdout,
	    "\t-t - Table from which data is to be selected (Default: table_to).\n");
    fprintf(stdout,
	    "\t-T - Table into which data is to be inserted (Default: table_from).\n");
    fprintf(stdout,
	    "\t-s - Select statement (Default: SELECT * FROM <table>).\n");
    fprintf(stdout,
	    "\t     You may include '%%s' to have %s insert <table_from>.\n",
	    p);
    fprintf(stdout,
	    "\t-w - Number of seconds to wait for locks to free up.\n"
	    "\t     If negative wait forever, if zero no wait.  (Default: NO WAIT.)\n");
    fprintf(stdout, "\t-p - PDQPRIORITY to set for the task.\n\n");
    fprintf(stdout, "One of -t or -T option or both must be specified.\n\n");
    fprintf(stdout,
	    "If column names selected differ from those in the target table then the\n"
	    "SELECT statment MUST include column aliases corresponding to the column names\n"
	    "of the target table columns.\n\n");
    fprintf(stdout,
	    "Note: If present, the contents of the environment variables INFORMIXUSER and \n"
	    "      INFORMIXPASS will be used as username and password when connecting to \n"
	    "      both the source and target servers.  The -U & -P options take precedence.\n\n");
}

void dumpSQLDA(char *msg, struct sqlda *desc, int nrecs)
{
    fprintf(stderr, "%s: SQLDA @ %p.\n", msg, desc);

    for (;;) {
	fprintf(stderr, "SQLDA name = %s\n", desc->desc_name);
	fprintf(stderr, "      sqld = %d\n", desc->sqld);
	fprintf(stderr, "      vars = %p\n", desc->sqlvar);
	fprintf(stderr, "      occ  = %d\n", desc->desc_occ);
	fprintf(stderr, "      next = %p\n", desc->desc_next);
	if (nrecs) dumpSqlvar(desc, nrecs);
	if (desc->desc_next)
	    desc = desc->desc_next;
	else
	    break;
    };

}

void dumpSqlvar(struct sqlda *desc, int nrecs)
{
    struct sqlvar_struct *currentVar;
    int             i, j;
    char            tempbuf[5000];
    char *ptr;

    bzero(tempbuf, 5000);

    fprintf(stderr, "SQLDA sqlvar dump\n");
    for (i = 0; i < desc->sqld; i++) {
      currentVar = &(desc->sqlvar[i]);
      fprintf(stderr, "     Variable #%d\n", i);
      fprintf(stderr, "     Type %d\n", currentVar->sqltype);
      if (currentVar->sqltype == CDTIMETYPE || currentVar->sqltype == CINVTYPE)
	fprintf(stderr, "     Length %d, prec = 0x%x\n", (int)sizeof(dtime_t), currentVar->sqllen);
      else
	fprintf(stderr, "     Length %d\n", currentVar->sqllen);
      fprintf(stderr, "     Data Address %p\n", currentVar->sqldata);
      ptr = currentVar->sqldata;

      for (j=0; j < nrecs; j++) {
	switch (currentVar->sqltype) {
	  case SQLCHAR:
	  case SQLVCHAR:
	  case SQLLVARCHAR:
	  case CCHARTYPE:
	  case CVCHARTYPE:
	  case CLVCHARTYPE:
	  case CFIXCHARTYPE:
	      memcpy(tempbuf, ptr, currentVar->sqllen);
	      fprintf(stderr, "     Value[%d]: <%s>, at %p\n", j, tempbuf, ptr);
	      break;
	  case SQLINT:
	  case SQLSERIAL:
	  case CINTTYPE:
	  case CDATETYPE:
	      fprintf(stderr, "     Value[%d]: %d, at %p\n", j, *(int *) ptr, ptr);
	      break;
	  case SQLINT8:
	  case SQLSERIAL8:
	  case CINT8TYPE:
	      ifx_int8toasc( (ifx_int8_t *)ptr, tempbuf, 20 );
	      fprintf(stderr, "     Value[%d]: %s, at %p\n", j, tempbuf, ptr);
	      break;
#if defined( SQLINFXBIGINT )
	  case SQLINFXBIGINT:
	  case SQLBIGSERIAL:
	  case CBIGINTTYPE:
#endif
	      fprintf(stderr, "     Value[%d]: %lld, at %p\n", j, *(long long *) ptr, ptr);
	      break;
	  case SQLFLOAT:
	  case CDOUBLETYPE:
	      fprintf(stderr, "     Value[%d]: %f, at %p\n", j, *(double *) ptr, ptr);
	      break;
	  case CDTIMETYPE:
	      dttoasc((dtime_t *)ptr, tempbuf);
	      fprintf(stderr, "     Value[%d]: %s, at %p.\n", j, tempbuf, ptr );
	      break;
	  default:
	      fprintf(stderr, "     Type[%d]: %d not handled\n", j, currentVar->sqltype);
	      break;
	}
	if (currentVar->sqltype == CDTIMETYPE)
	  ptr += sizeof(dtime_t);
	else if (currentVar->sqltype == CINVTYPE)
	  ptr += sizeof(intrvl_t);
	else
	  ptr += currentVar->sqllen;
      }
    }
}

