/* dbdelete.ec - Program to delete records from a database table without using
                too many locks.  Uses a maximum of 32767 locks when deleting 
		using a 1byte key or 8191 sets of locks when using ROWID.
                Don't forget to multiply by the number of indexes on the table.
*/
/*
   RCS Header:
   -----------
   $Header: /home/art/GoogleDrive/Latest/RCS/dbdelete.ec,v 1.38 2023/10/26 19:52:02 art Exp $

   RCS Log:
   --------
   $Log: dbdelete.ec,v $
   Revision 1.38  2023/10/26 19:52:02  art
   Thanks to David Williams for adding some trace and expanding the error
   reporting.

   Revision 1.37  2023/03/29 16:40:36  art
   Updated to better support compilation on Windows.

   Revision 1.36  2021/12/21 17:12:40  art
   Updated Copyright notice.

   Revision 1.35  2020/02/12 14:18:42  art
   Fixed a bug when -I was passed. Added new -v (verbose) flag to increase
   runtime trace. Added new -P option to specify a pause between delete
   runs. May improve performance on some servers or reduce impact on other
   processes. Fixed a problem with -p which was not obeying "-p 0".
   Thanks to David Williams for the ideas for -P & -v and for identifying
   the -p 0 issue. David also provided the code to implement -P.

   Revision 1.34  2019/11/08 02:22:20  art
   Added usage for -I

   Revision 1.33  2019/11/05 23:39:37  art
   Implemented -I to support using ifx_row_id for partitioned tables!

   Revision 1.32  2019/03/15 15:47:04  art
   Fixed some compiler warnings.

   Revision 1.31  2018/07/27 11:21:59  art
   Fixed a typo in ifdef protecting BIGINT handler code.
   Thanks to Robert Johansson for pointing this out.

   Revision 1.30  2016/02/02 12:31:03  art
   Fixed typo introduced in v1.29.

   Revision 1.29  2016/02/02 12:02:00  art
   Fixed problem with delete by key and identifiers longer than 18 bytes.
   Updated the copyright dates.

   Revision 1.28  2013/08/12 00:56:47  art
   Added support for INT8/SERIAL8, BIGINT/BIGSERIAL, and LVARCHAR types in -q queries.

   Revision 1.27  2013/02/10 22:18:19  art
   Updated Copyright

   Revision 1.26  2011/12/22 15:16:19  art
   Removed progress printout if -S is specified.  Also redirected version printout to stdout (was stderr).

   Revision 1.25  2011/12/21 20:47:07  art
   Added new option -R which begins all transactions with the WITHOUT REPLICATION
   clause which prevents the operation from being replicated via Enterprise
   Replication.  Thanks to Øyvind Gjerstad for the suggestion.
   Also increased the size of the maximum SQL DELETE command (for rowid deletes)
   from 32750 to 65500 to reflect the increase in the SQL buffer size starting
   with IDS v9.40.  Note it is still not a good idea to make the delete
   statements that large, but Caviat Emptor applies.

   Revision 1.24  2011/05/23 16:04:57  art
   Updated the copyright statement.

*/

static char RCSHeader[] = "$Header: /home/art/GoogleDrive/Latest/RCS/dbdelete.ec,v 1.38 2023/10/26 19:52:02 art Exp $";
static char RCSWhat[] = "<@(#) $RCSfile: dbdelete.ec,v $	 $Revision: 1.38 $ $Date: 2023/10/26 19:52:02 $ >";

#include <stdio.h>
#include <stdlib.h>
#if !defined( _WINDOWS_ )
#include <unistd.h>
#endif
#include <string.h>
#include <errno.h>
#include <time.h>
#include <getopt.h>

EXEC SQL include sqlca;
EXEC SQL include sqlda;
EXEC SQL include sqltypes;
EXEC SQL include decimal;
EXEC SQL include locator;

#define FLDSIZE  40
#define BUFFSIZE 256
#define DELIM "|"
#define FETBUFSIZE 32767

/* PROTOTYPES */
int delfrom_table( void );
void usage( char *progname );
void dumprec( struct sqlda * );

#if defined( _AIX ) & !defined( __GNUC__ )
#	define inline
#endif

/* GLOBALS - I hate these things. */
FILE *errlog;
char errfile[1025] = "dbdelete.unl";
char delim[2] = DELIM;
int begin_ok = 0, commit_ok = 0; /*  True if DB has transactions.  */
int use_ins_curs = 0;
int use_ifx_row_id = 0;
int LoseDups = 0;
int DelByKey = 0;
char DelKeys[100][129];
int Silent = 0;
int SBsz = 2048;
int lclFetBufSize = FETBUFSIZE;
int is_open = 0;
int one_open = 0;
int no_replicate = 0;
int Debug = 0;
int verbose = 0;
int LoopPause=0;

EXEC SQL BEGIN DECLARE SECTION;
int priority = 0, waittime = 10;
char dbname_from[130];
char host_from[102];
char tabname[130];
char qrystmt[3001];
char selstmt[2001] = "";
char delete_stmt[32770];
char temp[2001];
EXEC SQL END DECLARE SECTION;

#if ! (defined( __DGUX__ ) & defined( __ix86 ))
extern short FetBufSize;
extern short FetArrSize;
#endif

#define Version "3.2"
#define Revision "$Revision: 1.38 $"

#define ERR_OUT(txt) { fprintf( stderr, txt ": %d, %d, %s\n", sqlca.sqlcode, sqlca.sqlerrd[1], sqlca.sqlerrm ); }

int main( int argc, char **argv)
{
    int n, ch, errval;
    char PDQ[50], *p;
    static char revision[1024], *envFetBufSize;
    int fullstmt = 0;

    if ( argc == 1 ) {
	usage( argv[0] );
	exit( 2 );
    }

    if ((envFetBufSize = getenv( "FET_BUF_SIZE" )) != (char *)NULL) {
	lclFetBufSize = atoi( envFetBufSize );
	if (lclFetBufSize < 2048)
	    lclFetBufSize = 2048;
	if (lclFetBufSize > 32767)
	    lclFetBufSize = 32767;
    }

    while ((ch = getopt( argc, argv, "q:b:d:h:op:s:t:u:vw:B:ISRVP:" )) != EOF) 
	switch (ch) {
	case 'I':
	    use_ifx_row_id = 1;
	    break;
	case 'R':
	    no_replicate = 1;
	    break;
	case 'o':
	    one_open = 1;
	    break;
	case 'q':
	    snprintf( qrystmt, sizeof qrystmt, "%s", optarg );
	    fullstmt = 1;
	    break;
	case 'b':
	    lclFetBufSize = atoi( optarg );
	    break;
	case 'V':
	    strcpy( revision, Revision );
	    fprintf( stdout,
		     "\n%s: Features Version %s, Source %s!\n",
		     argv[0],
		     Version,
		     strtok( revision, "$" ) );
	    fprintf( stdout, "\tCopyright 1996-2022 by Art S. Kagel.\n" );
            fprintf( stdout,
                     "\tLicense for private usage (including use by "
                     "commercial enterprise\n"
                     "\tfor its own needs) is granted.  Commercial and "
                     "exploitation rights\n"
                     "\tare reserved.\n\n" );
	    fflush( stdout );
	    exit(0);
	    break;
	case 'v':
            verbose=1;
            fprintf( stderr, "Verbose ON\n");
            fflush( stderr );
	    break;
	case 'B':
	    SBsz = atoi( optarg );
	    break;
	case 'S':
	    Silent = 1;
	    break;
	case 'w':
	    waittime = atoi( optarg );
	    if (!waittime)
		waittime = 10;
	    break;
	case 'u':
	    p = strtok( optarg, "," );
	    if (p)
		do {
		    strncpy( DelKeys[DelByKey], p, 129 );
		    DelKeys[DelByKey][128] = (char)0;
		    DelByKey++;
		} while ((p = strtok( NULL, "," )) != (char *)NULL);
	    break;
	case 'h':
	    /* Delete from host. */
	    snprintf( host_from, sizeof host_from, "@%.99s", optarg );
	    host_from[sizeof host_from -1] = (char)0;
	    break;
	case 'd':
	    /* Delete from database. */
	    strncpy( dbname_from, optarg, 129 );
	    dbname_from[128] = (char)0;
	    break;
	case 's':
	    strncpy( selstmt, optarg, 2000 );
	    selstmt[2000] = (char)0;
	    break;
	case 't':
	    /* Table to delete from. */
	    strncpy( tabname, optarg, 129 );
	    tabname[128] = (char)0;
	    break;
	case 'p':
	    priority = atoi( optarg );
	    break;
	case 'P':
	    LoopPause = atoi( optarg );
	    break;
	default:
	    usage( argv[0] );
	    exit( 2 );
	    break;
	}

    if (getenv( "DBDELETE_DEBUG" ))
        Debug = 1;

    if (verbose) {
        fprintf( stderr, "LoopPause=%d\n",LoopPause);
        fflush( stderr );
    }

    if (lclFetBufSize < 2048 || lclFetBufSize > 32767) {
	fprintf( stderr,
		 "Invalid communication buffer size (%d)!\n",
		 lclFetBufSize );
	usage( argv[0] );
	exit( 2 );
    }

    if (!SBsz)
	SBsz = 2032;
    else if (SBsz > 65500) {
	fprintf( stderr, 
		"SQL Buffer size (%d) exceeds maximum (65500).  "
		"Adjusting to maximum.\n", 
		SBsz );
	SBsz = 65500;
    } else
	SBsz -= 16;

    /* Assume that the user has given us a where clause. */
    if (!fullstmt) {
	if (!DelByKey)
	    if (strlen( selstmt ))
		sprintf( qrystmt,
			 "SELECT %s FROM %s WHERE %s",
			 use_ifx_row_id == 1 ? "ifx_row_id" : "ROWID",
			 tabname,
			 selstmt );
	      else
		sprintf( qrystmt,
			 "SELECT %s FROM %s",
			 use_ifx_row_id == 1 ? "ifx_row_id" : "ROWID",
			 tabname );
	else {
	    strcpy( qrystmt, "SELECT " );
	    for (n=0; n < DelByKey; n++ ) {
		strcat( qrystmt, DelKeys[n] );
		if ((n+1) < DelByKey)
		    strcat( qrystmt, ", " );
	    }
	    strcat( qrystmt, " FROM " );
	    strcat( qrystmt, tabname );
	    if (strlen( selstmt )) {
		strcat( qrystmt, " WHERE " );
		strcat( qrystmt, selstmt );
	    }
	} 
    }
	
    if (!Silent)
	printf( "Selecting data with:\n\t%s;\n\n", qrystmt );
	    
    /* Set up PDQPRIORITY. */
    sprintf( PDQ, "PDQPRIORITY=%d", priority );
    putenv( PDQ );

    /* Set up connection for target database connection. */
    if (strlen( host_from ) == 0) {
	snprintf( host_from,
		  sizeof host_from,
		  "@%s",
		  getenv( "INFORMIXSERVER" ) );
    }

    EXEC SQL CONNECT TO :host_from AS "target" WITH CONCURRENT TRANSACTION;
    if (sqlca.sqlcode < 0) {
        ERR_OUT("Error opening target host connection");
	return 1;
    }

    EXEC SQL  database :dbname_from;
    if (sqlca.sqlcode < 0) {
	ERR_OUT("Error opening target database");
	return 1;
    }
    commit_ok = (sqlca.sqlwarn.sqlwarn1 == 'W');
    begin_ok  = (commit_ok && sqlca.sqlwarn.sqlwarn2 != 'W');

    if (no_replicate && !begin_ok) {
	fprintf( stderr,
		 "Transactions without replication (-R) specified but database %s %s ignoring.\n",
		 dbname_from,
		 commit_ok ? "is ANSI mode" : "does not \nsupport transactions" );
	no_replicate = 0;
    }

    EXEC SQL SET ISOLATION TO DIRTY READ;
    if (sqlca.sqlcode < 0 && sqlca.sqlcode != -256) {
        ERR_OUT("Error setting Isolation Level for target");
	return 1;
    }    

    if (waittime > 0)
	sprintf( temp, "SET LOCK MODE TO WAIT %d;", waittime );
    else
	strcpy( temp, "SET LOCK MODE TO WAIT;" );

    EXEC SQL EXECUTE IMMEDIATE :temp;
    if (sqlca.sqlcode < 0) {
        ERR_OUT("Error setting lock mode for target");
	return 1;
    }    

    EXEC SQL IFNDEF ifmx_6_0;
    if (priority > 50)
	EXEC SQL SET PDQPRIORITY 60;
    else if (priority > 40)
	EXEC SQL SET PDQPRIORITY 50;
    else if (priority > 30)
	EXEC SQL SET PDQPRIORITY 40;
    else if (priority > 20)
	EXEC SQL SET PDQPRIORITY 30;
    else if (priority > 10)
	EXEC SQL SET PDQPRIORITY 20;
    else if (priority > 0)
	EXEC SQL SET PDQPRIORITY 10;
    else 
	EXEC SQL SET PDQPRIORITY 0;
    if (sqlca.sqlcode < 0) {
        ERR_OUT("Warning setting PDQPRIORITY on target");
    }
    EXEC SQL ENDIF;

#ifdef TWOCONNECT
    /* Set up connection for source database connection. */
    if (strlen( host_from ) == 0) {
	strncpy( host_from, sizeof host_from, getenv( "INFORMIXSERVER" ) );
	host_from[sizeof host_from - 1] = (char)0;
    }

    EXEC SQL CONNECT TO :host_from AS "source" WITH CONCURRENT TRANSACTION;
    if (sqlca.sqlcode < 0) {
        ERR_OUT("Error opening source host connection");
	return 1;
    }

    EXEC SQL  database :dbname_from;
    if (sqlca.sqlcode < 0) {
        ERR_OUT("Error opening source database");
	return 1;
    }

    if (waittime > 0)
	sprintf( temp, "SET LOCK MODE TO WAIT %d;", waittime );
    else
	strcpy( temp, "SET LOCK MODE TO WAIT;" );

    EXEC SQL EXECUTE IMMEDIATE :temp;
    if (sqlca.sqlcode < 0) {
        ERR_OUT("Error setting lock mode for source");
	return 1;
    }    

    EXEC SQL IFNDEF ifmx_6_0;
    if (priority > 50)
	EXEC SQL SET PDQPRIORITY 60;
    else if (priority > 40)
	EXEC SQL SET PDQPRIORITY 50;
    else if (priority > 30)
	EXEC SQL SET PDQPRIORITY 40;
    else if (priority > 20)
	EXEC SQL SET PDQPRIORITY 30;
    else if (priority > 10)
	EXEC SQL SET PDQPRIORITY 20;
    else if (priority > 0)
	EXEC SQL SET PDQPRIORITY 10;
    else 
	EXEC SQL SET PDQPRIORITY 0;
    if (sqlca.sqlcode < 0) {
        ERR_OUT("Warning setting PDQPRIORITY on source");
    }
    EXEC SQL ENDIF;

    EXEC SQL SET CONNECTION "source";
    if ( sqlca.sqlcode != 0 ) {
        ERR_OUT("Error on SET CONNECTION TO source in main");
	return 1;
    }
#endif

    errval = delfrom_table();
    if (errlog != (FILE *)NULL) {
        fflush( errlog );
	fclose( errlog );
    }
    return errval;
}

/* ---------------------------------------------------------------------- */

int sleep_ms(unsigned int ms)
{
  int result = 0;

  struct timespec ts_rem =
  { 
    ms / 1000, 
    (ms % 1000) * 1000000L 
  };

  do
  {
    struct timespec ts_sleep = ts_rem;
#if defined( _WINDOWS_ )
    result = 0;
    Sleep( ms )
#else
    result = nanosleep(&ts_sleep, &ts_rem);
#endif
  } while ((result==-1) && (errno==EINTR));

  if (result==-1) {
    fprintf( stderr,
	     "Sleep failed :%d\n",
	     errno);
  }

  return result;
}
/* ---------------------------------------------------------------------- */

int delfrom_table( void )
{
    register int pos;
    register char *cp, *cp2;
    register int len;
    register int i, j;
    char tmp[14];
    struct sqlda *udesc, *cdesc;
    struct sqlvar_struct *col, *col2;
    char *buffer = NULL, *buffer2 = NULL;
    char *ip, *ip2 = delete_stmt;
    int  counter = 0, ii, errcount = 0, errlvl = 0, reccount = 0, dupslost = 0;
    int uncommitted = 0, tries = 0, fetchcount = 0, dsiz, tl;
    char string[1025];
    char **bufary;
    short **ipary;
    short *colszs;

    EXEC SQL PREPARE fetcher FROM :qrystmt;
    if ( sqlca.sqlcode != 0 ) {
        ERR_OUT("Error on prepare of fetch");
	if (sqlca.sqlcode == -201)
	    fprintf( stderr,
		     "Syntax error at char %d in:\n%s\n",
		     sqlca.sqlerrd[4],
		     qrystmt );
	else if (sqlca.sqlcode == -857)
	    fprintf( stderr,
		     "\n\tTable is fragmented without ROWID.  "
		     "Rerun specifying key column(s).\n" );
	return 1;
    }
 
    EXEC SQL DESCRIBE fetcher INTO udesc;
    if ( sqlca.sqlcode != 0 ) {
        ERR_OUT("Error on describe");
        return 1;
    }
    /* Create an sqlda struture for the insert statement that does not use the
       insert array code. */
    cdesc = (struct sqlda *)malloc( sizeof (struct sqlda) );
    memset( cdesc, 0, sizeof (struct sqlda) );
    cdesc->sqld = udesc->sqld;
    cdesc->sqlvar = (struct sqlvar_struct *)
        malloc( cdesc->sqld * sizeof (struct sqlvar_struct) );
    memset( cdesc->sqlvar, 0, cdesc->sqld * sizeof (struct sqlvar_struct) );
    strncpy( cdesc->desc_name, udesc->desc_name, 19 );
    cdesc->desc_occ = udesc->desc_occ;
    cdesc->desc_next = (struct sqlda *)NULL;
 
    /* Step 1: analyze udesc to determine type and memory requirements
     *         of each item in the select list.  rtypalign() returns a
     *         pointer to the next appropriate boundary (starting at
     *         pos) for the indicated data type.
     */
    pos = 0;
 
    /* Allocate pointer array to hold the fetch array pointers. */
 
    for (col = udesc->sqlvar, i = 0, col2 = cdesc->sqlvar;
         i < udesc->sqld;
         col++, i++, col2++)
    {
        switch(col->sqltype & SQLTYPE)
        {
        case SQLSMFLOAT:
            col->sqltype = CFLOATTYPE;
            col2->sqltype = CFLOATTYPE;
            break;
 
        case SQLFLOAT:
            col->sqltype = CDOUBLETYPE;
            col2->sqltype = CDOUBLETYPE;
            break;
 
        case SQLMONEY:
        case SQLDECIMAL:
            col->sqltype = CDECIMALTYPE;
            col2->sqltype = CDECIMALTYPE;
            break;
 
        case SQLSERIAL:
        case SQLINT:
            col->sqltype = CINTTYPE;
            col2->sqltype = CINTTYPE;
            break;

#if defined( SQLINT8 )
	case SQLINT8:
	case SQLSERIAL8:
#if defined( SQLINFXBIGINT )
	case SQLINFXBIGINT:
	case SQLBIGSERIAL:
	    col->sqltype = CBIGINTTYPE;
	    col2->sqltype = CBIGINTTYPE;
#else
	    col->sqltype = CDECIMALTYPE;
	    col2->sqltype = CDECIMALTYPE;
#endif
	    break;
#endif
 
        case SQLSMINT:
            col->sqltype = CSHORTTYPE;
            col2->sqltype = CSHORTTYPE;
            break;
 
        case SQLVCHAR:
	    col->sqltype = CSTRINGTYPE;
	    col2->sqltype = CSTRINGTYPE;
	    break;
        case SQLCHAR:
            col->sqltype = CFIXCHARTYPE; /* get all bytes */
            col2->sqltype = CFIXCHARTYPE; /* get all bytes */
            break;
 
        case SQLDATE:
            col->sqltype = CDATETYPE;
            col2->sqltype = CDATETYPE;
            break;
 
        case SQLDTIME:
            col->sqltype = CFIXCHARTYPE;
            col2->sqltype = CFIXCHARTYPE;
            col2->sqllen = col->sqllen = rtypmsize( col->sqltype, 26 );
            break;
 
        case SQLINTERVAL:
            col->sqltype = CFIXCHARTYPE;
            col2->sqltype = CFIXCHARTYPE;
            col2->sqllen = col->sqllen = rtypmsize( col->sqltype, 26 );
            break;
 
	case SQLLVARCHAR:
	    col->sqltype = CLVCHARTYPE;
	    col2->sqltype = CLVCHARTYPE;
	    break;

        default:
            /* Other data types.  Do nothing if we see an unsupported type. */
            fprintf( stderr, "delfrom_table: Invalid data type: %d \n",
                     col->sqltype );
            return 1;
        }
        switch(col->sqltype & SQLTYPE)
        {
        case CDTIMETYPE:
        case CINVTYPE:
          col2->sqllen = col->sqllen = rtypmsize( col->sqltype, 24 );
          break;
        case CDECIMALTYPE:
          col2->sqllen = col->sqllen = rtypmsize( col->sqltype, 22 );
          break;
        default:
          col2->sqllen = col->sqllen = rtypmsize(col->sqltype,col->sqllen);
        }
        pos = (int) rtypalign(pos,col->sqltype) + col->sqllen;
        col2->sqlname = (char *)malloc( strlen( col->sqlname ) + 1 );
        strcpy( col2->sqlname, col->sqlname );
    }
 
    /* Step 2: Allocate memory to receive rows of the table returned
     *         by the SELECT statement.  The variable pos has an integer
     *         value equal to the number of bytes occupied by the row.
     */
    /* Allocate a buffer for the insert sqlda structure. */
    buffer = malloc( pos );
    memset( buffer, 0, pos );
    /* OK buffer is now the insert buffer.  Bufary is now the array of
       pointers to the fetch buffers for each column. */
    /* Set FetBufSize and GetArrSize and allocate bufary. */
    if (pos > lclFetBufSize) {
        /* For large records, can only fgetch one at a time. */
        FetBufSize = lclFetBufSize;
        FetArrSize = 1;
    } else {
        FetArrSize = (lclFetBufSize / pos);
        FetBufSize = FetArrSize * pos;
    }
    if (!Silent)
	printf( "Deleting blocks of %d records.\n", FetArrSize );
 
    /* Bufary and coszs need one pointer per column in the result set. */
    bufary = (char **)malloc( udesc->sqld * sizeof (char *) );
    colszs = (short *)malloc( udesc->sqld * sizeof (short) );
 
    /* Allocate indicator list. */
    /* Array of indicators for fetch sqldat structure. */
    ipary = (short **)malloc( udesc->sqld * sizeof (short *) );
    if (buffer == NULL || ipary == NULL || colszs == NULL) {
        fprintf(stderr,
		"Out of memory(%lx|%lx|%lx).\n",
		(long)buffer,
		(long)ipary,
		(long)colszs );
        return 1;
    }
 
    /* Step 3: Set pointers in the allocated memory to receive each
     *         item in the select list.
     */

    /* Initialize pointer to insert buffer. */
    cp = buffer;
    for (col = udesc->sqlvar, col2 = cdesc->sqlvar, i = 0;
         i < udesc->sqld;
         col2++, col++, i++)
    {
        colszs[i] = col->sqllen;
        /* Allocate Fetch array for this column. */
        bufary[i] = malloc( col->sqllen * FetArrSize );
        /* Allocate indicator array for this column. */
        ipary[i] = (short *)malloc( FetArrSize * sizeof (short) );
 
        /* Adjust pointer for each column for alignment in insert sqlda */
        cp = buffer + rtypalign((int)(cp - buffer),col->sqltype);
        /* Assign indicator addresses. */
        col->sqlind = ipary[i];
        /* Assign the data pointers. */
        col->sqldata = bufary[i];
        col2->sqldata = cp;
        if (Debug)
          fprintf( stderr,
                   "Setting pointer to %p for %d bytes.\n",
                   cp,
                   col->sqllen );
        cp += col->sqllen;
    }
 
    /* Step 4: Declare fetch cursor. */
    EXEC SQL DECLARE fetch_curs CURSOR WITH HOLD FOR fetcher;
    if ( sqlca.sqlcode != 0 ) {
        fprintf( stderr,
                 "Error: %d(%d) on declare of fetch_curs.\n",
                 sqlca.sqlcode,
                 sqlca.sqlerrd[1] );
        return 1;
    }

#ifdef TWOCONNECT
    /* OK, the source connection and statement are setup, now set up the target
       connection and statement. */

    EXEC SQL SET CONNECTION "target";
    if ( sqlca.sqlcode != 0 ) {
        ERR_OUT("Error on SET CONNECTION TO target before loop");
	return 1;
    }
#endif

    if (!DelByKey) {
      sprintf( delete_stmt,
	       "DELETE FROM %s WHERE %s IN ( ",
	       tabname,
	       use_ifx_row_id == 1 ? "ifx_row_id" : "ROWID" );
	ip2 = delete_stmt + strlen( delete_stmt );
    } else {
	strcpy( delete_stmt, "DELETE FROM " );
	strcat( delete_stmt, tabname );
	strcat( delete_stmt, " WHERE " );
	for (i = 0; i < DelByKey; i++) {
	    strcat( delete_stmt, DelKeys[i] );
	    strcat( delete_stmt, " = ? " );
	    if ((i+1) < DelByKey)
		strcat( delete_stmt, " AND " );
	}
    }

    if (DelByKey) {
	if (!Silent) 
	    printf( "Deleting data with:\n\t%s;\n\n", delete_stmt );
	fflush( stdout );
	
	EXEC SQL PREPARE deletestmt FROM :delete_stmt;
	if ( sqlca.sqlcode != 0 ) {
	    ERR_OUT("Error on PREPARE deletestmt TO target");
	    return 1;
	}
    } else if (!Silent) {
	printf( "Deleting data with:\n\t%s... );\n\n", delete_stmt );
    }

    if (begin_ok) {
	if (no_replicate)
	    EXEC SQL BEGIN WORK WITHOUT REPLICATION;
	else
	    EXEC SQL BEGIN WORK;
	if ( sqlca.sqlcode != 0 ) {
	    ERR_OUT("Error on BEGIN WORK before loop");
	    return 1;
	}
    }

    /* DELETE LOOP */
    sqlca.sqlcode = 0;
    for ( ;; ) {

#ifdef TWOCONNECT
	EXEC SQL SET CONNECTION "source";
	if ( sqlca.sqlcode != 0 ) {
	    ERR_OUT("Error on SET CONNECTION TO source in loop");
	    errlvl = 1;
	    break;
	}
#endif

	/* New algorithm.  Fetch one array's worth of keys/rowids, close cursor,
	   delete these rows, reopen. */

	/* Open the fetch cursor. */
	if (!is_open) {
	    if (Debug)
	        fprintf( stderr, "Opening cursor at li %d.\n", __LINE__ );
	    EXEC SQL OPEN fetch_curs;
	    is_open = 1;
	    if ( sqlca.sqlcode != 0 ) {
	      ERR_OUT("Error on OPEN fetch_curs");
	      return 1;
	    }
 	}

	if (Debug)
	  fprintf( stderr, "Fetching at line: %d.\n", __LINE__ );
	
	EXEC SQL FETCH fetch_curs USING DESCRIPTOR udesc;
	if (Debug)
	  fprintf( stderr, "Fetched %d IDs.\n", sqlca.sqlerrd[2] );

	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,"reccount %d ",reccount);
	    ERR_OUT("Error fetching row");
	    if (sqlca.sqlcode == -243 && tries < 5)
	    {
		tries++;
		if (!one_open && is_open) {
		    if (Debug)
		        fprintf( stderr, "Closing cursor at li %d.\n", __LINE__ );
		    EXEC SQL CLOSE fetch_curs;
		    is_open = 0;
		    if ( sqlca.sqlcode != 0 ) {
	              ERR_OUT("Error on CLOSE fetch_curs 1st in loop");
		      return 1;
		    }
		}

		if (!is_open) {
		    if (Debug)
		        fprintf( stderr, "Opening cursor at li %d.\n", __LINE__ );
		    EXEC SQL OPEN fetch_curs;
		    is_open = 1;
		    if ( sqlca.sqlcode != 0 ) {
	              ERR_OUT("Error on REOPEN fetch_curs 1st in loop");
		      return 1;
		    }
		}
#ifdef TWOCONNECT
		EXEC SQL SET CONNECTION "target";
		if (sqlca.sqlcode != 0) {
	            ERR_OUT("Error on SET CONNECTION TO target 1st in loop");
		    errlvl = 1;
		    break;
		}
#endif
		if (commit_ok) {
		    EXEC SQL COMMIT WORK;
		    if ( sqlca.sqlcode != 0 ) {
	                ERR_OUT("Error on COMMIT WORK 1st in loop");
			errlvl = 1;
			break;
		    }
		}
		uncommitted = 0;
		if (begin_ok) {
		    if (no_replicate)
			EXEC SQL BEGIN WORK WITHOUT REPLICATION;
		    else
			EXEC SQL BEGIN WORK;
		    if ( sqlca.sqlcode != 0 ) {
	                ERR_OUT("Error on BEGIN WORK 1st in loop");
			return 1;
		    }
		}
		continue;	/* Try again. */
	    } else if (sqlca.sqlcode < 0)
	      exit(1);
	    break;
	} else if (sqlca.sqlcode == 100)
	    break;
	tries = 0;
	/* Save number of keys fetched this round. */
	fetchcount = sqlca.sqlerrd[2];
	if (Debug)
	  fprintf( stderr, "Fetchcount = %d\n", fetchcount );

	if (!one_open && is_open) {
	    if (Debug)
	        fprintf( stderr, "Closing cursor at li %d.\n", __LINE__ );
		EXEC SQL CLOSE fetch_curs;
		if ( sqlca.sqlcode != 0 ) {
	        	ERR_OUT("Error on CLOSE fetch_curs 2nd in loop");
			return 1;
		}
	    is_open = 0;
	}
	if (!is_open) {
	  if (Debug)
	    fprintf( stderr, "Opening cursor at li %d.\n", __LINE__ );
	  EXEC SQL OPEN fetch_curs;
	  is_open = 1;
	  if ( sqlca.sqlcode != 0 ) {
	    ERR_OUT("Error REOPEN fetch_curs");
	    return 1;
	  }
	}

#ifdef TWOCONNECT
	EXEC SQL SET CONNECTION "target";
	if (sqlca.sqlcode != 0) {
	    ERR_OUT("Error SET CONNECTION TO target");
	    errlvl = 1;
	    break;
	}
#endif

	col = udesc->sqlvar;
	col2 = cdesc->sqlvar;
	if (DelByKey) {
	    for (i=0; i<fetchcount; i++) {
		for (j = 0; j < udesc->sqld; j++) {
		    col2[j].sqldata = (col[j].sqldata + (i * col[j].sqllen));
		    col2[j].sqlind = &(col[j].sqlind[i]);
		}
	    
		if (Debug)
		  fprintf( stderr, "Executing: %s.\n", delete_stmt );

		EXEC SQL EXECUTE deletestmt USING DESCRIPTOR cdesc;
		if ( sqlca.sqlcode != 0 ) {
	            ERR_OUT("Error on EXECUTE delete 1st in loop");
		    return 1;
		}
		counter += sqlca.sqlerrd[2];
	    }
	} else {
	    /* Delete by rowid. */
	    i=0;
	    while (i < fetchcount) {
		dsiz = ip2 - delete_stmt;
		ip = ip2;
		for (; i < fetchcount && dsiz < SBsz; i++) {
		  if (use_ifx_row_id == 1) {
		    sprintf( tmp,
			     "'%s'",
			     (char *)(col[0].sqldata+(i*col[0].sqllen)) );
		  } else {
		    sprintf( tmp,
			     "%d",
			     *(int *)(col[0].sqldata+(i*col[0].sqllen)) );
		  }
		  strcpy( ip, tmp );
		  ip += (tl = strlen( tmp ));
		  dsiz += (tl + 1);
		  if (i < (fetchcount - 1) && dsiz < SBsz)
		    strcpy( ip++, "," );
		}
		while (dsiz < SBsz && fetchcount == 1) {
		  if (Debug)
		    fprintf( stderr, "Fetching at line: %d.\n", __LINE__ );
	
		    EXEC SQL FETCH fetch_curs USING DESCRIPTOR udesc;
		    if (Debug)
		      fprintf( stderr, "Fetched %d IDs.\n", sqlca.sqlerrd[2] );

		    if (sqlca.sqlcode < 0) {
	                fprintf( stderr,"reccount %d ",reccount);
	                ERR_OUT("Error on EXECUTE delete 2nd in loop");
			if (sqlca.sqlcode == -243 && tries < 5)
			{
			    tries++;
			    if (!one_open && is_open) {
			        if (Debug)
				    fprintf( stderr, "Closing cursor at li %d.\n", __LINE__ );
				EXEC SQL CLOSE fetch_curs;
				is_open = 0;
				if ( sqlca.sqlcode != 0 ) {
	                          ERR_OUT("Error on CLOSE fetch_curs 3rd in loop");
				  return 1;
				}
			    }

			    if (!is_open) {
			        if (Debug)
				    fprintf( stderr, "Opening cursor at li %d.\n", __LINE__ );
			        EXEC SQL OPEN fetch_curs;
				is_open = 1;
				if ( sqlca.sqlcode != 0 ) {
	                          ERR_OUT("Error on REOPEN fetch_curs 2nd in loop");
				  return 1;
				}
			    }

#ifdef TWOCONNECT
			    EXEC SQL SET CONNECTION "target";
			    if (sqlca.sqlcode != 0) {
	                        ERR_OUT("Error on SET CONNECTION TO target 2nd in loop");
				errlvl = 1;
				break;
			    }
#endif
			    if (commit_ok) {
				EXEC SQL COMMIT WORK;
				if ( sqlca.sqlcode != 0 ) {
	                            ERR_OUT("Error on COMMIT WORK 2nd in loop");
				    errlvl = 1;
				    break;
				}
			    }
			    uncommitted = 0;
			    if (begin_ok) {
				if (no_replicate)
				    EXEC SQL BEGIN WORK WITHOUT REPLICATION;
				else
				    EXEC SQL BEGIN WORK;
				if ( sqlca.sqlcode != 0 ) {
	                            ERR_OUT("Error on BEGIN WORK 2nd in loop");
				    return 1;
				}
			    }
			    continue;	/* Try again. */
			}
			break;
		    } else if (sqlca.sqlcode == 100)
			break;
		    tries = 0;
		    /* Save number of keys fetched this round. */
		    i = 0;
		    fetchcount = sqlca.sqlerrd[2];
		    if (Debug)
		      fprintf( stderr, "Fetchcount = %d\n", fetchcount );
		    for (; i < fetchcount && dsiz < SBsz; i++) {
			strcpy( ip++, "," );
			if (use_ifx_row_id == 1) {
			  sprintf( tmp,
				   "'%s'",
				   (char *)(col[0].sqldata+(i*col[0].sqllen)) );
			} else {
			  sprintf( tmp,
				   "%d",
				   *(int *)(col[0].sqldata+(i*col[0].sqllen)) );
			}
			strcpy( ip, tmp );
			ip += (tl = strlen( tmp ));
			dsiz += (tl + 1);
		    }
		}
		strcpy( ip, " )" );
		
		if (Debug)
		    fprintf( stderr, "%s\n\n", delete_stmt );
		
		EXEC SQL EXECUTE IMMEDIATE :delete_stmt;
		if ( sqlca.sqlcode != 0 ) {
	            ERR_OUT("Error on EXECUTE delete 3rd in loop");
		    return 1;
		}
		counter += sqlca.sqlerrd[2];
	    }
	}
	if (commit_ok) {
	    EXEC SQL COMMIT WORK;
	    if ( sqlca.sqlcode != 0 ) {
	        ERR_OUT("Error on COMMIT WORK 3rd in loop");
		errlvl = 1;
		break;
	    }
	}
	reccount += fetchcount;
	uncommitted = 0;
	if (!Silent)
	    fprintf(stderr, "Deleted: %d\n", counter );

        if (LoopPause) {
           if (verbose) {
               fprintf( stderr, "Sleep %d\n",LoopPause);
               fflush( stderr );
           }
           sleep_ms(LoopPause);
        }

	if (begin_ok) {
	    if (no_replicate)
		EXEC SQL BEGIN WORK WITHOUT REPLICATION;
	    else
		EXEC SQL BEGIN WORK;
	    if ( sqlca.sqlcode != 0 ) {
	        ERR_OUT("Error on BEGIN WORK end of loop");
		return 1;
	    }
	}
    } /* end fetch loop */

#ifdef TWOCONNECT
    EXEC SQL SET CONNECTION "target";
    if (sqlca.sqlcode != 0) {
	ERR_OUT("Error on SET CONNECTION TO target after loop");
	errlvl = 1;
	exit(97);
    }
#endif

    if (!Silent) {
	printf( "Input:  %d records.\n", reccount );
	printf( "Deleted: %d records.\n", counter );
    }

#ifdef TWOCONNECT
    EXEC SQL SET CONNECTION "source";
    if ( sqlca.sqlcode != 0 ) {
	ERR_OUT("Error on SET CONNECTION TO source after loop");
	return 1;
    }    
#endif

    if (is_open) {
        if (Debug)
	    fprintf( stderr, "Closing cursor at li %d.\n", __LINE__ );
        EXEC SQL CLOSE fetch_curs;
	is_open = 0;
	if ( sqlca.sqlcode != 0 && sqlca.sqlcode != -404) {
	  ERR_OUT("Error on CLOSE fetch_curs after loop");
	  return 1;
	}    
    }
   
    free(buffer);

    return errlvl;
}

#if defined DEFINE_UNUSED
/* ----------------------------------------------------------------------- */

char *strip( char *str, int len )
{
    char *p;

    for (p = (str + len - 1); p > str; p--) {
	if (*p == ' ' || *p == (char)0)
	    *p = (char)0;
	else
	    break;
    }
    return str;
}

char *charcpy( char *target, char *source, long len )
{
    while (len && *source) {
	if (*source == *delim) {
	    *target++ = '\\';
	    *target = *source;
	} else {
	    *target = *source;
	}
	source++;
	target++;
	len--;
    }
    *target++ = *delim;
    *target = (char)0;
}

/* ----------------------------------------------------------------------- */
#endif


void usage( char *progname )
{
    char *p;

    p = strrchr( progname, '/' );
    if (!p)
	p = progname;
    fprintf( stderr, "\n");
    fprintf( stderr,
	     "%s - Delete rows from an Informix table.\n\n",
	     p );

    fprintf( stderr, "Usage: %s\n", p );	
    fprintf( stderr, 
	     "\t\t -d database\n"
	     "\t\t -b FetBufSize\n"
	     "\t\t -t table\n"
	     "\t\t<-h host>\n"
	     "\t\t<-S> <-V> <-R> <-I> <-v> \n"
	     "\t\t<-w waittime>\n"
	     "\t\t<-u keylist>\n"
	     "\t\t<-o>\n"
	     "\t\t<-s where clause | -q select stmt>\n"
	     "\t\t<-B SQL bufsize>\n"
	     "\t\t<-p PDQRPIORITY>\n\n" );

    fprintf( stderr, "\t-d - Database from which to delete.\n" );
    fprintf( stderr,
	     "\t-b - Size of a transaction buffer.  Controls the number of rows deleted\n"
	     "\t    in a single transaction. (Default: 32767. Range: 2048 -> 32767)\n" );
    fprintf( stderr, "\t-t - Table from which data is to be deleted.\n" );
    fprintf( stderr, "\t-h - Host to delete rows from (Default: $INFORMIXSERVER).\n" );
    fprintf( stderr, "\t-S - Silent mode.  Supress messages.\n" );
    fprintf( stderr, "\t-V - Print version information and exit.\n" );
    fprintf( stderr, "\t-v - Enable verbose output.\n" );
    fprintf( stderr,
	     "\t-I - Use ifx_row_id instead of rowid. Supports partitioned tables\n"
	     "\t     that were not created WITH ROWIDS. Should be faster for other\n"
	     "\t     partitioned tables as well.\n" );
    fprintf( stderr, "\t-P - Pause in milli-seconds between deletes.\n" );
    fprintf( stderr,
	     "\t-w - # of seconds to wait for locked rows to be released (if "
	     "< 0 wait\n\t     forever).  (Default 10.)\n" );
    fprintf( stderr,
	     "\t-u - Comma separated list of columns forming a key which can be"
	     " used to \n\t     identify rows to delete.  Separate -u options"
	     " may be used. \n\t     (Default: ROWID.)\n" );
    fprintf( stderr,
	     "\t-o - Do not close the fetch cursor between fetch rounds.  This is\n"
	     "\t    slower, however, it will prevent dbdelete from deleting rows that\n"
	     "\t    have been inserted since the original cursor was opened.  Note:\n"
	     "\t    this option does NOT eliminate the dangers of using dbdelete on\n"
	     "\t    a table that is being actively inserted to.  Rows inserted BEFORE\n"
	     "\t    the cursor is open are still at risk.  Caviat emptor applies.\n" );
    fprintf( stderr,
	     "\t-s - Where clause for use in selecting ROWIDS or row keys to"
	     " delete.\n\t     Do not include the keyword \"WHERE\".\n" );
    fprintf( stderr,
	     "\t-q - Full SELECT statement to select rowids (or keys) to delete.  If \n"
	     "\t     keys are returned you must also include -u\n" );
    fprintf( stderr,
	    "\t-B - Set the size of the SQL DELETE statements that are "
	    "generated.\n" );
    fprintf( stderr, 
	    "\t     (DEFAULT: 2048, MAXIMUM: 65500)  N.B. 2K seemed most "
	    "efficient in\n\t     testing.  This option is effective only for "
	     "delete by rowid (ie no \n\t     '-u' options).\n" );
    fprintf( stderr, "\t-p - PDQPRIORITY to set for the task.\n" );
    fprintf( stderr,
	     "\t-R - Do not replicate deletions.  Begins transactions with the WITHOUT \n"
	     "\t    REPLICATION clause included.\n\n" );
    fprintf( stderr,
	     "\nThe standard Informix environment variable FET_BUF_SIZE will override the \n"
	     "default value of the (-b flag) Commit size if within the valid range.  Higher \n"
	     "values for -b will commit larger blocks of rows deleted and use more locks. \n\n" );
}
