/* dbmove.ec - Program to move records from one database table to another
                without using too many locks.  Uses a maximum of 32767 locks
                when inserting using a 1byte key or 8191 sets of locks when using
                ROWID.  Don't forget to multiply by the number of indexes on the
                table.  
*/
/*
   RCS Header:
   -----------
   $Header: /home/art/GoogleDrive/Latest/RCS/dbmove.ec,v 1.15 2021/12/21 17:12:40 art Exp $

   RCS Log:
   --------
   $Log: dbmove.ec,v $
   Revision 1.15  2021/12/21 17:12:40  art
   Updated Copyright notice.

   Revision 1.14  2021/07/30 01:45:36  art
   Added -K & -k to use any integer type single key column instead of ROWID or ifx_row_id.
   Good for partitioned tables on servers older than 14.10.FC6 which fixed the ifx_row_id
   addressing problem.

   Revision 1.13  2020/07/21 17:39:23  art
   Added support for ifx_row_id for partitioned tables created without the
   WITH ROWID option. Updated the copyright date.

   Revision 1.12  2019/03/15 15:48:47  art
   Fixed some compiler warnings.

   Revision 1.11  2016/06/27 21:16:42  art
   Updated copyright notice.

   Revision 1.10  2015/01/21 17:56:22  art
   Fixed usage for -r.

   Revision 1.9  2013/12/29 21:31:33  art
   Fixed warnings.

   Revision 1.8  2013/11/20 16:25:26  art
   Added options (-c & -C) to allow specification of a list of columns for both
   the select statement projection list and the insert statement column list.
   Fixed -d/-D, -t/-T, & -h/-H to reasonably default if not specified.

   Revision 1.7  2013/08/14 13:02:52  art
   Lots of code cleanup of legacy dbdelete code that's not needed here.
   Cleaned up trace and enforced -S better.
   More cleanup of the Usage display.

   Revision 1.6  2013/08/11 23:48:07  art
   Removed -q code.  Fixed more of the usage up.  Streamlined code referencing removed options.

   Revision 1.5  2013/08/09 15:54:16  art
   Fixed BEGIN inside a transaction error and end-of-processing silence.

   Revision 1.4  2013/08/09 14:42:14  art
   Stripped out unneccesary code.  Improved runtime traces.

   Revision 1.3  2013/08/09 00:30:48  art
   Added commit after each INSERT to minimize rollbacks.  May make that an option at some point.

   Revision 1.1  2013/08/08 18:57:10  art
   Initial revision


*/

static char RCSHeader[] = "$Header: /home/art/GoogleDrive/Latest/RCS/dbmove.ec,v 1.15 2021/12/21 17:12:40 art Exp $";
static char RCSWhat[] = "<@(#) $RCSfile: dbmove.ec,v $	 $Revision: 1.15 $ $Date: 2021/12/21 17:12:40 $ >";

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

EXEC SQL include sqlca;
EXEC SQL include sqlda;
EXEC SQL include sqltypes;
EXEC SQL include decimal;
EXEC SQL include locator;

#define FLDSIZE  40
#define BUFFSIZE 256
#define DELIM "|"
#define FETBUFSIZE 32767

/* PROTOTYPES */
int copy_table_to_table( void );
void usage( char *progname );
void dumprec( struct sqlda * );

#if defined( _AIX ) & !defined( __GNUC__ )
#	define inline
#endif

/* GLOBALS - I hate these things. */
FILE *errlog;
char errfile[1025] = "dbmove.unl";
char delim[2] = DELIM;
int begin_ok = 0, commit_ok = 0; /*  True if DB has transactions.  */
int use_ins_curs = 0;
int LoseDups = 0;
char DelKey[129];
int Silent = 0;
int SBsz = 2048;
int lclFetBufSize = FETBUFSIZE;
int is_open = 0;
int one_open = 0;
int no_replicate = 0;
int Debug = 0;
int this_rowid = 0;
int Use_Key_Type = 0;
char this_ifx_row_id[46];
long long this_key;
char id_str[133] = "ROWID";
char id_str2[133] = "ROWID";
char colsin[33001], colsout[33001];

EXEC SQL BEGIN DECLARE SECTION;
int last_rowid = 0;
int hold_rowid = 0;
bigint last_key = 0;
bigint hold_key = 0;
string last_ifx_row_id[46];
string hold_ifx_row_id[46];
int priority = 0, waittime = 10;
string dbname_from[130];
string host_from[102];
string tabname_from[130];
string dbname_to[130];
string host_to[100];
string tabname_to[130];
string qrystmt[3001];
string selstmt[2001] = "";
string copy_stmt[65530];
string temp[2001];
EXEC SQL END DECLARE SECTION;

#if ! (defined( __DGUX__ ) & defined( __ix86 ))
extern short FetBufSize;
extern short FetArrSize;
#endif

#define Revision "$Revision: 1.15 $"

int main( int argc, char **argv)
{
    int n, ch, errval;
    char PDQ[50], *p;
    static char revision[1024], *envFetBufSize;
    int fullstmt = 0;

    if ( argc == 1 ) {
	usage( argv[0] );
	exit( 2 );
    }

    host_to[0] = (char)0;
    host_from[0] = (char)0;

    dbname_to[0] = (char)0;
    dbname_from[0] = (char)0;

    tabname_to[0] = (char)0;
    tabname_from[0] = (char)0;

    colsin[0] = (char)0;
    colsout[0] = (char)0;

    DelKey[0] = (char)0;
    last_key = 0LL;

    strcpy( last_ifx_row_id, "0:0" );

    if ((envFetBufSize = getenv( "FET_BUF_SIZE" )) != (char *)NULL) {
	lclFetBufSize = atoi( envFetBufSize );
	if (lclFetBufSize < 4096)
	    lclFetBufSize = 4096;
	if (lclFetBufSize > 32767)
	    lclFetBufSize = 32767;
    }

    while ((ch = getopt( argc, argv, "c:C:i:r:b:d:h:k:op:s:t:w:D:H:B:IK:RST:V" )) != EOF) 
	switch (ch) {
	case 'K':
	    if (Use_Key_Type == 0) {
	        strncpy( DelKey, optarg, sizeof DelKey );
	        Use_Key_Type = 2;
	    } else {
		fprintf( stderr, "Only one of -I or -K permitted.\n" );
		usage( argv[0] );
		exit(1);
	    }
	    break;
	case 'k':
	    last_key = atoll( optarg );
	    break;
	case 'I':
	    if (Use_Key_Type == 0)
	        Use_Key_Type = 1;
	    else {
		fprintf( stderr, "Only one of -I or -K permitted.\n" );
                usage( argv[0] );
                exit(1);
            }

	    break;
	case 'R':
	    no_replicate = 1;
	    break;
	case 'o':
	    one_open = 1;
	    break;
	case 'b':
	    lclFetBufSize = atoi( optarg );
	    break;
	case 'i':
	  strncpy( last_ifx_row_id, optarg, sizeof last_ifx_row_id  );
	  break;
	case 'r':
	    last_rowid = atoi( optarg );
	    break;
	case 'V':
	    strcpy( revision, Revision );
	    fprintf( stdout,
		     "\n%s: Source Version %s!\n",
		     argv[0],
		     strtok( revision, "$" ) );

	    fprintf( stdout, "\tCopyright 1996-2022 by Art S. Kagel.\n" );
            fprintf( stdout,
                     "\tLicense for private usage (including use by "
                     "commercial enterprise\n"
                     "\tfor its own needs) is granted.  Commercial and "
                     "exploitation rights\n"
                     "\tare reserved.\n\n" );
	    fflush( stdout );
	    exit(0);
	    break;
	case 'B':
	    SBsz = atoi( optarg );
	    break;
	case 'S':
	    Silent = 1;
	    break;
	case 'w':
	    waittime = atoi( optarg );
	    if (!waittime)
		waittime = 10;
	    break;
	case 'h':
	    /* Copy from host. */
	    snprintf( host_from, sizeof host_from, "@%.99s", optarg );
	    host_from[sizeof host_from -1] = (char)0;
	    break;
	case 'd':
	    /* Copy from database. */
	    strncpy( dbname_from, optarg, 129 );
	    dbname_from[128] = (char)0;
	    break;
	case 't':
	    /* Table to copy from. */
	    strncpy( tabname_from, optarg, 129 );
	    tabname_from[128] = (char)0;
	    break;
	case 'H':
	    /* Copy to host. */
	    sprintf( host_to, "@%-1s", optarg );
	    host_to[sizeof host_to -1] = (char)0;
	    break;
	case 'D':
	    /* Copy to database. */
	    strncpy( dbname_to, optarg, 129 );
	    dbname_to[128] = (char)0;
	    break;
	case 'T':
	    /* Table to copy to. */
	    strncpy( tabname_to, optarg, 129 );
	    tabname_to[128] = (char)0;
	    break;
	case 's':
	    strncpy( selstmt, optarg, 2000 );
	    selstmt[2000] = (char)0;
	    break;
	case 'p':
	    priority = atoi( optarg );
	    break;
	case 'c':
	    strncpy( colsin, optarg, 33000 );
	    colsin[33000] = (char)0;
	    break;
	case 'C':
	    strncpy( colsout, optarg, 33000 );
	    colsout[33000] = (char)0;
	    break;
	default:
	    usage( argv[0] );
	    exit( 2 );
	    break;
	}

    if (getenv( "DBMOVE_DEBUG" ))
        Debug = 1;

    if (lclFetBufSize < 4096 || lclFetBufSize > 32767) {
	fprintf( stdout,
		 "Invalid communication buffer size (%d)!\n",
		 lclFetBufSize );
	fflush( stdout );
	usage( argv[0] );
	exit( 2 );
    }

    if (!SBsz)
	SBsz = 2032;
    else if (SBsz > 65500) {
	fprintf( stdout, 
		"SQL Buffer size (%d) exceeds maximum (65500).  "
		"Adjusting to maximum.\n", 
		SBsz );
	fflush( stdout );
	SBsz = 65500;
    } else
	SBsz -= 16;

    /* Set up connection for target database connection. */
    if (strlen( host_from ) == 0) {
	snprintf( host_from,
		  sizeof host_from,
		  "@%s",
		  getenv( "INFORMIXSERVER" ) );
    }
    if (strlen(host_to) == 0)
	strcpy( host_to, host_from );
    if (strlen(dbname_to) != 0 && strlen(dbname_from) == 0)
	strcpy( dbname_from, dbname_to );
    else if (strlen(dbname_to) == 0 && strlen(dbname_from) != 0)
	strcpy( dbname_to, dbname_from );

    if (strcmp( host_to, host_from ) == 0 &&
        strcmp( dbname_to, dbname_from ) == 0 &&
	strcmp( tabname_to, tabname_from ) == 0) {
	fprintf( stdout,
		 "Cannot move data from a table to itself!\n" );
	fflush( stdout );
	return 1;
    }

    /* If the user specifies an input column list for the select,
       but no output column list for the insert, assume they should
       be the same. */
    if (strlen( colsout ) == 0 && strlen( colsin ) != 0)
	strcpy( colsout, colsin );

    /* Assume that the user has given us a where clause. */
    
    if (Use_Key_Type == 1) {
      strcpy( id_str, "ifx_row_id" );
      strcpy( id_str2, "ifx_row_id" );
    } else if (Use_Key_Type == 2) {
      snprintf( id_str, sizeof id_str, "%s::BIGINT", DelKey );
      snprintf( id_str2, sizeof id_str2, " %s ", DelKey );
    }
    
    if (strlen( selstmt ))
	sprintf( qrystmt,
		 "SELECT %s FROM %s WHERE %s AND %s > ? ORDER BY %s ",
		 id_str,
		 tabname_from,
		 selstmt,
		 id_str2,
		 id_str2);
    else
        sprintf( qrystmt, 
		 "SELECT %s FROM %s WHERE %s > ? ORDER BY %s ",
		 id_str,
		 tabname_from,
		 id_str2,
		 id_str2 );
	
    if (!Silent) {
	fprintf( stdout, "Selecting data with:\n\t%s;\n\n", qrystmt );
	fflush( stdout );
    }
	    
    /* Set up PDQPRIORITY. */
    sprintf( PDQ, "PDQPRIORITY=%d", priority );
    putenv( PDQ );

    EXEC SQL CONNECT TO :host_from AS "target" WITH CONCURRENT TRANSACTION;
    if (sqlca.sqlcode < 0) {
	fprintf( stdout,
		 "Error opening target host connection: %d, %d\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	fflush( stdout );
	return 1;
    }

    EXEC SQL  database :dbname_from;
    if (sqlca.sqlcode < 0) {
	fprintf( stdout,
		 "Error opening target database: %d, %d\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	fflush( stdout );
	return 1;
    }
    commit_ok = (sqlca.sqlwarn.sqlwarn1 == 'W');
    begin_ok  = (commit_ok && sqlca.sqlwarn.sqlwarn2 != 'W');

    if (no_replicate && !begin_ok) {
	fprintf( stdout,
		 "Transactions without replication (-R) specified but database %s %s ignoring.\n",
		 dbname_from,
		 commit_ok ? "is ANSI mode" : "does not \nsupport transactions" );
	fflush( stdout );
	no_replicate = 0;
    }

    EXEC SQL SET ISOLATION TO DIRTY READ;
    if (sqlca.sqlcode < 0 && sqlca.sqlcode != -256) {
	fprintf( stdout,
		 "Error setting Isolation Level for target: %d, %d\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	fflush( stdout );
	return 1;
    }    

    if (waittime > 0)
	sprintf( temp, "SET LOCK MODE TO WAIT %d;", waittime );
    else
	strcpy( temp, "SET LOCK MODE TO WAIT;" );
    EXEC SQL EXECUTE IMMEDIATE :temp;
    if (sqlca.sqlcode < 0) {
	fprintf( stdout,
		 "Error setting lock mode for target: %d, %d\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	fflush( stdout );
	return 1;
    }    

    EXEC SQL IFNDEF ifmx_6_0;
    if (priority > 50)
	EXEC SQL SET PDQPRIORITY 60;
    else if (priority > 40)
	EXEC SQL SET PDQPRIORITY 50;
    else if (priority > 30)
	EXEC SQL SET PDQPRIORITY 40;
    else if (priority > 20)
	EXEC SQL SET PDQPRIORITY 30;
    else if (priority > 10)
	EXEC SQL SET PDQPRIORITY 20;
    else 
	EXEC SQL SET PDQPRIORITY 10;
    if (sqlca.sqlcode < 0) {
	fprintf( stdout,
		 "Warning setting PDQPRIORITY on target: %d, %d\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	fflush( stdout );
    }
    EXEC SQL ENDIF;

    errval = copy_table_to_table();
    if (errlog != (FILE *)NULL) {
        fflush( errlog );
	fclose( errlog );
    }
    return errval;
}

/* ---------------------------------------------------------------------- */

int copy_table_to_table( void )
{
    register int pos;
    register char *cp, *cp2;
    register int len;
    register int i, j;
    char tmp[100];
    struct sqlda *udesc, *cdesc;
    struct sqlvar_struct *col, *col2;
    char *buffer = NULL, *buffer2 = NULL;
    char *ip, *ip2 = copy_stmt;
    int counter = 0, ii, errcount = 0, errlvl = 0, reccount = 0, dupslost = 0;
    int uncommitted = 0, tries = 0, fetchcount = 0;
    int dsiz, tl;
    char string[1025];
    char **bufary;
    short **ipary;
    short *colszs;

    EXEC SQL PREPARE fetcher FROM :qrystmt;
    if ( sqlca.sqlcode != 0 ) {
	fprintf( stdout,
		 "Error: %d(%d) on prepare of fetch at pos: %d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1],
		 sqlca.sqlerrd[4] );
	fflush( stdout );
	if (sqlca.sqlcode == -201) {
	    fprintf( stdout,
		     "Syntax error at char %d in:\n%s\n",
		     sqlca.sqlerrd[4],
		     qrystmt );
	    fflush( stdout );
	} else if (sqlca.sqlcode == -857)
	    fprintf( stdout,
		     "\n\tTable is fragmented without ROWID.  "
		     "Rerun specifying key column(s) if -I.\n" );
	fflush( stdout );
	return 1;
    }
 
    EXEC SQL DESCRIBE fetcher INTO udesc;
    if ( sqlca.sqlcode != 0 ) {
        fprintf( stdout,
                 "Error: %d(%d) on describe.\n",
                 sqlca.sqlcode,
                 sqlca.sqlerrd[1] );
	fflush( stdout );
        return 1;
    }
    /* Create an sqlda struture for the insert statement that does not use the
       insert array code. */
    cdesc = (struct sqlda *)malloc( sizeof (struct sqlda) );
    memset( cdesc, 0, sizeof (struct sqlda) );
    cdesc->sqld = udesc->sqld;
    cdesc->sqlvar = (struct sqlvar_struct *)
        malloc( cdesc->sqld * sizeof (struct sqlvar_struct) );
    memset( cdesc->sqlvar, 0, cdesc->sqld * sizeof (struct sqlvar_struct) );
    strncpy( cdesc->desc_name, udesc->desc_name, 19 );
    cdesc->desc_occ = udesc->desc_occ;
    cdesc->desc_next = (struct sqlda *)NULL;
 
    /* Step 1: analyze udesc to determine type and memory requirements
     *         of each item in the select list.  rtypalign() returns a
     *         pointer to the next appropriate boundary (starting at
     *         pos) for the indicated data type.
     */
    pos = 0;
 
    /* Allocate pointer array to hold the fetch array pointers. */
 
    for (col = udesc->sqlvar, i = 0, col2 = cdesc->sqlvar;
         i < udesc->sqld;
         col++, i++, col2++)
    {
        switch(col->sqltype & SQLTYPE)
        {
        case SQLSMFLOAT:
            col->sqltype = CFLOATTYPE;
            col2->sqltype = CFLOATTYPE;
            break;
 
        case SQLFLOAT:
            col->sqltype = CDOUBLETYPE;
            col2->sqltype = CDOUBLETYPE;
            break;
 
        case SQLMONEY:
        case SQLDECIMAL:
            col->sqltype = CDECIMALTYPE;
            col2->sqltype = CDECIMALTYPE;
            break;
 
        case SQLSERIAL:
        case SQLINT:
            col->sqltype = CINTTYPE;
            col2->sqltype = CINTTYPE;
            break;
 
        case SQLBIGSERIAL:
        case SQLINFXBIGINT:
            col->sqltype = CBIGINTTYPE;
            col2->sqltype = CBIGINTTYPE;
            break;
 
        case SQLSMINT:
            col->sqltype = CSHORTTYPE;
            col2->sqltype = CSHORTTYPE;
            break;
 
        case SQLVCHAR:
        case SQLCHAR:
	  if (strcmp( col->sqlname, "ifx_row_id" ) == 0 && Use_Key_Type == 1) {
            col->sqltype = CSTRINGTYPE; /* get all bytes */
            col2->sqltype = CSTRINGTYPE; /* get all bytes */
	  } else {
            col->sqltype = CFIXCHARTYPE; /* get all bytes */
            col2->sqltype = CFIXCHARTYPE; /* get all bytes */
	  }
            break;
 
        case SQLDATE:
            col->sqltype = CDATETYPE;
            col2->sqltype = CDATETYPE;
            break;
 
        case SQLDTIME:
            col->sqltype = CFIXCHARTYPE;
            col2->sqltype = CFIXCHARTYPE;
            col2->sqllen = col->sqllen = rtypmsize( col->sqltype, 26 );
            break;
 
        case SQLINTERVAL:
            col->sqltype = CFIXCHARTYPE;
            col2->sqltype = CFIXCHARTYPE;
            col2->sqllen = col->sqllen = rtypmsize( col->sqltype, 26 );
            break;
 
        default:
            /* Other data types.  Do nothing if we see an unsupported type. */
            fprintf( stdout, "copy_table_to_table: Invalid data type: %d \n",
                     col->sqltype );
	    fflush( stdout );
            return 1;
        }
        switch(col->sqltype & SQLTYPE)
        {
        case CDTIMETYPE:
        case CINVTYPE:
          col2->sqllen = col->sqllen = rtypmsize( col->sqltype, 24 );
          break;
        case CDECIMALTYPE:
          col2->sqllen = col->sqllen = rtypmsize( col->sqltype, 22 );
          break;
        default:
          col2->sqllen = col->sqllen = rtypmsize(col->sqltype,col->sqllen);
        }
        pos = (int) rtypalign(pos,col->sqltype) + col->sqllen;
        col2->sqlname = (char *)malloc( strlen( col->sqlname ) + 1 );
        strcpy( col2->sqlname, col->sqlname );
    }
 
    /* Step 2: Allocate memory to receive rows of the table returned
     *         by the SELECT statement.  The variable pos has an integer
     *         value equal to the number of bytes occupied by the row.
     */
    /* Allocate a buffer for the insert sqlda structure. */
    buffer = malloc( pos );
    memset( buffer, 0, pos );
    /* OK buffer is now the insert buffer.  Bufary is now the array of
       pointers to the fetch buffers for each column. */
    /* Set FetBufSize and GetArrSize and allocate bufary. */
    if (pos > lclFetBufSize) {
        /* For large records, can only fetch one at a time. */
        FetBufSize = lclFetBufSize;
        FetArrSize = 1;
    } else {
        FetArrSize = (lclFetBufSize / pos);
        FetBufSize = FetArrSize * pos;
    }
    if (!Silent) {
	fprintf( stdout, "Copying blocks of %d records.\n", FetArrSize );
	fflush( stdout );
    }
 
    /* Bufary and coszs need one pointer per column in the result set. */
    bufary = (char **)malloc( udesc->sqld * sizeof (char *) );
    colszs = (short *)malloc( udesc->sqld * sizeof (short) );
 
    /* Allocate indicator list. */
    /* Array of indicators for fetch sqldat structure. */
    ipary = (short **)malloc( udesc->sqld * sizeof (short *) );
    if (buffer == NULL || ipary == NULL || colszs == NULL) {
        fprintf(stdout,
		"Out of memory(%lx|%lx|%lx).\n",
		(long)buffer,
		(long)ipary,
		(long)colszs );
	fflush( stdout );
        return 1;
    }
 
    /* Step 3: Set pointers in the allocated memory to receive each
     *         item in the select list.
     */

    /* Initialize pointer to insert buffer. */
    cp = buffer;
    for (col = udesc->sqlvar, col2 = cdesc->sqlvar, i = 0;
         i < udesc->sqld;
         col2++, col++, i++)
    {
        colszs[i] = col->sqllen;
        /* Allocate Fetch array for this column. */
        bufary[i] = malloc( col->sqllen * FetArrSize );
        /* Allocate indicator array for this column. */
        ipary[i] = (short *)malloc( FetArrSize * sizeof (short) );
 
        /* Adjust pointer for each column for alignment in insert sqlda */
        cp = buffer + rtypalign((int)(cp - buffer),col->sqltype);
        /* Assign indicator addresses. */
        col->sqlind = ipary[i];
        /* Assign the data pointers. */
        col->sqldata = bufary[i];
        col2->sqldata = cp;
        if (Debug) {
          fprintf( stdout,
                   "Setting pointer to %p for %d bytes.\n",
                   cp,
                   col->sqllen );
	  fflush( stdout );
	}
        cp += col->sqllen;
    }
 
    /* Step 4: Declare fetch cursor. */
    EXEC SQL DECLARE fetch_curs CURSOR WITH HOLD FOR fetcher;
    if ( sqlca.sqlcode != 0 ) {
        fprintf( stdout,
                 "Error: %d(%d) on declare of fetch_curs.\n",
                 sqlca.sqlcode,
                 sqlca.sqlerrd[1] );
	fflush( stdout );
        return 1;
    }

    if (strlen( colsin ) > 0)
	sprintf( copy_stmt, "INSERT INTO %s%s:%s (%s)", dbname_to, host_to, tabname_to, colsin );
    else
	sprintf( copy_stmt, "INSERT INTO %s%s:%s ", dbname_to, host_to, tabname_to );

    if (strlen( colsout ) > 0) {
	char tmpstr[33200];

	snprintf( tmpstr,
		  sizeof tmpstr,
		  "SELECT %s FROM ",
		  colsout );
	strncat( copy_stmt,
		 tmpstr,
		 sizeof tmpstr );
    } else
	strcat( copy_stmt,
		 "SELECT * FROM " );
    strncat( copy_stmt,
	     tabname_from,
	     sizeof tabname_from );
    strcat( copy_stmt,
	     " WHERE " );
    if (Use_Key_Type == 1)
      strcat( copy_stmt, " ifx_row_id " );
    else if (Use_Key_Type == 2)
      strcat( copy_stmt, id_str2 );
    else
      strcat( copy_stmt, " ROWID " );
    strcat( copy_stmt,
	     " IN ( " );
    ip2 = copy_stmt + strlen( copy_stmt );

    if (!Silent) {
	fprintf( stdout, "Moving data with:\n\t%s... );\n\n", copy_stmt );
	fflush( stdout );
    }

    if (begin_ok) {
	if (no_replicate)
	    EXEC SQL BEGIN WORK WITHOUT REPLICATION;
	else
	    EXEC SQL BEGIN WORK;
	if ( sqlca.sqlcode != 0 ) {
	    fprintf( stdout,
		     "Error: %d(%d) on BEGIN WORK at %d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1],
		     __LINE__ );
	    fflush( stdout );
	    fprintf( stdout, "Last good rowid = %d.\n", last_rowid );
	    fflush( stdout );
	    return 1;
	} else if (Debug) {
	    fprintf( stdout, "Beginning work at: %d.\n", __LINE__ );
	}
    }

    sqlca.sqlcode = 0;
    for ( ;; ) {
	/* Open the fetch cursor. */
	if (!is_open) {
	    fflush( stdout );
	    if (Debug) {
	      fprintf( stdout, "Opening cursor at li %d.\n", __LINE__ );
	      if (Use_Key_Type == 1)
		fprintf( stdout, "   Using last_ifx_row_id: %s.\n", last_ifx_row_id );
	      else if (Use_Key_Type == 2)
		fprintf( stdout, "   Using %s: %lld.\n", DelKey, last_key);
	      else
		fprintf( stdout, "   Using rowid: %d.\n", last_rowid );
	      fflush( stdout );
	    }
	    if (Use_Key_Type == 1) {
	      if (!Silent && !Debug)
		fprintf( stdout, "Opening cursor with ifx_row_id > %s at line %d.\n", last_ifx_row_id, __LINE__ );
	      EXEC SQL OPEN fetch_curs USING :last_ifx_row_id;
	    } else if (Use_Key_Type == 2) {
	      if (!Silent && !Debug)
		fprintf( stdout, "Opening cursor with %s > %lld at line %d.\n", DelKey, last_key, __LINE__ );
	      EXEC SQL OPEN fetch_curs USING :last_key;
	    } else {
	      if (!Silent && !Debug)
		fprintf( stdout, "Opening cursor with rowid > %d at line %d.\n", last_rowid, __LINE__ );
	      EXEC SQL OPEN fetch_curs USING :last_rowid;
	    }
	    is_open = 1;
	}
	if ( sqlca.sqlcode != 0 ) {
	    fprintf( stdout,
		     "Error: %d(%d) on OPEN fetch_curs.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    fflush( stdout );
	    if (Use_Key_Type == 1)
	      fprintf( stdout, "Last good ifx_row_id = %s.\n", last_ifx_row_id );
	    else if (Use_Key_Type == 2)
	      fprintf( stdout, "Last good %s = %lld.\n", DelKey, last_key );
	    else
	      fprintf( stdout, "Last good rowid = %d.\n", last_rowid );
	    fflush( stdout );
	    return 1;
	}
 
	EXEC SQL FETCH fetch_curs USING DESCRIPTOR udesc;

	if (sqlca.sqlcode < 0) {
	    fprintf( stdout,
		     "Error fetching row(%d)!  Code=%d, ISAM=%d at line %d.\n",
		     reccount,
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1],
		     __LINE__ );
	    fflush( stdout );
	    if (sqlca.sqlcode == -243 && tries < 5)
	    {
		tries++;
		if (!one_open) {
		    if (Debug) {
		        fprintf( stdout, "Closing cursor at li %d.\n", __LINE__ );
			fflush( stdout );
		    }
		    EXEC SQL CLOSE fetch_curs;
		    is_open = 0;
		}
		if ( sqlca.sqlcode != 0 ) {
		    fprintf( stdout,
			     "Error: %d(%d) on CLOSE fetch_curs.\n",
			     sqlca.sqlcode,
			     sqlca.sqlerrd[1] );
		    fflush( stdout );
		    if (Use_Key_Type == 1)
		      fprintf( stdout, "Last good ifx_row_id = %s.\n", last_ifx_row_id );
		    else if (Use_Key_Type == 2)
		      fprintf( stdout, "Last good %s = %lld.\n", DelKey, last_key );
		    else
		      fprintf( stdout, "Last good rowid = %d.\n", last_rowid );
		    fflush( stdout );
		    return 1;
		} else {
		  if (Use_Key_Type == 1) 
		    strcpy( last_ifx_row_id, this_ifx_row_id );
		  else if (Use_Key_Type == 2) 
		    last_key = this_key;
		  else
		    last_rowid = this_rowid;
		}

		if (!is_open) {
		    fflush( stdout );
		    if (Debug) {
		        fprintf( stdout, "Opening cursor at li %d.\n", __LINE__ );
			if (Use_Key_Type == 1)
			  fprintf( stdout, "   Using last_ifx_row_id: %s.\n", last_ifx_row_id );
			else if (Use_Key_Type == 2)
			  fprintf( stdout, "   Using %s: %lld.\n", DelKey, last_key );
			else
			  fprintf( stdout, "   Using rowid: %d.\n", last_rowid );
			fflush( stdout );
		    }
		    if (Use_Key_Type == 1) {
		      if (!Silent && !Debug)
			fprintf( stdout, "Opening cursor with ifx_row_id > %s at line %d.\n", last_ifx_row_id, __LINE__ );
		      EXEC SQL OPEN fetch_curs USING :last_ifx_row_id;
		    } else if (Use_Key_Type == 2) {
		      if (!Silent && !Debug)
			fprintf( stdout, "Opening cursor with %s > %lld at line %d.\n", DelKey, last_key, __LINE__ );
		      EXEC SQL OPEN fetch_curs USING :last_key;
		    } else {
		      if (!Silent && !Debug)
			fprintf( stdout, "Opening cursor with rowid > %d at line %d.\n", last_rowid, __LINE__ );
		      EXEC SQL OPEN fetch_curs USING :last_rowid;
		    }
		    is_open = 1;
		}
		if ( sqlca.sqlcode != 0 ) {
		    fprintf( stdout,
			     "Error: %d(%d) on REOPEN fetch_curs.\n",
			     sqlca.sqlcode,
			     sqlca.sqlerrd[1] );
		    fflush( stdout );
		    if (Use_Key_Type == 1)
		      fprintf( stdout, "Last good ifx_row_id = %s.\n", last_ifx_row_id );
		    else if (Use_Key_Type == 2)
		      fprintf( stdout, "Last good %s = %lld.\n", DelKey, last_key );
		    else
		      fprintf( stdout, "Last good rowid = %d.\n", last_rowid );
		    fflush( stdout );
		    return 1;
		}
		if (uncommitted && commit_ok) {
		    EXEC SQL COMMIT WORK;
		    if ( sqlca.sqlcode != 0 ) {
			fprintf( stdout,
				 "Error: %d(%d) on COMMIT WORK in loop.\n",
				 sqlca.sqlcode,
				 sqlca.sqlerrd[1] );
			fflush( stdout );
			if (Use_Key_Type == 1)
			  fprintf( stdout, "Last good ifx_row_id = %s.\n", last_ifx_row_id );
			else if (Use_Key_Type == 2)
			  fprintf( stdout, "Last good %s = %lld.\n", DelKey, last_key );
			else
			  fprintf( stdout, "Last good rowid = %d.\n", last_rowid );
			fflush( stdout );
			errlvl = 1;
			break;
		    } else {
		        if (Use_Key_Type == 1) 
			  strcpy( last_ifx_row_id, this_ifx_row_id );
		        else if (Use_Key_Type == 2) 
			  last_key = this_key;
		        else
			  last_rowid = this_rowid;
			if (!Silent) {
			  if (Use_Key_Type == 1) 
			    fprintf( stdout, "Committed %d rows through ifx_row_id: %s.\n", uncommitted, last_ifx_row_id );
			  else if (Use_Key_Type == 2) 
			    fprintf( stdout, "Committed %d rows through %s: %lld.\n", uncommitted, DelKey, last_key );
			  else
			    fprintf( stdout, "Committed %d rows through rowid: %d.\n", uncommitted, last_rowid );

			    fflush( stdout );
			}
		    }
		    if (begin_ok) {
			if (no_replicate)
			    EXEC SQL BEGIN WORK WITHOUT REPLICATION;
			else
			    EXEC SQL BEGIN WORK;
			if ( sqlca.sqlcode != 0 ) {
			    fprintf( stdout,
				     "Error: %d(%d) on BEGIN WORK at %d.\n",
				     sqlca.sqlcode,
				     sqlca.sqlerrd[1],
				     __LINE__ );
			    fflush( stdout );
			    if (Use_Key_Type == 1)
			      fprintf( stdout, "Last good ifx_row_id = %s.\n", last_ifx_row_id );
			    else if (Use_Key_Type == 2)
			      fprintf( stdout, "Last good %s = %lld.\n", DelKey, last_key );
			    else
			      fprintf( stdout, "Last good rowid = %d.\n", last_rowid );
			    fflush( stdout );
			    return 1;
			} else if (Debug) {
			    fprintf( stdout, "Beginning work at: %d.\n", __LINE__ );
			}
		    }
		}
		uncommitted = 0;
		continue;	/* Try again. */
	    }
	    break;
	} else if (sqlca.sqlcode == 100) {
	    if (!Silent) {
		fprintf( stdout, "No more rows to process.\n" );
		fflush( stdout );
	    }
	    break;
	} else {
	    fprintf( stdout, "Fetched %d %s's at line %d.\n", sqlca.sqlerrd[2], id_str, __LINE__ );
	}
	tries = 0;
	/* Save number of keys fetched this round. */
	fetchcount = sqlca.sqlerrd[2];
	if (Debug) {
	    fprintf( stdout, "%d rowid/ifx_row_ids fetched.\n", fetchcount );
	    fflush( stdout );
	}

	if (!one_open && is_open) {
	    if (Debug) {
	        fprintf( stdout, "Closing cursor at li %d.\n", __LINE__ );
		fflush( stdout );
	    }
	    EXEC SQL CLOSE fetch_curs;
	    is_open = 0;
	}

	col = udesc->sqlvar;
	col2 = cdesc->sqlvar;

	/* Copy by rowid/ifx_row_id. */
	i=0;
	while (i < fetchcount) {
	    dsiz = ip2 - copy_stmt;
	    ip = ip2;
	    for (; i < fetchcount && dsiz < SBsz; i++) {
	        if (Use_Key_Type == 1) {
		  strcpy( this_ifx_row_id, (char *)(col[0].sqldata+(i*col[0].sqllen)));
		  sprintf( tmp,
			   "'%s'",
			   this_ifx_row_id );
	        } else if (Use_Key_Type == 2) {
		  this_key = *(long long *)(col[0].sqldata+(i*col[0].sqllen));
		  sprintf( tmp,
			   "%lld",
			   this_key );
	        } else {
		  this_rowid = *(int *)(col[0].sqldata+(i*col[0].sqllen));
		  sprintf( tmp,
			   "%d",
			   this_rowid );
	        }
		strcpy( ip, tmp );
		ip += (tl = strlen( tmp ));
		dsiz += (tl + 1);

		if (i < (fetchcount - 1) && dsiz < SBsz)
		    strcpy( ip++, "," );
	    }
	    if (Debug) {
		fprintf( stdout, "IN string (%s)\n", ip2 );
		fflush( stdout );
	    }
	    while (dsiz < SBsz && fetchcount == 1) {
		/* Test code. */
		if (Debug) {
		    fprintf( stdout, "Inside && fetchcount == 1 loop!\n" );
		    fflush( stdout );
		}
		if (!is_open) {
		  fflush( stdout );
		  if (Debug) {
		    fprintf( stdout, "Opening cursor at li %d.\n", __LINE__ );
		    if (Use_Key_Type == 1)
		      fprintf( stdout, "   Using last_ifx_row_id: %s.\n", last_ifx_row_id );
		    else if (Use_Key_Type == 2)
		      fprintf( stdout, "   Using %s: %lld.\n", DelKey, last_key );
		    else
		      fprintf( stdout, "   Using rowid: %d.\n", last_rowid );
		    fflush( stdout );
		  }
		  if (Use_Key_Type == 1) {
		    if (!Silent && !Debug)
		      fprintf( stdout, "Opening cursor with ifx_row_id > %s at line %d.\n", last_ifx_row_id, __LINE__ );
		    EXEC SQL OPEN fetch_curs USING :last_ifx_row_id;
		  } else if (Use_Key_Type == 2) {
		    if (!Silent && !Debug)
		      fprintf( stdout, "Opening cursor with %s > %lld at line %d.\n", DelKey, last_key, __LINE__ );
		    EXEC SQL OPEN fetch_curs USING :last_key;
		  } else {
		    if (!Silent && !Debug)
		      fprintf( stdout, "Opening cursor with rowid > %d at line %d.\n", last_rowid, __LINE__ );
		    EXEC SQL OPEN fetch_curs USING :last_rowid;
		  }
		  is_open = 1;
		}
		EXEC SQL FETCH fetch_curs USING DESCRIPTOR udesc;

		if (sqlca.sqlcode < 0) {
		    fprintf( stdout,
			     "Error fetching row(%d)!  Code=%d, ISAM=%d at line %d.\n",
			     reccount,
			     sqlca.sqlcode,
			     sqlca.sqlerrd[1],
			     __LINE__ );
		    fflush( stdout );
		    if (sqlca.sqlcode == -243 && tries < 5)
			{
			    tries++;
			    if (!one_open && is_open) {
			        if (Debug) {
				    fprintf( stdout, "Closing cursor at li %d.\n", __LINE__ );
				    fflush( stdout );
				}
				EXEC SQL CLOSE fetch_curs;
				is_open = 0;
			    }
			    if ( sqlca.sqlcode != 0 ) {
				fprintf( stdout,
					 "Error: %d(%d) on CLOSE fetch_curs.\n",
					 sqlca.sqlcode,
					 sqlca.sqlerrd[1] );
				fflush( stdout );
				if (Use_Key_Type == 1)
				  fprintf( stdout, "Last good ifx_row_id = %s.\n", last_ifx_row_id );
				else if (Use_Key_Type == 2)
				  fprintf( stdout, "Last good %s = %lld.\n", DelKey, last_key );
				else
				  fprintf( stdout, "Last good rowid = %d.\n", last_rowid );
				fflush( stdout );
				return 1;
			    }

			    if (!is_open) {
				fflush( stdout );
			        if (Debug) {
				    fprintf( stdout, "Opening cursor at li %d.\n", __LINE__ );
				    if (Use_Key_Type == 1)
				      fprintf( stdout, "Using ifx_row_id = %s.\n", last_ifx_row_id );
				    else if (Use_Key_Type == 2)
				      fprintf( stdout, "Using %s = %lld.\n", DelKey, last_key );
				    else
				      fprintf( stdout, "Using rowid = %d.\n", last_rowid );
				    fflush( stdout );
				}
				if (Use_Key_Type == 1) {
				  if (!Silent && !Debug)
				    fprintf( stdout, "Opening cursor with ifx_row_id > %s at line %d.\n", last_ifx_row_id, __LINE__ );
				  EXEC SQL OPEN fetch_curs USING :last_ifx_row_id;
				} else if (Use_Key_Type == 1) {
				  if (!Silent && !Debug)
				    fprintf( stdout, "Opening cursor with %s > %lld at line %d.\n", DelKey, last_key, __LINE__ );
				  EXEC SQL OPEN fetch_curs USING :last_key;
				} else {
				  if (!Silent && !Debug)
				    fprintf( stdout, "Opening cursor with rowid > %d at line %d.\n", last_rowid, __LINE__ );
				  EXEC SQL OPEN fetch_curs USING :last_rowid;
				}
				is_open = 1;
			    }
			    if ( sqlca.sqlcode != 0 ) {
				fprintf( stdout,
					 "Error: %d(%d) on REOPEN fetch_curs.\n",
					 sqlca.sqlcode,
					 sqlca.sqlerrd[1] );
				fflush( stdout );
				if (Use_Key_Type == 1)
				  fprintf( stdout, "Last good ifx_row_id = %s.\n", last_ifx_row_id );
				else if (Use_Key_Type == 2)
				  fprintf( stdout, "Last good %s = %lld.\n", DelKey, last_key );
				else
				  fprintf( stdout, "Last good rowid = %d.\n", last_rowid );
				fflush( stdout );
				return 1;
			    }
			    if (uncommitted && commit_ok) {
				EXEC SQL COMMIT WORK;
				if ( sqlca.sqlcode != 0 ) {
				    fprintf( stdout,
					     "Error: %d(%d) on COMMIT WORK in loop.\n",
					     sqlca.sqlcode,
					     sqlca.sqlerrd[1] );
				    fflush( stdout );
				    if (Use_Key_Type == 1)
				      fprintf( stdout, "Last good ifx_row_id = %s.\n", last_ifx_row_id );
				    else if (Use_Key_Type == 2)
				      fprintf( stdout, "Last good %s = %lld.\n", DelKey, last_key );
				    else
				      fprintf( stdout, "Last good rowid = %d.\n", last_rowid );
				    fflush( stdout );
				    errlvl = 1;
				    break;
				} else {
				    if (Use_Key_Type == 1) 
				      strcpy( last_ifx_row_id, hold_ifx_row_id );
				    else if (Use_Key_Type == 2) 
				      last_key = hold_key;
				    else
				      last_rowid = hold_rowid;
				    if (!Silent) {
				        if (Use_Key_Type == 1)
					  fprintf( stdout, "Committed %d rows through ifx_row_id: %s.\n", uncommitted, last_ifx_row_id );
				        else if (Use_Key_Type == 2)
					  fprintf( stdout, "Committed %d rows through %s: %lld.\n", uncommitted, DelKey, last_key );
				        else
					  fprintf( stdout, "Committed %d rows through rowid: %d.\n", uncommitted, last_rowid );
					fflush( stdout );
				    }
				}
				if (begin_ok) {
				    if (no_replicate)
					EXEC SQL BEGIN WORK WITHOUT REPLICATION;
				    else
					EXEC SQL BEGIN WORK;
				    if ( sqlca.sqlcode != 0 ) {
					fprintf( stdout,
						 "Error: %d(%d) on BEGIN WORK at %d.\n",
						 sqlca.sqlcode,
						 sqlca.sqlerrd[1],
						 __LINE__ );
					fflush( stdout );
					if (Use_Key_Type == 1)
					  fprintf( stdout, "Last good ifx_row_id = %s.\n", last_ifx_row_id );
					else if (Use_Key_Type == 2)
					  fprintf( stdout, "Last good %s = %lld.\n", DelKey, last_key );
					else
					  fprintf( stdout, "Last good rowid = %d.\n", last_rowid );
					fflush( stdout );
					return 1;
				    } else if (Debug) {
					fprintf( stdout, "Begining work at: %d.\n", __LINE__ );
				    }
				}
			    }
			    uncommitted = 0;
			    continue;	/* Try again. */
			}
		    break;
		} else if (sqlca.sqlcode == 100) {
		    if (!Silent) {
			fprintf( stdout, "No more rows to process.\n" );
			fflush( stdout );
		    }
		    break;
		} else {
		    fprintf( stdout, "At line %d: %d rows fetched.\n", __LINE__, sqlca.sqlerrd[2] );
		}
		tries = 0;
		/* Save number of keys fetched this round. */
		i = 0;
		fetchcount = sqlca.sqlerrd[2];
		for (; i < fetchcount && dsiz < SBsz; i++) {
		    strcpy( ip++, "," );
		    /* Save last rowid processed this loop. */
		    if (Use_Key_Type == 1) {
		      strcpy( this_ifx_row_id, (char *)(col[0].sqldata+(i*col[0].sqllen)));
		      sprintf( tmp,
			       "'%s'",
			       this_ifx_row_id );
		    } else if (Use_Key_Type == 2) {
		      this_key = *(long long *)(col[0].sqldata+(i*col[0].sqllen));
		      sprintf( tmp,
			       "'%lld'",
			       this_key );
		    } else {
		      this_rowid = *(int *)(col[0].sqldata+(i*col[0].sqllen));
		      sprintf( tmp,
			       "%d",
			       this_rowid );
		    }
		    strcpy( ip, tmp );
		    ip += (tl = strlen( tmp ));
		    dsiz += (tl + 1);
		}
		if (Debug) {
		  if (Use_Key_Type == 1)
		    fprintf( stdout, "Leaving ifx_row_id loop.  This_ifx_row_id = %s.\n", this_ifx_row_id );
		  else if (Use_Key_Type == 2)
		    fprintf( stdout, "Leaving %s loop. This_key  = %lld.\n", DelKey, this_key );
		  else
		    fprintf( stdout, "Leaving rowid loop.  This_rowid = %d.\n", this_rowid );
		  fflush( stdout );
		}
	    }
	    strcpy( ip, " )" );
		
	    if (Debug) {
		fprintf( stdout, "List built: %s\n\n", copy_stmt );
		fflush( stdout );
	    }

	    EXEC SQL EXECUTE IMMEDIATE :copy_stmt;
	    if ( sqlca.sqlcode != 0 ) {
		fprintf( stdout,
			 "Error: %d(%d) on EXECUTE copystmt.\n",
			 sqlca.sqlcode,
			 sqlca.sqlerrd[1] );
		fflush( stdout );
		if (Use_Key_Type == 1)
		  fprintf( stdout, "Last good ifx_row_id = %s.\n", last_ifx_row_id );
		else if (Use_Key_Type == 2)
		  fprintf( stdout, "Last good %s = %lld.\n", DelKey, last_key );
		else
		  fprintf( stdout, "Last good rowid = %d.\n", last_rowid );
		fflush( stdout );
		return 1;
	    }
	    uncommitted = sqlca.sqlerrd[2];
	    counter += sqlca.sqlerrd[2];
	    if (Use_Key_Type == 1) 
	      strcpy( hold_ifx_row_id, this_ifx_row_id );
	    else if (Use_Key_Type == 2) 
	      hold_key = this_key;
	    else
	      hold_rowid = this_rowid;
	    if (uncommitted && commit_ok) {
		EXEC SQL COMMIT WORK;
		if ( sqlca.sqlcode != 0 ) {
		    fprintf( stdout,
			     "Error: %d(%d) on COMMIT WORK in loop.\n",
			     sqlca.sqlcode,
			     sqlca.sqlerrd[1] );
		    fflush( stdout );
		    if (Use_Key_Type == 1)
		      fprintf( stdout, "Last good ifx_row_id = %s.\n", last_ifx_row_id );
		    else if (Use_Key_Type == 2)
		      fprintf( stdout, "Last good %s = %lld.\n", DelKey, last_key );
		    else
		      fprintf( stdout, "Last good rowid = %d.\n", last_rowid );
		    fflush( stdout );
		    errlvl = 1;
		    break;
		} else {
		    if (Use_Key_Type == 1) 
		      strcpy( last_ifx_row_id, hold_ifx_row_id );
		    else if (Use_Key_Type == 2) 
		      last_key = hold_key;
		    else
		      last_rowid = hold_rowid;
		    if (!Silent) {
		        if (Use_Key_Type == 1)
			  fprintf( stdout, "Committed %d rows through ifx_row_id: %s.\n", uncommitted, last_ifx_row_id );
		        else if (Use_Key_Type == 2)
			  fprintf( stdout, "Committed %d rows through %s: %lld.\n", uncommitted, DelKey, last_key );
		        else
			  fprintf( stdout, "Committed %d rows through rowid: %d.\n", uncommitted, last_rowid );
			fflush( stdout );
		    }
		}
		if (begin_ok) {
		    if (no_replicate)
			EXEC SQL BEGIN WORK WITHOUT REPLICATION;
		    else
			EXEC SQL BEGIN WORK;
		    if ( sqlca.sqlcode != 0 ) {
			fprintf( stdout,
				 "Error: %d(%d) on BEGIN WORK at %d.\n",
				 sqlca.sqlcode,
				 sqlca.sqlerrd[1],
				 __LINE__ );
			fflush( stdout );
			if (Use_Key_Type == 1)
			  fprintf( stdout, "Last good ifx_row_id = %s.\n", last_ifx_row_id );
			else if (Use_Key_Type == 2)
			  fprintf( stdout, "Last good %s = %lld.\n", DelKey, last_key );
			else
			  fprintf( stdout, "Last good rowid = %d.\n", last_rowid );
			fflush( stdout );
			return 1;
		    } else if (Debug) {
			fprintf( stdout, "Beginning work at: %d.\n", __LINE__ );
		    }
		}
	    }
	    uncommitted = 0;
	}

	if (uncommitted && commit_ok) {
	    if (Debug) {
		fprintf( stdout,
			 "Commit outside loop triggered at: %d.  Uncommitted = %d.\n",
			 __LINE__,
			 uncommitted );
		fflush( stdout );
	    }
	    EXEC SQL COMMIT WORK;
	    if ( sqlca.sqlcode != 0 ) {
		fprintf( stdout,
			 "Error: %d(%d) on COMMIT WORK in loop.\n",
			 sqlca.sqlcode,
			 sqlca.sqlerrd[1] );
		fflush( stdout );
		if (Use_Key_Type == 1)
		  fprintf( stdout, "Last good ifx_row_id = %s.\n", last_ifx_row_id );
		else if (Use_Key_Type == 2)
		  fprintf( stdout, "Last good %s = %lld.\n", DelKey, last_key );
		else
		  fprintf( stdout, "Last good rowid = %d.\n", last_rowid );
		fflush( stdout );
		errlvl = 1;
		break;
	    } else {
	        if (Use_Key_Type == 1) 
		  strcpy( last_ifx_row_id, hold_ifx_row_id );
	        else if (Use_Key_Type == 2) 
		  last_key = hold_key;
	        else
		  last_rowid = hold_rowid;
		if (!Silent) {
		  if (Use_Key_Type == 1)
		    fprintf( stdout, "Committed %d rows through ifx_row_id: %s.\n", uncommitted, last_ifx_row_id );
		  else if (Use_Key_Type == 2)
		    fprintf( stdout, "Committed %d rows through %s: %lld.\n", uncommitted, DelKey, last_key );
		  else
		    fprintf( stdout, "Committed %d rows through rowid: %d.\n", uncommitted, last_rowid );
		  fflush( stdout );
		}
	    }
	    if (begin_ok) {
		if (no_replicate)
		    EXEC SQL BEGIN WORK WITHOUT REPLICATION;
		else
		    EXEC SQL BEGIN WORK;
		if ( sqlca.sqlcode != 0 ) {
		    fprintf( stdout,
			     "Error: %d(%d) on BEGIN WORK at %d.\n",
			     sqlca.sqlcode,
			     sqlca.sqlerrd[1],
			     __LINE__ );
		    fflush( stdout );
		    if (Use_Key_Type == 1)
		      fprintf( stdout, "Last good ifx_row_id = %s.\n", last_ifx_row_id );
		    else if (Use_Key_Type == 2)
		      fprintf( stdout, "Last good %s = %lld.\n", DelKey, last_key );
		    else
		      fprintf( stdout, "Last good rowid = %d.\n", last_rowid );
		    fflush( stdout );
		    return 1;
		} else if (Debug) {
		    fprintf( stdout, "Beginning work at: %d.\n", __LINE__ );
		}
	    }
	}
	reccount += fetchcount;
	uncommitted = 0;
	if (!Silent) {
            if (Use_Key_Type == 1)
              fprintf( stdout, "Last good ifx_row_id = %s.\n", last_ifx_row_id );
            else if (Use_Key_Type == 2)
              fprintf( stdout, "Last good %s = %lld.\n", DelKey, last_key );
            else
              fprintf( stdout, "Last good rowid = %d.\n", last_rowid );
	}
	fflush( stdout );
    } /* end fetch loop */

    if (!Silent) {
	fprintf( stdout, "Input:  %d records.\n", reccount );
	fprintf( stdout, "Output: %d records.\n", counter );
	fflush( stdout );
    }

    if (is_open) {
        if (Debug) {
	    fprintf( stdout, "Closing cursor at li %d.\n", __LINE__ );
	    fflush( stdout );
	}
        EXEC SQL CLOSE fetchcurs;
	is_open = 0;
    }
    if ( sqlca.sqlcode != 0 && sqlca.sqlcode != -404) {
	fprintf( stdout,
		 "Error: %d(%d) on CLOSE fetchcurs.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	fflush( stdout );
	if (Use_Key_Type == 1)
	  fprintf( stdout, "Last good ifx_row_id = %s.\n", last_ifx_row_id );
	else if (Use_Key_Type == 2)
	  fprintf( stdout, "Last good %s = %lld.\n", DelKey, last_key );
	else
	  fprintf( stdout, "Last good rowid = %d.\n", last_rowid );
	fflush( stdout );
	return 1;
    }    
   
    free(buffer);

    return errlvl;
}


void usage( char *progname )
{
    char *p;

    p = strrchr( progname, '/' );
    if (!p)
	p = progname;
    else
	p++;
    fprintf( stdout, "\n");
    fprintf( stdout,
	     "%s - Copy rows from one Informix table to another.\n\n",
	     p );

    fprintf( stdout, "Usage: %s\n", p );	
    fprintf( stdout, 
	     "\t\t<-b FetBufSize>\n"
	     "\t\t<-h source host>\n"
	     "\t\t<-d source database>\n"
	     "\t\t<-t source table>\n"
	     "\t\t<-c source column list>\n"
	     "\t\t<-H target host>\n"
	     "\t\t<-D target database>\n"
	     "\t\t<-T target table>\n"
	     "\t\t<-C target column list>\n"
	     "\t\t<-I > <-i last_ifx_rowid>\n"
	     "\t\t<-K keycol> <-k keyval>\n"
	     "\t\t<-r last rowid>\n"
	     "\t\t<-s WHERE>\n"
	     "\t\t<-S> <-V> <-R>\n"
	     "\t\t<-w waittime>\n"
	     "\t\t<-o>\n"
	     "\t\t<-B SQL bufsize>\n"
	     "\t\t<-p PDQRPIORITY>\n\n" );

    fprintf( stdout, "\t-h - Host to copy rows from (Default: $INFORMIXSERVER).\n" );
    fprintf( stdout, "\t-d - Database from which to copy rows.(Default: target database)\n" );
    fprintf( stdout, "\t-t - Table from which data is to be copied.(Default: target table)\n" );
    fprintf( stdout, "\t-H - Host to insert rows to (Default: source host).\n" );
    fprintf( stdout, "\t-D - Database to which to insert rows.(Default: source database)\n" );
    fprintf( stdout, "\t-T - Table to which data is to be inserted.(Default: source table)\n" );
    fprintf( stdout,
	     "\t-c - Comma separated list of columns to select from the source table. \n"
	     "\t     (Default: *)\n" );
    fprintf( stdout,
	     "\t-C - Comma separated list of columns to insert into the target table. \n"
	     "\t     (Default: Source column list)\n" );
    fprintf( stdout,
	     "\t-I - Use ifx_row_id instead of ROWID. Needed to copy from a partitioned\n"
	     "\t     table that was not created WITH ROWID.\n" );
    fprintf( stdout,
	     "\t-i - Specify last ifx_row_id processed to continue a crashed or killed\n"
	     "\t     session with the next ifx_row_id following.\n"
	     "\t     (Default: start with first ifx_row_id)\n" );
    fprintf( stdout,
	     "\t-K - Use supplied key column instead of ROWID. Needed to copy from a \n"
	     "\t     partitioned table that was not created WITH ROWID.\n" );
    fprintf( stdout,
	     "\t-k - Specify last key value processed to continue a crashed or killed\n"
	     "\t     session with the next key value following.\n"
	     "\t     (Default: start with lowest key value)\n" );
    fprintf( stdout,
	     "\t-b - Size of a transaction buffer.  Controls the number of rowids \n"
	     "\t     fetched in a single operation (FetBufSize/4). (Default: 32767. \n"
	     "\t     Range: 4096 -> 32767)\n" );
    fprintf( stdout,
	     "\t-B - Set the size of the SQL INSERT...SELECT statements that are \n"
	     "\t     generated.  This controls the number of rows copied in a single\n"
	     "\t     transaction. (Default: 2048)\n" );
    fprintf( stdout,
	     "\t-r - Specify last rowid processed to continue a crashed or killed \n"
	     "\t     session with the next rowid following.\n"
	     "\t     (Default: start with first rowid)\n" );
    fprintf( stdout,
	     "\t-s - Specify a WHERE clause filter to select ROWIDs to copy. Use -r \n"
	     "\t     instead to filter by rowid.\n" );
    fprintf( stdout, "\t-S - Silent mode.  Supress normal progress messages.\n" );
    fprintf( stdout, "\t-V - Print version information and exit.\n" );
    fprintf( stdout,
	     "\t-w - # of seconds to wait for locked rows to be released (if < 0 wait\n"
	     "\t     forever).  (Default 10.)\n" );
    fprintf( stdout,
	     "\t-o - Do not close the fetch cursor between fetch rounds.  This is\n"
	     "\t     slower, however, it will prevent dbmove from copying rows that\n"
	     "\t     have been inserted since the original cursor was opened.  Note:\n"
	     "\t     this option does NOT eliminate the dangers of using dbmove on\n"
	     "\t     a table that is being actively inserted to.  Rows inserted BEFORE\n"
	     "\t     the cursor is open may be copied.  Caviat emptor applies.\n" );
    fprintf( stdout, "\t-p - PDQPRIORITY to set for the task.\n" );
    fprintf( stdout,
	     "\t-R - Do not replicate insertions.  Begins transactions with the WITHOUT\n"
	     "\t     REPLICATION clause included.\n\n" );
    fprintf( stdout,
	     "The standard Informix environment variable FET_BUF_SIZE will override the \n"
	     "default value of the -b flag if within the valid range.  Higher values for -b\n"
	     "will fetch larger blocks of rowids at a time and briefly use more locks. \n\n" );
    fflush( stdout );
}
