/*
  Program Id: dbscript

  Program Description: this program will generate a script that will execute a 
  command against all tables in a database or a set of tables. 

  Author: Art S. Kagel from dbstruct.ec 

  Date: November 12, 2012

  Version: 1.0

  Comments:    

  Revisions: $Revsion$

Usage:
        dbscript [-h host] -d database [-t table] [-c [CMD|cmd_file] [filename]
                [-i table_1 [-i table_2 [...]]] [-x [table_a|dbase_a:] [-x...]] 
                [-i @[file1] [-i @[file2]]] [-x @[file1] [-x @[file2]]] [-G] 
                [-i '!where clause'|subquery] [-x '!where clause'] 
        dbscript -V

  RCS Header
  ----------
  $Header: /home/art/GoogleDrive/Latest/RCS/dbscript.ec,v 1.15 2021/12/21 17:12:40 art Exp $

  RCS Log
  -------
  $Log: dbscript.ec,v $
  Revision 1.15  2021/12/21 17:12:40  art
  Updated Copyright notice.

  Revision 1.14  2020/04/19 22:33:17  art
  Cleaned up usage.

  Revision 1.13  2020/04/19 22:28:16  art
  Added new options: -e, -v, -s, -S, & -n controlling what objects are
  processed. See usage for details.

  Revision 1.12  2020/03/20 16:40:46  art
  Changed the query populating hierarchy1 using CONNECT BY to be a 
  dynamic prepared statement to be compatible with CSDK 3.50.xC6 which
  should support it but doesn't, while IDS 11.50.xC6 does.

  Revision 1.11  2019/03/15 15:12:17  art
  Fixed compiler warning and updated inline function to work with newer gcc compilers.

  Revision 1.10  2016/06/27 21:17:44  art
  Updated the copyright notice.

  Revision 1.9  2015/07/16 21:22:24  art
  Added -N & -D options to sort tables by name or in dependency order respectively.

  Revision 1.8  2015/03/08 01:38:21  art
  Fixed usage in comments at the top.

  Revision 1.7  2013/12/29 21:36:25  art
  Fixed warnings.

  Revision 1.6  2013/12/25 14:01:40  art
  Fixed compile time warnings on HPUX.

  Revision 1.5  2013/12/24 17:12:57  art
   Fixed compile time warnings on HPUX.  Updated copyright notice.

  Revision 1.4  2013/07/09 16:18:24  art
  Fixed problem with old 18byte object names.  Added support for 128byte names.

  Revision 1.3  2013/05/23 12:54:06  art
  Added -i and -x handling from dostats to add flexibility.

  Revision 1.2  2013/02/10 22:18:19  art
  Updated Copyright

  Revision 1.1  2012/11/12 19:43:02  art
  Initial revision


*/
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <ctype.h>
#include <errno.h>
#if defined(__sysv__) || defined(i386) || defined(_AIX) || defined(__GNUC__) || defined(__hpux)
#include <string.h>
#else
#include <strings.h>
#endif

#include "getopt.h"

EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqltypes;
EXEC SQL INCLUDE datetime;

static char rcs_header[] = "$Header: /home/art/GoogleDrive/Latest/RCS/dbscript.ec,v 1.15 2021/12/21 17:12:40 art Exp $";

extern struct sqlca_s sqlca;

EXEC SQL BEGIN DECLARE SECTION;

struct { 
    string tabname[129];
} systables, prev;

char databasename[129], host[100], *statement = (char *)0, connection[256];
char tablename[129] = "";
int major_vers, minor_vers;
string level_vers[129], dbstmp[129], tmpstr[1025];

EXEC SQL END DECLARE SECTION;

extern int optind;
extern char *optarg;

#if !defined(SQLINFXBIGINT)
#define SQLINFXBIGINT 52
#define SQLBIGSERIAL 53
#endif

char *a_space;
char dtparts[7][10] = { "YEAR",
			"MONTH",
			"DAY",
			"HOUR",
			"MINUTE",
			"SECOND",
			"FRACTION" };

char *includes = (char *)NULL, *excludes = (char *)NULL;
char *subquery = (char *)NULL, *exquery = (char *)NULL;
char *dbexcludes = (char *)NULL;
int incl_len = 0, excl_len = 0, dbexcl_len = 0, qry_len = 0, xqry_len = 0;
int sawlistincl = 0, sawlistexcl = 0, sawsubincl = 0, sawsubexcl = 0, stmt_len = 0;

#define TBSFACT 16777216L
#if !defined( FALSE )
#define FALSE (0)
#endif
#if !defined( TRUE )
#define TRUE (!FALSE)
#endif

#if !defined( __GNUC__ )
#define strlower( t, s, m ) {					\
	char *target = (t), *source = (s); int max = (m);	\
	while (max-- && *source) {				\
	    *target++ = tolower(*source++);			\
	}							\
	*target = 0;						\
    }
#else
static inline void strlower( char *target, char *source, int max )
{
    while (max-- && *source) {
        *target++ = tolower(*source++);
    }
    *target = 0;
}
#endif

void rexit(int stat);
void usage( char *program );
void ucase( char *target, char *source );
#define STR_STEP 1024
void readlist( char **, int *, char **, int *, char * );
void appendStr( char **, int *, char * );

typedef enum Orders { TabidOrder, NameOrder, DependOrder } Orders_t;

int main(int argc, char *argv[])
{
    static int pass_no = 1;
    int i, j, c, errflag=0, dflag=0, tflag=0, hflag=0, allows_nulls;
    int do_external = 0, do_views = 0, do_synonyms = 0, do_sequences = 0, no_tables = 0;
    long colwidth, pt1, pt2, fr1, fr2;
    FILE *outfile = stdout, *infile;
    char filename[1025], filename2[1025];
    static char tempstr[1024];
    static int HaveFile = 0, HaveCmd = 0, Delimident_Set = 0;
    static Orders_t doOrder = TabidOrder;
    char instr[65536];
    size_t res;
    char *cres;

    prev.tabname[0] = (char)0;
    host[0] = (char)0;

    if (getenv( "DELIMIDENT" ) != (char *)NULL) {
        Delimident_Set = 1;
    }

    while ((c = getopt(argc, argv, "d:i:x:t:h:c:senvSV?DN")) != EOF){
        switch(c) {
	case 'D':
	    doOrder = DependOrder;
	    break;
	case 'N':
	    doOrder = NameOrder;
	    break;
	case 'S':
	    do_sequences = 1;
	    break;
	case 'V':
	    fprintf( stderr,
		     "dbscript.ec Revision: %s.  Copyright 1990-2022, Art S. Kagel\n",
		     "$Revision: 1.15 $" );
	    return 0;
	case 'c':
	    if (optarg[0] == '@') {
		HaveFile = 1;
		snprintf( filename2, sizeof filename2, "%s", &(optarg[1]) );
	    } else {
		HaveCmd = 1;
		snprintf( instr, sizeof instr, "%s", optarg );
	    }
	    break;
	case 'd':
	    if (dflag) {
		errflag++;
	    } else {
		dflag++;
		strncpy(databasename, optarg, sizeof databasename);
		databasename[sizeof databasename - 1] = (char)0;
	    }
	    break;
	case 'n':
	    no_tables = 1;
	    break;
	case 's':
	    do_synonyms = 1;
	    break;
	case 't':
	    if (tflag) {
		errflag++;
	    } else {
		tflag++;
		if (!Delimident_Set) {
		    strlower( tablename, optarg, sizeof tablename );
		} else {
		    strncpy(tablename, optarg, sizeof tablename);
		    tablename[sizeof tablename - 1] = (char)0;
		}
	    }
	    break;
	case 'v':
	    do_views = 1;
	    break;
	case 'e':
	    do_external = 1;
	    break;
	case 'h':
	    if (hflag) {
		errflag++;
	    } else {
		hflag++;
		strncpy(host, optarg, sizeof host);
		host[sizeof host - 1] = (char)0;
	    }
	    break;
	case 'i':
	    /* Form table include list. */
	    if (optarg[0] == '@') {
		readlist( &includes,
			  &incl_len,
			  (char **)NULL,
			  (int *) NULL,
			  optarg );
		sawlistincl = 1;
	    } else if (optarg[0] == '!') {
		if (strncasecmp( &(optarg[1]), "select", 6 ) != 0) 
		    appendStr( &subquery, &qry_len, "select tabname from systables where " );
		appendStr( &subquery, &qry_len, &(optarg[1]) );
		sawsubincl = 1;
	    } else {
		if (includes)
		    appendStr( &includes, &incl_len, "','" );
		else
		    appendStr( &includes, &incl_len, "('" );
		appendStr( &includes, &incl_len, optarg );
		sawlistincl = 1;
	    }
	    break;
	case 'x':
	    /* Form table exclude list. */
	    if (optarg[0] == '@') {
		readlist( &excludes,
			  &excl_len,
			  &dbexcludes,
			  &dbexcl_len,
			  optarg );
		sawlistexcl = 1;
	    } else if (optarg[0] == '!') {
		if (strncasecmp( &(optarg[1]), "select", 6 ) == 0) {
		    fprintf( stderr, "-x '!select...' not supported. Invert your filters and use -i!\n" );
		    usage( argv[0] );
		    exit( 1 );
		}
		appendStr( &exquery, &xqry_len, " AND a.tabname NOT IN (" );
		appendStr( &exquery, &xqry_len, "select tabname from systables where " );
		appendStr( &exquery, &xqry_len, &(optarg[1]) );
		appendStr( &exquery, &xqry_len, ") " );
		sawsubexcl = 1;
	    } else if (strchr( optarg, ':' )) {
		/* Database exclusion not supported. */
		fprintf( stderr, "-x dbase: not supported!\n" );
		usage( argv[0] );
		exit( 1 );
	    } else {
		/* Table exclusion. */
		if (excludes)
		    appendStr( &excludes, &excl_len, "','" );
		else
		    appendStr( &excludes, &excl_len, "('" );
		appendStr( &excludes, &excl_len, optarg );
		sawlistexcl = 1;
	    }
	    break;
	case '?':
	default:
	    errflag++;
	    break;
        }
    }

    if (optind < argc) {
	strncpy( filename, argv[optind], sizeof filename );
	filename[sizeof filename - 1] = (char)0;
	outfile = fopen( filename, "w" );
	if (!outfile) {
	    fprintf( stderr, "Cannot open output: %s.\n", filename );
	    exit(3);
	}
    }

    if (errflag || dflag == 0){
	usage( argv[0] );
        exit(1);
    }
    if(strcmp(databasename,"")==0 || ( tflag == 1 && strcmp(tablename,"")==0)){
	usage( argv[0] );
        exit(1);
    }

    if (includes && sawlistincl)
	appendStr( &includes, &incl_len, "') " );
    else if (includes)
	appendStr( &includes, &incl_len, ") " );
    
    if (excludes && sawlistexcl)
	appendStr( &excludes, &excl_len, "') " );
    else if (excludes)
	appendStr( &excludes, &excl_len, ") " );

    if (dbexcludes)
	appendStr( &dbexcludes, &dbexcl_len, "') " );

    if (includes && strcmp( tablename, "*" ) == 0) {
	fprintf( stderr, 
		 "Tablename template other than '*' required with includes!\n" );
	usage( argv[0] );
	exit( 1 );
    }

    /* fix up -t with/without -i/-x */
    //    if (!includes && strlen( tablename ) == 0)
    //	strcpy( tablename, "*" );


    strcpy( connection, databasename );
    if (hflag) {
	strcat( connection, "@" );
	strcat( connection, host );
	EXEC SQL
	    CONNECT TO :connection;
    } else {
	EXEC SQL DATABASE :connection;
    }
    if ( sqlca.sqlcode )
	{
	    fprintf( stderr,
		     "Cannot access host (%s) & database (%s); error=%d\n",
		     host,
		     databasename,
		     sqlca.sqlcode );
	    rexit(1);
	}

    /* Detect version. */
    EXEC SQL SELECT DBINFO( 'version', 'major' ),
	DBINFO( 'version', 'minor' ),
	DBINFO( 'version', 'level' )
	INTO :major_vers, :minor_vers, :level_vers
	FROM "informix".systables where tabid = 1;
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "FAILED: Error fetching server version.  SQL=%d, ISAM=%d.\n",
		 (int)sqlca.sqlcode,
		 (int)sqlca.sqlerrd[1] );
	exit( 1 );
    }

    if (doOrder == DependOrder &&
        !(major_vers > 11 ||
          ( major_vers == 11 && minor_vers > 50) ||
          ( major_vers == 11 && minor_vers == 50 &&
            (level_vers[0] > 'C' || (level_vers[0] == 'C' && level_vers[1] >= '5'))))) {
        fprintf( stderr,
                 "Dependency ordering not supported by this server.  Version 11.50.xC5 or later required!\n" );
        usage( argv[0] );
        rexit( 1 );
    }

    if (HaveFile) {
	infile = fopen( filename2, "r" );
	res = fread( instr, sizeof instr, 1, infile );
	if (ferror( infile ) ) {
	    fprintf( stderr, "Error reading command string file: %d.\n", errno );
	    exit( 2 );
	}
	if (strlen( instr ) == 0) exit(0);
	HaveCmd = 1;
    }
    if (!HaveCmd) {
	fprintf( stderr, "Cmd: " );
	cres = fgets( instr, sizeof instr, stdin );
	if (ferror( stdin ) || cres == (char *)NULL) {
	    fprintf( stderr, "Error reading command string file: %d.\n", errno );
	    exit( 2 );
	}
	if (strlen( instr ) == 0) exit(0);
	HaveCmd = 1;
    }
    if (instr[strlen(instr)-1] == '\n')
	instr[strlen(instr)-1] = (char)0;
    strcat( instr, "\n" );

    /* If dependecy-order specified, build the temp table for ordering. */
    if (doOrder == DependOrder) {
	EXEC SQL SELECT st.tabid ctabid, sr.ptabid ptabid, st.tabtype
	    FROM systables st, OUTER( sysreferences sr, sysconstraints sc)
	    WHERE  st.tabid = sc.tabid 
	    AND sc.constrid = sr.constrid
	    INTO TEMP hierarchy;
	if (sqlca.sqlcode != 0) {
	    fprintf( stderr, 
		     "Error building hierarchy map! Error: %d.\n",
		     sqlca.sqlcode );
	    exit(1);
	}
	EXEC SQL DELETE FROM hierarchy 
	    WHERE ctabid = ptabid;
	if (sqlca.sqlcode != 0) {
	    fprintf( stderr, 
		     "Error cleaning loops from hierarchy map! Error: %d.\n",
		     sqlca.sqlcode );
	    exit(1);
	}

	EXEC SQL SELECT FIRST 1 name
	INTO :dbstmp
	FROM sysmaster:sysdbstab
	WHERE bitand(flags,'0x2000') > 1;
	
	snprintf( tmpstr,
		  sizeof tmpstr,
		  "create temp table hierarchy1( ctabid int, ptabid int, tabtype char(1), hlevel int ) in %s;",
		  dbstmp );

	EXEC SQL EXECUTE IMMEDIATE :tmpstr;
	if (sqlca.sqlcode != 0) {
	    fprintf( stderr, 
		     "Error building hierarchy1! Error: %d.\n",
		     sqlca.sqlcode );
	    exit(1);
	}

	snprintf( tmpstr,
		  sizeof tmpstr,
		  "%s;",
		  "INSERT INTO hierarchy1 "
		  "SELECT ctabid, ptabid, tabtype, LEVEL as hlevel "
		  "FROM hierarchy "
		  "START WITH ptabid IS NULL "
		  "CONNECT BY PRIOR ctabid = ptabid "
		  "ORDER SIBLINGS BY ctabid " );
	EXEC SQL EXECUTE IMMEDIATE :tmpstr;
	if (sqlca.sqlcode != 0) {
	    fprintf( stderr, 
		     "Error organizing hierarchy map! Error: %d.\n",
		     sqlca.sqlcode );
	    exit(1);
	}
	EXEC SQL DROP TABLE hierarchy;
	EXEC SQL SELECT ctabid, ptabid, case when tabtype = 'V' then 999999999 else h1.hlevel end AS hlevel
		 FROM hierarchy1 AS h1
		 WHERE rowid = (
		 SELECT MAX(rowid)
		 FROM hierarchy1 AS h2
		 WHERE h2.ctabid = h1.ctabid
		 AND hlevel = (
		       SELECT MAX(hlevel)
		       FROM hierarchy1 AS h3
		       WHERE h3.ctabid = h2.ctabid
		       )
		 )
		 INTO TEMP hierarchy2;
	if (sqlca.sqlcode != 0) {
	    fprintf( stderr, 
		     "Error ordering hierarchy map! Error: %d.\n",
		     sqlca.sqlcode );
	    
	    exit(1);
	}
	EXEC SQL DROP TABLE hierarchy1;
    }

    /* Select for tables and columns. */
    appendStr( &statement, &stmt_len, "SELECT a.tabname FROM systables a " );
    if (doOrder == DependOrder)
	appendStr( &statement, &stmt_len, ", hierarchy2 as h " );

    if (major_vers >= 11 && subquery) {
	/* Use a dynamic table so ORDER BY is supported. */
	appendStr( &statement,
		   &stmt_len,
		   ", ( " );
	appendStr( &statement,
		   &stmt_len,
		   subquery );
	appendStr( &statement,
		   &stmt_len,
		   ") as flt " );
    }
    if (no_tables == 1)
      appendStr( &statement, &stmt_len, "WHERE a.tabid >= 100 AND ( a.tabtype IS NULL " ); /* Dummy test */
    else
      appendStr( &statement, &stmt_len, "WHERE a.tabid >= 100 AND ( a.tabtype = 'T' " );
    if (do_external == 1)
        appendStr( &statement, &stmt_len, " OR a.tabtype = 'E' " );
    if (do_views == 1)
        appendStr( &statement, &stmt_len, " OR a.tabtype = 'V' " );
    if (do_synonyms == 1)
      appendStr( &statement, &stmt_len, " OR a.tabtype = 'S' " );
    if (do_sequences == 1)
      appendStr( &statement, &stmt_len, " OR a.tabtype = 'Q' " );
    appendStr( &statement, &stmt_len, " )" );
    
    if (strlen( tablename ) > 0)
	appendStr( &statement, &stmt_len, "AND a.tabname MATCHES ? " );
    if (tflag && strcmp( tablename, "none" ) != 0) {
	if (includes || (subquery && major_vers < 11))
	    appendStr( &statement,
		       &stmt_len,
		       "AND (a.tabname MATCHES '" );
	else
	    appendStr( &statement,
		       &stmt_len,
		       "AND a.tabname MATCHES '" );

	appendStr( &statement, &stmt_len, tablename );
	appendStr( &statement, &stmt_len, "' " );
    } else if (includes || (subquery && major_vers < 11)) {
	appendStr( &statement, &stmt_len, "AND (a.tabname IS NULL " );
    }
    if (subquery && major_vers < 11) {
	appendStr( &statement, &stmt_len, " OR a.tabname IN (" );
	appendStr( &statement, &stmt_len, subquery );
	appendStr( &statement, &stmt_len, ") " );
	appendStr( &statement, &stmt_len, ") " );
    } else if (subquery) {
	appendStr( &statement, &stmt_len, "AND a.tabname = flt.tabname " );
    }
    if (includes ) {
	appendStr( &statement, &stmt_len, "OR a.tabname IN " );
	appendStr( &statement, &stmt_len, includes );
	appendStr( &statement, &stmt_len, ") " );
    }
    if (exquery) {
	appendStr( &statement, &stmt_len, exquery );
    }
    if (excludes) {
	appendStr( &statement,
		   &stmt_len,
		   "AND a.tabname NOT IN " );
	appendStr( &statement, &stmt_len, excludes );
    }

    switch (doOrder) {
    case TabidOrder:
	appendStr( &statement, &stmt_len, "ORDER BY a.tabid; " );
	break;
    case NameOrder:
	appendStr( &statement, &stmt_len, "ORDER BY a.tabname; " );
	break;
    case DependOrder:
	appendStr( &statement, &stmt_len, "AND a.tabid = h.ctabid " );
	appendStr( &statement, &stmt_len, "ORDER BY h.hlevel, a.tabid; " );
	break;
    }

    EXEC SQL
	PREPARE stmt1 FROM :statement;
    if (sqlca.sqlcode) {
	fprintf( stderr,
		 "Cannot PREPARE select statement for table data, error: %d.\n",
		 sqlca.sqlcode );
	fprintf( stderr, "Error at character %d.\nStatement is:\n",
		 sqlca.sqlerrd[4] );
	fprintf( stderr, "%s\n", statement );
	rexit(2);
    }

    EXEC SQL
	DECLARE a1 CURSOR FOR stmt1;
    if (sqlca.sqlcode ) {
        fprintf( stderr,
		 "Cannot declare a1 cursor; Error = %d\n", sqlca.sqlcode );
        rexit(1);
    }

    if (strlen( tablename ) > 0)
	EXEC SQL
	    OPEN a1 USING :tablename;
    else
	EXEC SQL
	    OPEN a1;
    if (sqlca.sqlcode == SQLNOTFOUND) {
	fprintf( stderr, "No table matching \"%s\".\n", tablename );
	rexit(3);
    } else if ( sqlca.sqlcode ) {
        fprintf( stderr, "Cannot open a1 cursor; Error = %d\n", sqlca.sqlcode );
	fprintf( stderr, "Statement: %s;\n", statement );
	fprintf( stderr, "Tablename: %s.\n", tablename );
        rexit(1);
    }

    i = 0;

    for(;;) {
        EXEC SQL
	    fetch a1 into
            :systables.tabname;

        if (sqlca.sqlcode == SQLNOTFOUND)
	    break;
	else if (sqlca.sqlcode && sqlca.sqlcode != SQLNOTFOUND ) {
	    fprintf( stderr,
		     "Error fetching a1 cursor;Error=%d\n",
		     sqlca.sqlcode);
	    rexit(1);
        } else {
	    fprintf( outfile,
		     instr,
		     systables.tabname );
	}
	memcpy( &prev, &systables, sizeof prev );
    }
}

void rexit(int stat)
{

    if (stat) {
        EXEC SQL
	    close a1;
        if (sqlca.sqlcode && sqlca.sqlcode != -404) {
            fprintf( stderr,
		     "cannot close a1 cursor;Error=%d\n",
		     sqlca.sqlcode);
        }
    }
    EXEC SQL
	CLOSE DATABASE;
    EXEC SQL
	DISCONNECT ALL;
    exit(stat);
}

void usage( char *program )
{
    fprintf( stderr,
	     "Usage:\n\t%s [-h host] -d database [-t table] [-c [CMD|cmd_file] [filename]\n",
	     program );
    fprintf( stderr,
	     "\t\t[-i table_1 [-i table_2 [...]]] [-x [table_a|dbase_a:] [-x...]] \n"
	     "\t\t[-i @[file1] [-i @[file2]]] [-x @[file1] [-x @[file2]]] [-G] \n"
	     "\t\t[-i '!where clause'|subquery] [-x '!where clause'] \n"
	     "\t\t[-D|-N] [-n] [-v] [-e] [-s] [-S' \n" );
    fprintf( stderr,
	     "\t%s -V\n", program );
    fprintf( stderr,
	     "\t-V - Display revision and copyright and exit.\n" );
    fprintf( stderr,
	     "\t-c - Specify a brief commandline to execute for each table, or the path\n"
	     "\t     to a file containing the commandline. If not present, the command \n"
	     "\t     is read from stdin.\n" );
    fprintf( stderr,
	     "\t-t - By default all 'normal' tables are processed. Use -t with a MATCHES \n"
	     "\t   wildcard specification to limit the tables processed. \n" );
    fprintf( stderr,
	     "\t-e - Include EXTERNAL TABLES in the 'table' names processed.\n" );
    fprintf( stderr,
	     "\t-v - Include VIEWS in the 'table' names processed.\n" );
    fprintf( stderr,
	     "\t-s - Include SYNONYMS in the 'table' names processed.\n" );
    fprintf( stderr,
	     "\t-S - Include SEQUENCES in the 'table' names processed.\n" );
    fprintf( stderr,
	     "\t-n - Exclude 'normal' tables from processing. Useful only when one or \n"
	     "\t     more of -s, -e, -v, or -S are included. \n" );
    fprintf( stderr,
	     "\t-i Specify tables to forceably include in the tables processed.  One\n"
	     "\t   may enter a single table, the @ sign followed by the name of a file\n"
	     "\t   containing a list of tables one per line, or an exclamation point (!)\n"
	     "\t   followed by either a WHERE filter to be applied to systables to \n"
	     "\t   select tables for inclusion, or a complete sub-query beginning with \n"
	     "\t   the keyword 'select' which will return a list of valid table names \n"
	     "\t   to include.  To aid script use if the argument to -i is simply '@' \n"
	     "\t   stdin is read.  Multiple -i options are concatenated to form a net \n"
	     "\t   list.  Note that all -i options can be used together.  NOTE: The \n"
	     "\t   included SELECT statement is now processed as a derived table with\n"
	     "\t   v11.xx.  ORDER BY and other SQL features which were restricted in \n"
	     "\t   sub-queries used in earlier releases of Informix are supported when\n"
	     "\t   connecting to v11 and later!\n" );
    fprintf( stderr,
	     "\t-x Specify tables or databases to forceably exclude from those\n"
	     "\t   processed.  One may enter a single table, the @ sign\n"
	     "\t   followed by the name of a file containing a list of tables and/or\n"
	     "\t   databases, one per line, or an exclamation point (!) followed by a\n"
	     "\t   WHERE filter to be applied to systables to select tables to be\n"
	     "\t   excluded.  To aid script use if the argument to -x is simply '@' \n"
	     "\t   stdin is read.  Multiple -x options are concatenated to form a net \n"
	     "\t   list.  Note that all -x options can be used together.\n" );
    fprintf( stderr,
	     "\t-D - Process tables in dependency order, parents before children.\n"
	     "\t-N - Process tables alphabetically. (Default: tabid order)\n" );
    fprintf( stderr, "\tHost defaults to current host.\n" );
    fprintf( stderr,
	     "\tTable may contain valid \"matches\" clause wildcards.\n" );
    fprintf( stderr,
	     "\tTable defaults to all user tables %s\n\n",
	     "(excluding system catalog tables.)" );
    fprintf( stderr,
	     "The command to be executed may contain a single '%%s' which will be replaed by\n"
	     "each tablename that is processed.  For more complex command strings, save the\n"
	     "command(s) into a shell script that takes the table name as an argument passed\n"
	     "using the '%%s'.\n\n" );
}

void ucase( char *target, char *source )
{
    while (*source)
	*target++ = toupper( *source++ );
    *target = (char)0;
}


void appendStr( char **str, int *str_size, char *newstr )
{
    int sum_len;

    if (*str == NULL) {
	*str_size = STR_STEP;
	*str = (char *) malloc( STR_STEP );
	**str = 0;
    }

    sum_len = strlen(*str) + strlen(newstr) + 1;
    if (sum_len > *str_size) {
	/* round up sum_len to multiple of STR_STEP */
	/* optimiser won't eliminate this - it would change the meaning */
	*str_size = ((sum_len + STR_STEP - 1) / STR_STEP) * STR_STEP;

	*str = (char *)realloc( *str, *str_size );

	if (! *str) {
	    fprintf( stderr, "FAILED\nError allocating memory errno = %d.\n", errno );
	    exit( 255 );
	}
    }
    strcat(*str, newstr);
}

void readlist( char **list, int *len, char **dblist, int *dblen, char *file )
{
    FILE *fil;
    char string[1025];

    if ((int)strlen( file ) == 1)
	fil = stdin;
    else {
	fil = fopen( &file[1], "rt" );
	if (fil == (FILE *)NULL) {
	    fprintf( stderr, "Cannot open list file: %s.\n", &file[1] );
	    exit( 19 );
	}
    }

    if (fil == stdin) {
	fputs( "table> ", stderr );
	fflush( stderr );
    }
    while (fgets( string, 1025, fil )) {
	/* remove trailing newline */
	string[(int)strlen( string ) - 1] = (char)0;

	/* new feature: skip blank lines and comments */
	/* TODO: skip blank lines with blanks and allow leading space? */
	if (! *string || *string == '#')
	    continue;

	if (strchr( string, ':' )) {
	    /* database exclude. */
	    if (dblist == (char **) NULL) {
		fprintf( stderr,
			 "Databases not supported in include lists!\n" );
		exit(20);
	    }

	    if (*dblist)
		appendStr( dblist, dblen, "','" );
	    else
		appendStr( dblist, dblen, " AND name NOT IN ('" );

	    appendStr( dblist, dblen, strtok( string, ":" ) );
	} else {
	    /* Table include/exclude. */
	    if (*list)
		appendStr( list, len, "','" );
	    else
		appendStr( list, len, "('" );

	    appendStr( list, len, string );
	}
	if (fil == stdin) {
	    fputs( "table> ", stderr );
	    fflush( stderr );
	}
    }

    if ((int)strlen( file ) > 1)
	fclose( fil );
}
