/* printfree.ec - Calculate Informix free space within a tables extents.

   Written by Art S. Kagel, version written 3/26/1998
*/
/*
   RCS Header:
   -----------
   $Header: /home/art/GoogleDrive/Latest/RCS/printfreeB.ec,v 1.13 2021/12/21 17:12:40 art Exp $

   RCS Log:
   --------
   $Log: printfreeB.ec,v $
   Revision 1.13  2021/12/21 17:12:40  art
   Updated Copyright notice.

   Revision 1.12  2016/06/27 21:23:36  art
   Updated copyright notice.

   Revision 1.11  2013/12/29 21:40:49  art
   Updated copyright notice.

   Revision 1.10  2013/12/24 18:06:51  art
   More compiler warnings fixed.

   Revision 1.9  2013/02/10 22:18:19  art
   Updated Copyright

   Revision 1.8  2011/05/23 16:05:32  art
   Updated the copyright statement.

   Revision 1.7  2004/12/14 20:16:32  kagel
   Checking last changes - bug fix

   Revision 1.6  2004/04/30 15:13:09  kagel
   Removed RCSWhat.

   Revision 1.5  2004/04/30 15:11:27  kagel
   Fixed uninitialized var prob.

   Revision 1.4  2003/10/13 17:16:34  kagel
   Resolved compiled problems stemming from string prototypes in string.h
   instead of strings.h.

   Revision 1.3  2001/10/24 19:32:39  kagel
   Added -V version and copyright printout.

   Revision 1.2  2000/12/05 13:20:14  kagel
   *** empty log message ***

   Revision 1.1  1998/05/07 15:17:15  kagel
   Initial revision

# Revision 1.11.1.6  1998/04/23  15:16:13  kagel
# Incorporated fixes from 1.11.2.5 including support for detached indexes
# and improved fragment report headers.
#
# Revision 1.11.1.4  1998/03/31  16:42:16  kagel
# Completed.  Too SLOWWWWWWWW.  Waiting for that select from sysptnbit
# take much too long.  Also reformated output.  Will port new output to the
# main branch.
#
# Revision 1.11.1.3  1998/03/30  22:49:22  kagel
# Changed to used SQL to add up bitmap buckets.  FASTER!
# Still does not work for fragmented tables!  Keep at it.
#
# Revision 1.11.1.2  1998/03/30  21:32:46  kagel
# Finished recoding and basic testing.  This works for simple tables but
# is MUCH slower.  Still, no Informix secrets revealed.
#
# Revision 1.11.1.1  1998/03/26  22:39:46  kagel
# Base checkin for new branch.  This branch to use sysptnbit for bitmap
# information rather than read bitmap pages directly.

*/

static char RCSHeader[] = "$Header: /home/art/GoogleDrive/Latest/RCS/printfreeB.ec,v 1.13 2021/12/21 17:12:40 art Exp $";

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#if defined(linux) || defined(i386)
#include <string.h>
#else
#include <strings.h>
#endif
#include <errno.h>

#define LENMASK 0x7FFF

EXEC SQL BEGIN DECLARE SECTION;
typedef struct partn_s {
    long partnum;             /* partition number */
    short flags;              /* flags                        */
    short pctfree;            /* percentage free space        */
    short rowsize;            /* uncompressed row size        */
    short ncols;              /* number columns in table      */
    short nkeys;              /* number keys for table        */
    short nextns;             /* number extents in partition  */
    long created;             /* date/time created            */
    long serial;              /* current serial value         */
    long fextsize;            /* first extent size            */
    long nxtsize;             /* subsequent extents size      */
    long nptotal;             /* number pages total           */
    long npused;              /* number pages ever used       */
    long npdata;              /* number pages of data         */
    long octptnm;             /* partnum of the optical cluster table */
    long lockid;              /* lock id to use               */
    long nrows;               /* number rows in table         */
} partn_t;

typedef struct spaces_s {
    short dbsnum, first_chunk, nchunks;
    string name[129];
} spaces_t;

typedef struct SYSPTNBIT_S {
        long pb_partnum;
        long pb_bitmap;
        long count;
} sysptnbit_t;

typedef struct tabptns_s {
    string fragtype[2];
    string indexname[129];
    long partnum;
} tabptns_t;

EXEC SQL END DECLARE SECTION;

#define VARCHARS     0x0100  /* partition contains VARCHAR's         */
#define BLOBS        0x0200  /* partition contains BlobSpace Blob    */
#define PBLOBS       0x0400  /* partition contains Partition Blob    */
#define OBLOBS       0x1000  /* partition contains Optical Blob      */
#define USE4BITS     0x0800  /* Has 4 bit bitmap-VARCHAR,Blobs,Long Rows */

typedef enum { datapage = 1, partition = 2, bitmap = 4, chunk = 8,
	       remainder = 9, psblob = 11, bsblob=12, blbbit=13, blbmap=14,
	       btnode=16, btroot=32, bttwig=64, btleaf=128, llog=256,
	       llogend=512, llogsync=768, plog=1024, rsrvd=4096, noplog=8192,
	       nophyslog=16384, btdelf=32768, btree = 48 } pageflags; 

char pagetypes[][12] = { "", "DATA", "PARTITION", "", "BITMAP", "", "", "", 
		    "CHUNK", "REMAINDER", "", "TSBLOB",
		    "BSBLOB", "BSBITMAP", "BSBLOBMAP", "", "BTREENODE", "", "",
		    "", "", "", "", "", "",
		    "", "", "", "", "", "", "", "BTREEROOT", "", "", "", "", "",
		    "", "", "", "", "", "", "", "", "", "", "BTREE"
		};

typedef enum { unused_4 = 0, hompart_4 = 4, idxfull_4 = 8,  homfull_4 = 12,
		rememt_4 = 1, rempart_4 = 5, remsmal_4 = 9,  remfull_4 = 13,
		blbemt_4 = 2, blbpart_4 = 6, blbsmal_4 = 10, blbfull_4 = 14 }
bit4vals;

typedef struct Partno_u {
    long partno;
    long dbspaceno, pageno;
} PartNo_t;

int main( int argc, char **argv )
{
    long buckets[20], nbitmaps, fragno;
    long spaceidx;
    spaces_t spaces[1000];
    PartNo_t PartNo;
    int ndbs = 0, ii, usebit4 = 0, ix;
    long nentries;
    int printdetail = 0, printdebug = 0, freepages;
    char *env;
    char Revision[] = "$Revision: 1.13 $";

    EXEC SQL BEGIN DECLARE SECTION;
    string *dbase, *table;
    spaces_t aspace;
    long nfrags, totpages, nexts;
    partn_t apartition;
    short partind;
    sysptnbit_t bitent;
    tabptns_t *tabptns;
    EXEC SQL END DECLARE SECTION;

    env = getenv( "PRINTFREE" );
    if (env && strstr( env, "detail" ) != (char *)NULL)
	printdetail = 1;
    env = getenv( "PRINTFREE" );
    if (env && strstr( env, "debug" ) != (char *)NULL)
	printdebug = 1;

    for (ix=1; ix < argc; ix++) {
	if (strcmp( argv[ix], "-V" ) == 0) {
	    fprintf( stderr,
		     "\n%s: %s Copyright 1998-2022 by Art S. Kagel.\n",
		     argv[0],
		     strtok(Revision, "$"));
	    fprintf( stderr,
		     "\tLicense for private usage (including use by "
		     "commercial enterprise\n"
		     "\tfor its own needs) is granted.  Commercial and "
		     "exploitation rights\n"
		     "\tare reserved.\n\n" );
	    exit(0);
	}
    }

    if (argc != 3) {
	printf( "Usage: %s database tablename\n", argv[0] );
	exit(1);
    }

    dbase = argv[1];
    table = argv[2];

    printf( "Looking at DB: %s, Table: %s.\n", dbase, table );
    
    /* First get all fragments for the table. */

    EXEC SQL DATABASE :dbase;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Cannot open database: %s. Err: %d, ISAM: %d\n",
		 dbase,
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit(3);
    }

    EXEC SQL SELECT COUNT(*)
	INTO :nfrags
	FROM systables t, outer sysfragments f
	WHERE t.tabid = f.tabid
	  AND t.tabname = :table;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Cannot FETCH fragment count. Err: %d, ISAM: %d\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit(3);
    }

    tabptns = (tabptns_t *)malloc( (nfrags + 1) * sizeof (tabptns_t) );

    EXEC SQL DECLARE tfrags_curs CURSOR FOR
	SELECT f.fragtype, f.indexname, f.partn
	FROM systables t, sysfragments f
	WHERE  t.tabid = f.tabid
	  AND t.tabname = :table
        UNION
        SELECT "T", "", partnum
        FROM systables
        WHERE tabname = :table
          AND partnum != 0
	ORDER BY 1 desc, 2, 3;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Cannot DECLARE fragment cursor. Err: %d, ISAM: %d\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit(3);
    }

    EXEC SQL OPEN tfrags_curs;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Cannot OPEN fragment cursor. Err: %d, ISAM: %d\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit(3);
    }

    for (ii=0;;ii++) {
	EXEC SQL FETCH tfrags_curs
	    INTO :tabptns[ii].fragtype,
		:tabptns[ii].indexname,
		:tabptns[ii].partnum;
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Cannot FETCH fragment cursor. Err: %d, ISAM: %d\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit(3);
	} else if (sqlca.sqlcode == SQLNOTFOUND)
	    break;
    }
    EXEC SQL CLOSE tfrags_curs;
    EXEC SQL FREE tfrags_curs;
    
    /* OK now look in SYSMASTER for partition information for each fragment. */

    EXEC SQL DATABASE sysmaster;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Cannot open database: %s.\n",
		 dbase );
	exit(3);
    }
    /* Get dbspace information. */
    EXEC SQL DECLARE dbs_curs CURSOR FOR
	SELECT dbsnum, fchunk, nchunks, name
	FROM sysdbspaces
	ORDER BY dbsnum;
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Cannot DECLARE cursor dbs_curs! Err=%d, ISAM=%d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit( 2 );
    }

    EXEC SQL OPEN dbs_curs;
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Cannot OPEN cursor dbs_curs! Err=%d, ISAM=%d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit( 2 );
    }
	
    memset( &spaces, 0, sizeof spaces );
    for (;;) {
	EXEC SQL FETCH dbs_curs
	INTO :aspace.dbsnum,
	    :aspace.first_chunk,
	    :aspace.nchunks,
	    :aspace.name;
	if (sqlca.sqlcode == SQLNOTFOUND)
	    break;
	else if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Cannot FETCH cursor dbs_curs! Err=%d, ISAM=%d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit( 2 );
	}
	spaces[aspace.dbsnum].dbsnum = aspace.dbsnum;
	spaces[aspace.dbsnum].first_chunk = aspace.first_chunk;
	spaces[aspace.dbsnum].nchunks = aspace.nchunks;
	strcpy( spaces[aspace.dbsnum].name, aspace.name );
	ndbs = aspace.dbsnum;  /* Save maximum dbs number for diagnostics. */
    }
    EXEC SQL CLOSE dbs_curs;

    if (printdetail)
	for (ii=0; ii < ndbs; ii++)
	    if (spaces[ii].dbsnum)
		fprintf( stderr,
			 "DBS# %d  FCHNK# %d  NCHKS: %d\n",
			 spaces[ii].dbsnum,
			 spaces[ii].first_chunk,
			 spaces[ii].nchunks );

    EXEC SQL SELECT COUNT(*)
	INTO :nfrags
	FROM systabnames
	WHERE dbsname = :dbase
	      AND tabname = :table;

    EXEC SQL DECLARE prt_curs CURSOR FOR
	SELECT partnum
	FROM systabnames
	WHERE dbsname = :dbase
	      AND tabname = :table;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr, "Cannot DECLARE cursor prt_curs.\n" );
	exit(3);
    }
    EXEC SQL OPEN prt_curs;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr, "Cannot OPEN CURSOR prt_curs.\n" );
	exit(3);
    }

    for ( fragno=0;fragno < nfrags;fragno++) {
	bzero( buckets, sizeof buckets );
	/* Get the table's partition header from sysmaster to find
	   flag indicating bitmap type. */

	EXEC SQL SELECT *
	    INTO
	    :apartition.partnum,
	    :apartition.flags,
	    :apartition.pctfree,
	    :apartition.rowsize,
	    :apartition.ncols,
	    :apartition.nkeys,
	    :apartition.nextns,
	    :apartition.created,
	    :apartition.serial,
	    :apartition.fextsize,
	    :apartition.nxtsize,
	    :apartition.nptotal,
	    :apartition.npused,
	    :apartition.npdata,
	    :apartition.octptnm,
	    :apartition.lockid,
	    :apartition.nrows
	    FROM sysptnhdr
	    WHERE partnum = :tabptns[fragno].partnum
	    ;

	usebit4 = (apartition.flags & USE4BITS)
	    || (apartition.flags & BLOBS)
	    || (apartition.flags & PBLOBS)
	    || (apartition.flags & OBLOBS);
    
	if (printdetail) {
	    fprintf( stderr,
		     "Tablespace header:\n\t"
		     "partnum  = 0x%x\n\t"
		     "flags    = 0x%hx\n\t"
		     "pctfree  = %hd\n\t"
		     "rowsize  = %hd\n\t"
		     "ncols    = %hd\n\t"
		     "nkeys    = %hd\n\t"
		     "nextns   = %hd\n\t"
		     "created  = %d\n\t"
		     "serial   = %d\n\t"
		     "fextsize = %d\n\t"
		     "nextsize = %d\n\t"
		     "nptotal  = %d\n\t"
		     "npused   = %d\n\t"
		     "npdata   = %d\n\t"
		     "octprnm  = 0x%x\n\t"
		     "lockid   = 0x%x\n\t"
		     "nrows    = %u\n\t"
		     "usebit4 = %d\n\n",
		     (unsigned int)apartition.partnum,
		     apartition.flags,
		     apartition.pctfree,
		     apartition.rowsize,
		     apartition.ncols,
		     apartition.nkeys,
		     apartition.nextns,
		     (int)apartition.created,
		     (int)apartition.serial,
		     (int)apartition.fextsize,
		     (int)apartition.nxtsize,
		     (int)apartition.nptotal,
		     (int)apartition.npused,
		     (int)apartition.npdata,
		     (unsigned int)apartition.octptnm,
		     (unsigned int)apartition.lockid,
		     (unsigned int)apartition.nrows,
		     usebit4 );
	}
	if (usebit4) {
	    nentries = 4032;
	} else {
	    nentries = 8064;
	}

	PartNo.partno = apartition.partnum;
	PartNo.dbspaceno = PartNo.partno >>20;
	PartNo.pageno = PartNo.partno & 0xFFFFF;
	spaceidx = PartNo.dbspaceno;
	if (tabptns[fragno].fragtype[0] == 'T') {
	    printf( "\fReport for table: %s:%s in dbspace #%d: %s.\n\n",
		    dbase,
		    table,
		    (int)spaceidx,
		    spaces[spaceidx].name );
	    printf( "    Table partition header reports that %s has: \n"
		    "\t%d%% free\n"
		    "\t%d pages allocated in %d extents\n"
		    "\t%d pages used\n"
		    "\t%d rows of data in %d data pages\n",
		    (nfrags == 1 ? "table" : "table fragment"),
		    apartition.pctfree,
		    (int)apartition.nptotal,
		    apartition.nextns,
		    (int)apartition.npused,
		    (int)apartition.nrows,
		    (int)apartition.npdata );
	} else if (tabptns[fragno].fragtype[0] == 'I') {
	    printf( "\fReport for index: %s:%s.%s in dbspace #%d: %s.\n\n",
		    dbase,
		    table,
		    tabptns[fragno].indexname,
		    (int)spaceidx,
		    spaces[spaceidx].name );
	    printf( "    Index partition header reports that %s has: \n"
		    "\t%d%% free\n"
		    "\t%d pages allocated in %d extents\n"
		    "\t%d pages used\n"
		    "\t%d rows of data in %d data pages\n",
		    (nfrags == 1 ? "index" : "index fragment"),
		    apartition.pctfree,
		    (int)apartition.nptotal,
		    apartition.nextns,
		    (int)apartition.npused,
		    (int)apartition.nrows,
		    (int)apartition.npdata );
	} else {
	    printf( "    Unknown fragment type: '%s'.\n",
		    tabptns[fragno].fragtype );
	}

	if (printdetail)
	    fprintf( stderr,
		     "Read partnum=%lx, eval to %lx:%lx.\n",
		     PartNo.partno,
		     (long)PartNo.dbspaceno,
		     (long)PartNo.pageno );

	if (spaceidx > ndbs) {
	    fprintf( stderr,
		     "Database no: %ld not in dbspace list!\n",
		     (long)PartNo.dbspaceno );
	    exit(2);
	} else
	    if (printdetail)
		fprintf( stderr,
			 "Found it: %x == %lx, chunkno = %d.\n",
			 (unsigned int)spaces[spaceidx].dbsnum,
			 PartNo.dbspaceno,
			 spaces[spaceidx].first_chunk );

	/* Read in extent count. */

	EXEC SQL SELECT count(*), sum(pe_size)
	    INTO :nexts, :totpages
	    FROM sysptnext
	    WHERE pe_partnum = :apartition.partnum;
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Cannot SELECT extent count! Err=%d, ISAM=%d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit( 2 );
	}

	printf( "    Sysptnext reports:\t %ld pages in %ld extents.\n",
		totpages,
		nexts );
    
	EXEC SQL DECLARE bit_curs CURSOR FOR
	    SELECT pb_partnum, pb_bitmap, count(*)
	    FROM sysptnbit
	    WHERE pb_partnum = :apartition.partnum
	    GROUP BY pb_partnum, pb_bitmap
	    ORDER BY pb_partnum, pb_bitmap;
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Cannot DECLARE cursor bit_curs! Err=%d, ISAM=%d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit( 2 );
	}

	EXEC SQL OPEN bit_curs;
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Cannot OPEN cursor bit_curs! Err=%d, ISAM=%d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit( 2 );
	}

	for (;;) {
	    EXEC SQL FETCH bit_curs
		INTO
		    :bitent.pb_partnum,
		    :bitent.pb_bitmap,
		    :bitent.count;
	    if (sqlca.sqlcode < 0 ) {
		fprintf( stderr,
			 "Cannot FETCHING cursor bit_curs!  Err=%d, ISAM=%d.\n",
			 sqlca.sqlcode,
			 sqlca.sqlerrd[1] );
		exit( 2 );
	    } else if (sqlca.sqlcode == SQLNOTFOUND)
		break;

	    /* Bucket this entry appropriately. */
	    buckets[bitent.pb_bitmap] = bitent.count;
	}
	nbitmaps = (totpages + nentries - 1) / nentries;

	freepages = totpages - (nbitmaps + buckets[hompart_4]
				+ buckets[rempart_4] + buckets[blbpart_4]
				+ buckets[remsmal_4] + buckets[blbsmal_4]
				+ buckets[homfull_4] + buckets[remfull_4]
				+ buckets[idxfull_4] + buckets[blbfull_4]);
	printf( "    Bitmap scan reports:\n" );
	printf( "\tUnused pages: \t\t%10d.\n", freepages );
	printf( "\tBitmap pages: \t\t%10ld.\n", nbitmaps );
	if (usebit4)
	    printf( "\tUnused blob pages: \t%10ld.\n", buckets[blbemt_4] );
	printf( "\tPartial data pages: \t%10ld.\n",
		buckets[hompart_4] + buckets[rempart_4] );
	if (usebit4) {
	    printf( "\tPartial blob pages: \t%10ld.\n", buckets[blbpart_4] );
	    printf( "\tSmall data pages: \t%10ld.\n", buckets[remsmal_4] );
	    printf( "\tHalf full blob pages: \t%10ld.\n", buckets[blbsmal_4] );
	}
	printf( "\tFull data pages: \t%10ld.\n",
		buckets[homfull_4] + buckets[remfull_4] );
	printf( "\tFull index pages: \t%10ld.\n", buckets[idxfull_4] );
	if (usebit4)
	    printf( "\tFull blob pages: \t%10ld.\n", buckets[blbfull_4] );
	printf( "\t\t\t\t-----------\n" );
	printf( "\tTotal pages reported: \t%10ld.\n\n",
		buckets[unused_4] +
		buckets[rememt_4] +
		buckets[blbemt_4] +
		buckets[hompart_4] +
		buckets[rempart_4] +
		buckets[blbpart_4] +
		buckets[remsmal_4] +
		buckets[blbsmal_4] +
		buckets[homfull_4] +
		buckets[remfull_4] +
		buckets[idxfull_4] +
		buckets[blbfull_4] +
	        nbitmaps +
	        freepages );
	if (buckets[3])
	    printf( "Buckets[3] = %10ld.\n", buckets[3] );
	if (buckets[7])
	    printf( "Buckets[7] = %10ld.\n", buckets[7] );
	if (buckets[11])
	    printf( "Buckets[11] = %10ld.\n", buckets[11] );
	if (buckets[15])
	    printf( "Buckets[15] = %10ld.\n", buckets[15] );
	printf( "\n\n" );
	fflush( stdout );
    }

    return 0;
}

